package ReproduceBioTransformerResults;

import java.io.File;
import java.util.ArrayList;

import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.inchi.InChIGenerator;
import org.openscience.cdk.inchi.InChIGeneratorFactory;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smarts.SmartsPattern;
import org.openscience.cdk.smiles.SmiFlavor;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import ambit2.smarts.SMIRKSManager;
import ambit2.smarts.SMIRKSReaction;
import createSomDataset.CollectData_convertSMIRKS_to_SDF;

public class CompareWithGroundTruth {
	CollectData_convertSMIRKS_to_SDF cd = new CollectData_convertSMIRKS_to_SDF();
	
	public static void main(String[] args) throws Exception{
		CompareWithGroundTruth gwgt = new CompareWithGroundTruth();
		gwgt.generateResults_UGT();		
		//gwgt.generateResults_SULT();
		
	}
	
	public void generateResults_SULT() throws Exception{
		IAtomContainerSet predicted_metabolites_SULT = loadPredictedMetabolites_SULT();
		IAtomContainerSet groudTruth_SULT = loadTrueMetabolites("SULT");
		ArrayList<String> predictedInChIKeyList = new ArrayList<>();
		//ArrayList<String> groundTruthInChIKeyList = new ArrayList<>();
		InChIGeneratorFactory inchiFac = InChIGeneratorFactory.getInstance();
		int TP = 0, FP = 0, TN = 0, FN = 0;
		for(int i = 0; i < predicted_metabolites_SULT.getAtomContainerCount(); i++) {
			InChIGenerator inchiGen = inchiFac.getInChIGenerator(predicted_metabolites_SULT.getAtomContainer(i));
			String inChIKey_14 = inchiGen.getInchiKey().split("-")[0];
			if(!predictedInChIKeyList.contains(inChIKey_14)) predictedInChIKeyList.add(inChIKey_14);		
				
		}
		for(int i = 0; i < groudTruth_SULT.getAtomContainerCount(); i++) {
			InChIGenerator inchiGen = inchiFac.getInChIGenerator(groudTruth_SULT.getAtomContainer(i));
			String inChIKey_14 = inchiGen.getInchiKey().split("-")[0];
			if(predictedInChIKeyList.contains(inChIKey_14)) TP++;
			
		}
		FP = predicted_metabolites_SULT.getAtomContainerCount() - TP;
		FN =  groudTruth_SULT.getAtomContainerCount() - TP;
		TN = getTotalAtomNum("SULT") - FN;
		System.out.println("   P           N");
		System.out.println("P " + TP + "          " + FN);
		System.out.println("N " + FP + "          " + TN);
	}
	
	
	
	public void generateResults_UGT() throws Exception{
		IAtomContainerSet predicted_metabolites_UGT = loadPredictedMetabolites_UGT();
		IAtomContainerSet groudTruth_UGT = loadTrueMetabolites("UGT");
		
		ArrayList<String> predictedInChIKeyList = new ArrayList<>();
		//ArrayList<String> groundTruthInChIKeyList = new ArrayList<>();
		InChIGeneratorFactory inchiFac = InChIGeneratorFactory.getInstance();
		int TP = 0, FP = 0, TN = 0, FN = 0;
		for(int i = 0; i < predicted_metabolites_UGT.getAtomContainerCount(); i++) {
			InChIGenerator inchiGen = inchiFac.getInChIGenerator(predicted_metabolites_UGT.getAtomContainer(i));
			String inChIKey_14 = inchiGen.getInchiKey().split("-")[0];
			if(!predictedInChIKeyList.contains(inChIKey_14)) predictedInChIKeyList.add(inChIKey_14);		
				
		}
		for(int i = 0; i < groudTruth_UGT.getAtomContainerCount(); i++) {
			InChIGenerator inchiGen = inchiFac.getInChIGenerator(groudTruth_UGT.getAtomContainer(i));
			String inChIKey_14 = inchiGen.getInchiKey().split("-")[0];
			if(predictedInChIKeyList.contains(inChIKey_14)) TP++;
			
		}
		FP = predicted_metabolites_UGT.getAtomContainerCount() - TP;
		FN =  groudTruth_UGT.getAtomContainerCount() - TP;
		TN = getTotalAtomNum("UGT") - FN;
		System.out.println("   P           N");
		System.out.println("P " + TP + "          " + FN);
		System.out.println("N " + FP + "          " + TN);
	}
	
	public int getTotalAtomNum(String type) throws Exception{
		int counter = 0;
		String inputPath;
		if(type.equalsIgnoreCase("UGT")) {
			inputPath = "G:/PhaseIIPredictor/UGT/Testing.sdf";
		}
		else if(type.equalsIgnoreCase("SULT")){
			inputPath = "G:/PhaseIIPredictor/Sulfation/Sulfation_test.sdf";
		}
		else throw new Exception("Must specify the type of phaseII reaction");
		
		IAtomContainerSet molecules =  this.cd.readSDFFile(inputPath);
		for(int i = 0; i < molecules.getAtomContainerCount(); i++) {
			IAtomContainer oneMole = molecules.getAtomContainer(i);
			for(int j = 0; j < oneMole.getAtomCount(); j++) {
				if(!oneMole.getAtom(j).getSymbol().equalsIgnoreCase("H")) counter++;
			}
		}
		return counter;	
	}
	/**
	 * This function will read the the Testing_metabolite.sdf for UGT conjugation reaction
	 * @return
	 * @throws Exception
	 */
	public IAtomContainerSet loadTrueMetabolites(String type) throws Exception{
		IAtomContainerSet results_temp;
		if(type.equalsIgnoreCase("UGT")){
			String inputPath = "G:/PhaseIIPredictor/UGT/Testing_metabolite.sdf";
			results_temp = this.cd.readSDFFile(inputPath);
			//results_temp = generateMetabolite_Special(results, this.cd.UGT_SMARTS);
			//return results;
		}
		else if(type.equalsIgnoreCase("SULT")){
			String inputPath = "G:/PhaseIIPredictor/Sulfation/Sulfation_test_metabolite.sdf";
			results_temp = this.cd.readSDFFile(inputPath);
			//return this.cd.readSDFFile(inputPath);
		}
		else throw new Exception("Please specify the phaseII reaction type");

		return results_temp;
		
	}
	
	public IAtomContainerSet loadPredictedMetabolites_UGT() throws Exception{
		ArrayList<String> checkedInChIKey = new ArrayList<>();
		InChIGeneratorFactory inchiFac = InChIGeneratorFactory.getInstance();
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		String inputFolder = "E:/Users/Tian/git/biotransformerjar/PhaseII/test/";
		File folder = new File(inputFolder);
		File[] listFile = folder.listFiles();
		for(int i = 0; i < listFile.length; i++) {
			File oneFile = listFile[i];
			IAtomContainerSet predictedMetabolites = this.cd.readSDFFile(oneFile.getAbsolutePath());
			for(int j = 0; j < predictedMetabolites.getAtomContainerCount(); j++) {
				IAtomContainer oneMetabolite = predictedMetabolites.getAtomContainer(j);
				InChIGenerator inchiGen = inchiFac.getInChIGenerator(oneMetabolite);
				String inChIKey = inchiGen.getInchiKey();
				if(!checkedInChIKey.contains(inChIKey)) checkedInChIKey.add(inChIKey);
				else continue;
				if(this.cd.matchSMARTS(oneMetabolite, this.cd.UGT_SMARTS)) {
					resultSet.addAtomContainer(oneMetabolite);
				}
						
			}
		}
		return resultSet;
	}
	
	public IAtomContainerSet loadPredictedMetabolites_SULT() throws Exception{
		String SULFATE_SMARTS = "[H][#8]S([#8])(=O)=O";
		ArrayList<String> checkedInChIKey = new ArrayList<>();
		InChIGeneratorFactory inchiFac = InChIGeneratorFactory.getInstance();
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		String inputFolder = "E:/Users/Tian/git/biotransformerjar/PhaseII/SULT_test/";
		File folder = new File(inputFolder);
		File[] listFile = folder.listFiles();
		for(int i = 0; i < listFile.length; i++) {
			File oneFile = listFile[i];
			IAtomContainerSet predictedMetabolites = this.cd.readSDFFile(oneFile.getAbsolutePath());
			for(int j = 0; j < predictedMetabolites.getAtomContainerCount(); j++) {
				IAtomContainer oneMetabolite = predictedMetabolites.getAtomContainer(j);
				InChIGenerator inchiGen = inchiFac.getInChIGenerator(oneMetabolite);
				String inChIKey = inchiGen.getInchiKey();
				if(!checkedInChIKey.contains(inChIKey)) checkedInChIKey.add(inChIKey);
				else continue;
				if(this.cd.matchSMARTS(oneMetabolite, SULFATE_SMARTS)) {
					resultSet.addAtomContainer(oneMetabolite);
				}
						
			}
		}
		return resultSet;
	}

	

}
