package biotransformerapiphaseii;



import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;

import phase2utils.Utilities;
import weka.core.Instances;


public class BioTransformerAPI_phaseII {
	PredictMetabolites pm;
	
	public static void main(String[] args) throws Exception{
		BioTransformerAPI_phaseII p2Predictor = new BioTransformerAPI_phaseII();
		String inputPath = "G:/PhaseIIPredictor/UGT/withNS/UGT_test.sdf";
		IAtomContainerSet molecules = Utilities.parseSdf(inputPath);
		IAtomContainerSet results = p2Predictor.runPredictions(molecules);
		//IAtomContainerSet results = p2Predictor.runOnePrediction(molecules.getAtomContainer(0));
		for(int i = 0; i < results.getAtomContainerCount(); i++) {			
			System.out.println((String) results.getAtomContainer(i).getProperty("ReactionType") + "; " + results.getAtomContainer(i).getProperty("Enzyme") + "; " + results.getAtomContainer(i).getProperty("Score"));
		}
	}
	
	public BioTransformerAPI_phaseII() throws Exception{
		this.pm = new PredictMetabolites();
	}
	/**
	 * This function makes prediction for a single input molecule. Please note that the current PhaseII predictor only makes predictions for UGT and SULF reactions
	 * Because the makePrediction function in the predictor only takes IAtomContainerSet as input, here, we wrap the single input molecule with a IAtomContainerSet
	 * @param oneMole
	 * @return
	 * @throws Exception
	 */
	public IAtomContainerSet runOnePrediction(IAtomContainer oneMole) throws Exception{
		IAtomContainerSet substrates = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class); 
		substrates.addAtomContainer(oneMole);
		//IAtomContainerSet metabolites_ugt = this.pm.makePrediction(substrates, "UGT");
		//IAtomContainerSet metabolites_sulf = this.pm.makePrediction(substrates, "SULF");
		//resultSet.add(metabolites_ugt);
		//resultSet.add(metabolites_sulf);
		resultSet = this.pm.makePrediction_faster(substrates);				
		return resultSet;
	}
	/**
	 * This function will run predictions for a input set of molecules. Please note that the current PhaseII predictor only makes predictions for UGT and SULF reactions
	 * @param molecules
	 * @return
	 * @throws Exception
	 */
	public IAtomContainerSet runPredictions(IAtomContainerSet molecules) throws Exception{
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class); 
		//IAtomContainerSet metabolites_ugt = this.pm.makePrediction(molecules, "UGT");
		//IAtomContainerSet metabolites_sulf = this.pm.makePrediction(molecules, "SULF");
		//resultSet.add(metabolites_ugt);
		//resultSet.add(metabolites_sulf);
		resultSet = this.pm.makePrediction_faster(molecules);
		return resultSet;
	}
	
	


}
