package biotransformerapiphaseii;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;

import createSomDataset.CollectData_convertSMIRKS_to_SDF;
import createSomDataset.PrepareCV;
import generateInstance.GenerateInstances;
import generateInstance.GenerateRawFeatures;
import phase2utils.GenerateTitleline;
import weka.classifiers.Classifier;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;

public class CreateInstances {
	Classifier ugt_classifier;
	Classifier sulf_classifier;
	Double ugt_threshold;
	Double sulf_threshold;
	ArrayList<String> ugt_attributes = new ArrayList<String>();
	ArrayList<String> sulf_attributes = new ArrayList<>();
	
	int depth = 4;
	GenerateInstances generateInstances = new GenerateInstances(depth);
	GenerateRawFeatures generateRawFeatures = new GenerateRawFeatures(depth);
	PrepareCV pc = new PrepareCV();	
	CollectData_convertSMIRKS_to_SDF cd = new CollectData_convertSMIRKS_to_SDF();
	
	public CreateInstances() throws Exception{
		loadModelAndAttributes("UGT");
		loadModelAndAttributes("SULF");
	}
	public void loadModelAndAttributes(String type) throws Exception{
		String modelPath = "/" + type + "/" + type + ".model";  
		String supportPath = "/" + type + "/" + type + "_supportfiles.csv";
		loadModel(modelPath);
		loadAttributesAndBeta(supportPath);
	}
	
	public Instances getInstances(IAtomContainer oneMole) throws Exception{
		String titleLine = GenerateTitleline.generateTitle(this.depth);	
		ArrayList<Attribute> attList = this.generateInstances.createAttributes(titleLine);		
		ArrayList<String> rawDataList = new ArrayList<>();
		Instances bomInstances = new Instances("Rel", attList, 100000); 
		ArrayList<String> rawFeatureList = new ArrayList<>();
		try {
			rawFeatureList = this.generateRawFeatures.generateRawFeatures(oneMole);
			}catch (Exception e) 
		{
				//e.printStackTrace();				
				System.out.println("Failed processing molecule : " + this.cd.sg.create(oneMole));
				IAtomContainer converted = this.cd.processNitrogen(oneMole);
				if(converted == null || converted.isEmpty()) System.out.println("Failed to convert the nitrogen contained functional group");
				else{
					System.out.println("Converted: " + this.cd.sg.create(converted));
					try {
						rawFeatureList = this.generateRawFeatures.generateRawFeatures(converted);		
					}catch(Exception e2) {
						System.out.println("Failed to generate raw features for the converted structure");
					}
				}
			
							
		}
		rawDataList.addAll(rawFeatureList);	
		bomInstances = this.generateInstances.createInstances(rawDataList, titleLine);
		return bomInstances;
	}
	
	/**
	 * Load the model obtained from the path to the model file
	 * @param modelPath
	 * @throws Exception
	 */
	public void loadModel(String modelPath) throws Exception{			
		InputStream model_Stream = getClass().getResourceAsStream(modelPath);		
		Classifier cls = (Classifier) weka.core.SerializationHelper.read(model_Stream);//Load the model
		if(modelPath.contains("UGT")) {			
			this.ugt_classifier = cls;
		}
		else if(modelPath.contains("SULF")) {
			this.sulf_classifier = cls;
		}
		else throw new Exception("The PhaseII model type can not be specified");
	}
	
	public void loadAttributesAndBeta(String supportPath) throws Exception{
		InputStream supportPath_Stream = getClass().getResourceAsStream(supportPath);
		BufferedReader br = new BufferedReader(new InputStreamReader(supportPath_Stream));
		String oneLine = br.readLine();//prob threshold
		if(supportPath.contains("UGT")) {
			this.ugt_threshold = Double.parseDouble(oneLine);
			while((oneLine = br.readLine())!=null){			
				this.ugt_attributes.add(oneLine);
			}
		}
		else if(supportPath.contains("SULF")) {
			this.sulf_threshold = Double.parseDouble(oneLine);
			while((oneLine = br.readLine())!=null){			
				this.sulf_attributes.add(oneLine);
			}
		}
		else {
			br.close();
			throw new Exception("The PhaseII Type can not be specified");
		}
		br.close();		
	}
	
	/**
	 * Reformat the instances of the input molecules so they can be input to the learned models by applying feature selection and normalization
	 * @param testSet: The instances of the user's input molecules
	 * @param attList: The attributes that are used in the learned model
	 * @param meanList: The mean values of each attribute
	 * @param maxList: The max values of each attribute
	 * @param minList: The min values of each attribute
	 * @return
	 * @throws Exception
	 */
	public Instances matchAttributes(Instances testSet, ArrayList<String> attListToMatch) throws Exception{
		/*
		 * Create the class atrribute
		 */
		List my_nominal_values = new ArrayList(3); 
		my_nominal_values.add("F"); 
		my_nominal_values.add("T"); 
		Attribute lastAttribute  = new Attribute(attListToMatch.get(attListToMatch.size()-1), my_nominal_values);
		/*
		 * Convert the attribute name strings to attribute
		 */
		ArrayList<Attribute> attsToMatch = new ArrayList<Attribute>();
		for(int i = 0; i<attListToMatch.size();i++){
			if(i<attListToMatch.size()-1){
				Attribute attribute = new Attribute(attListToMatch.get(i));
				attsToMatch.add(attribute);
			}
			else{
				
				attsToMatch.add(lastAttribute);
			}
		}

		int numAttributes = attListToMatch.size();
		Instances matched = new Instances("Rel", attsToMatch, testSet.size());
		Instance sample = new DenseInstance(numAttributes);
		/*
		 * Assign values for each testing instance
		 */
		for(int idxInstance = 0; idxInstance< testSet.size(); idxInstance++){
			//ArrayList misFeatures = new ArrayList();
			Instance temp = testSet.get(idxInstance);
			String label;
			//System.out.println("Check Predictor temp.classValue() == " + temp.classValue());
			if(temp.classValue()==0.0){
				label = "F";
			}
			else{
				label = "T";
			}
			
			for(int idxAttribute = 0; idxAttribute < numAttributes; idxAttribute++){
				  Attribute oneAttribute = attsToMatch.get(idxAttribute);
				  if(idxAttribute<numAttributes-1){					  
					  String nameOfAtt = oneAttribute.toString();
					  String[] formatAtt = nameOfAtt.split(" ");
					  
					  if(formatAtt.length!=3){
						  String combineSpaces = "";
						  for(int kk = 1; kk<formatAtt.length-1; kk++){
							  if(kk<formatAtt.length - 2){
								  combineSpaces = combineSpaces + formatAtt[kk]+ " ";
							  }
							  else combineSpaces = combineSpaces + formatAtt[kk];
						  }
						  formatAtt[1] = combineSpaces;
					  }
					 
					  if(formatAtt[1].contains("'")){
						  formatAtt[1] = formatAtt[1].replace("'", "");
					  }
					  //System.out.println(formatAtt[1]);
					  Attribute convAtt = testSet.attribute(formatAtt[1]);
					  double vle = temp.value(convAtt);
					  if(!Double.isNaN(vle)){					
						  sample.setValue(oneAttribute, vle);
					  }

				  }
				  else{
					  //Set class with random value. Does not matter
					  sample.setValue(lastAttribute, label);
					  
				  }
			}						
			matched.add(sample);
		}
		if(matched.classIndex() == -1){
			matched.setClassIndex(matched.numAttributes()-1);
		}
		
		return matched;
		
	}
	
}
