package biotransformerapiphaseii;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;

import org.openscience.cdk.Bond;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.inchi.InChIGenerator;
import org.openscience.cdk.inchi.InChIGeneratorFactory;
import org.openscience.cdk.interfaces.IAtom;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.interfaces.IBond;
import org.openscience.cdk.smiles.SmiFlavor;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import phase2utils.UniqueIDFunctionSet;
import phase2utils.Utilities;
import weka.classifiers.Classifier;
import weka.core.Instances;

public class PredictMetabolites {
	CreateInstances ci;
	SmilesParser sp = new SmilesParser(DefaultChemObjectBuilder.getInstance());
	SmilesGenerator sg = new SmilesGenerator(SmiFlavor.Isomeric);
	GenerateMetaboliteStructure gms = new GenerateMetaboliteStructure();
	DecimalFormat df = new DecimalFormat("#.##"); 
//	public static void main(String[] args) throws Exception{
//		PredictMetabolites pm = new PredictMetabolites();
//		//String inputPath = "G:/PhaseIIPredictor/UGT/withNS/UGT_training.sdf"; 
//		//String inputPath = "G:/PhaseIIPredictor/Sulfation/withNS/Sulfation_training.sdf";
//		
////		String type = "SULF";
////		String inputPath = "G:/PhaseIIPredictor/Sulfation/withNS/Sulfation_validation.sdf";
//		String type = "UGT";
//		String inputPath = "G:/PhaseIIPredictor/UGT/withNS/UGT_test.sdf";
//		IAtomContainerSet molecules = Utilities.parseSdf(inputPath);
//	
//		pm.makePrediction(molecules, type);
//		
//	}
	public PredictMetabolites() throws Exception{
		this.ci = new CreateInstances();				
	}
	
	/**
	 * This function assumes the user wants to make predictions for both Sulfonation and UGT, and thus doesn't take reaction type as parameter.
	 * @param molecules
	 * @return
	 * @throws Exception
	 */
	public IAtomContainerSet makePrediction_faster(IAtomContainerSet molecules) throws Exception{
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		InChIGeneratorFactory inChIFac = InChIGeneratorFactory.getInstance();
		ArrayList<String> productInChIKey = new ArrayList<>();
		
		for(int i = 0; i < molecules.getAtomContainerCount(); i++) {
			//if(i < 33) continue;
			IAtomContainer oneMole = molecules.getAtomContainer(i).clone();
			Instances instances = this.ci.getInstances(oneMole);
			Instances matchedInstances_ugt = this.ci.matchAttributes(instances, this.ci.ugt_attributes);
			Instances matchedInstances_sulf = this.ci.matchAttributes(instances, this.ci.sulf_attributes);
			for(int j = 0; j < instances.numInstances(); j++) {
				String atom_idx = instances.attribute("Bond").value(j);				
				double prob_ugt = this.ci.ugt_classifier.distributionForInstance(matchedInstances_ugt.get(j))[1];
				double prob_sulf = this.ci.sulf_classifier.distributionForInstance(matchedInstances_sulf.get(j))[1];
				//System.out.println(atom_idx + ": " + prob);
				if(prob_ugt >= this.ci.ugt_threshold) {
					double score = (prob_ugt - this.ci.ugt_threshold) / (1.0 - this.ci.ugt_threshold);
					//System.out.println("Molecule idx: " + i + " ; reactive site: " + atom_idx);
					//String[] check = atom_idx.split("\\.");
					IAtomContainer oneResult = this.gms.generateMetabolite(Integer.parseInt(atom_idx.split("\\.")[1]), oneMole.clone(), "UGT");
					oneResult.setProperty("Score", Double.valueOf(this.df.format(score)));
					oneResult.setProperty("Enzyme", "UGT");
					oneResult.setProperty("ReactionType", "Glucuronidation");
					InChIGenerator inChIGen = inChIFac.getInChIGenerator(oneResult);
					String inChIKey = inChIGen.getInchiKey();
					if(!productInChIKey.contains(inChIKey)) {
						resultSet.addAtomContainer(oneResult);
						productInChIKey.add(inChIKey);
					}					
				}
				if(prob_sulf >= this.ci.sulf_threshold) {
					double score = (prob_sulf - this.ci.sulf_threshold) / (1.0 - this.ci.sulf_threshold);
					//System.out.println("Molecule idx: " + i + " ; reactive site: " + atom_idx);
					//String[] check = atom_idx.split("\\.");
					IAtomContainer oneResult = this.gms.generateMetabolite(Integer.parseInt(atom_idx.split("\\.")[1]), oneMole.clone(), "SULF");
					oneResult.setProperty("Score", Double.valueOf(this.df.format(score)));
					oneResult.setProperty("Enzyme", "SULTs");
					oneResult.setProperty("ReactionType", "Sulfonation");
					InChIGenerator inChIGen = inChIFac.getInChIGenerator(oneResult);
					String inChIKey = inChIGen.getInchiKey();
					if(!productInChIKey.contains(inChIKey)) {
						resultSet.addAtomContainer(oneResult);
						productInChIKey.add(inChIKey);
					}					
				}
			}
		}
		return resultSet;
	}
	/**
	 * This function allows the user to select UGT or SULF reaction type 
	 * @param molecules
	 * @param type
	 * @return
	 * @throws Exception
	 */
	public IAtomContainerSet makePrediction(IAtomContainerSet molecules, String type) throws Exception{
		Classifier cls;
		ArrayList<String> attributes;
		Double threshold;
		String enzyme;
		String reactionType;
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		InChIGeneratorFactory inChIFac = InChIGeneratorFactory.getInstance();
		ArrayList<String> productInChIKey = new ArrayList<>();
		if(type.equals("UGT")) {
			cls = this.ci.ugt_classifier;
			attributes = this.ci.ugt_attributes;
			threshold = this.ci.ugt_threshold;
			enzyme = "UGT";
			reactionType = "Glucuronidation";
		}
		else if(type.equals("SULF")){
			cls = this.ci.sulf_classifier;
			attributes = this.ci.sulf_attributes;
			threshold = this.ci.sulf_threshold;
			enzyme = "SULTs";
			reactionType = "Sulfonation";
		}
		else throw new Exception("Unknown phase II reaction type. Please select SULF or UGT");
		for(int i = 0; i < molecules.getAtomContainerCount(); i++) {
			//if(i < 33) continue;
			IAtomContainer oneMole = molecules.getAtomContainer(i).clone();
			Instances instances = this.ci.getInstances(oneMole);
			Instances matchedInstances = this.ci.matchAttributes(instances, attributes);
			for(int j = 0; j < instances.numInstances(); j++) {
				String atom_idx = instances.attribute("Bond").value(j);				
				//double[] prob_pred = this.ci.ugt_classifier.distributionForInstance(matchedInstances.get(j));
				double prob = cls.distributionForInstance(matchedInstances.get(j))[1];
				//System.out.println(atom_idx + ": " + prob);
				if(prob >= threshold) {
					double score = (prob - threshold) / (1.0 - threshold);
					//System.out.println("Molecule idx: " + i + " ; reactive site: " + atom_idx);
					//String[] check = atom_idx.split("\\.");
					IAtomContainer oneResult = this.gms.generateMetabolite(Integer.parseInt(atom_idx.split("\\.")[1]), oneMole.clone(), type);
					oneResult.setProperty("Score", Double.valueOf(this.df.format(score)));
					oneResult.setProperty("Enzyme", enzyme);
					oneResult.setProperty("ReactionType", reactionType);
					InChIGenerator inChIGen = inChIFac.getInChIGenerator(oneResult);
					String inChIKey = inChIGen.getInchiKey();
					if(!productInChIKey.contains(inChIKey)) {
						resultSet.addAtomContainer(oneResult);
						productInChIKey.add(inChIKey);
					}
					
				}
			}
		}
		return resultSet;
	}
	
	


}
