package createSomDataset;

import java.io.FileOutputStream;

import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.io.SDFWriter;
import org.openscience.cdk.smiles.SmiFlavor;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;

import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

/**
 * https://academic.oup.com/bioinformatics/article/32/20/3183/2196454
 * @author Tian
 *
 */
public class CollectDataSimpleModelPaper {
	CollectData_convertSMIRKS_to_SDF cd = new CollectData_convertSMIRKS_to_SDF();
	SmilesParser sp = new SmilesParser(DefaultChemObjectBuilder.getInstance());
	SmilesGenerator sg = new SmilesGenerator(SmiFlavor.Isomeric);
	public static void main(String[] args) throws Exception{
		String inputPath = "G:/PhaseIIPredictor/suppl_info/Peng_external_dataset/Peng_external_dataset.sdf";
		String outputPath =  "G:/PhaseIIPredictor/Peng_external_dataset.sdf";
		CollectDataSimpleModelPaper cdsm = new CollectDataSimpleModelPaper();
		IAtomContainerSet molecules = cdsm.getData(inputPath);
		cdsm.generateDataset(molecules, outputPath);
	}
	
	public void generateDataset(IAtomContainerSet molecules, String outputPath) throws Exception{
		IAtomContainerSet result = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		for(int i = 0; i < molecules.getAtomContainerCount(); i++) {
			IAtomContainer oneMole = molecules.getAtomContainer(i);
			
			String target_atom_property = "None";
			int count = 0;
			for(int j = 0; j < oneMole.getAtomCount(); j++) {
				if(oneMole.getAtom(j).getProperty("IsBoM") != null) {
					count++;
					if(target_atom_property.equals("None")) target_atom_property = "<" + (j + 1) + "," + oneMole.getAtom(j).getSymbol() + ";UGT;R" + count + ">";
					else {
						target_atom_property =  target_atom_property + "\n" + "<" + (j + 1) + "," + oneMole.getAtom(j).getSymbol() + ";UGT;R" + count + ">";
					}
					
				}
			}
			oneMole.setProperty("BoM", target_atom_property);
			result.addAtomContainer(oneMole);
		}
		SDFWriter sdfw = new SDFWriter(new FileOutputStream(outputPath));
		sdfw.write(result);
		sdfw.close();
	}
	
	public IAtomContainerSet getData(String inputPath) throws Exception{
		IAtomContainerSet molecules = this.cd.readSDFFile(inputPath);
		for(int i = 0; i < molecules.getAtomContainerCount(); i++) {
			IAtomContainer oneMole = molecules.getAtomContainer(i);
			String primary_som = oneMole.getProperty("PRIMARY_SOM");
			String[] primary_som_array = primary_som.split("\\s");
			for(int j = 0; j < primary_som_array.length; j++) {
				Integer som_idx = Integer.parseInt(primary_som_array[j]);
				oneMole.getAtom(som_idx - 1).setProperty("IsBoM", true);
				
			}
			AtomContainerManipulator.suppressHydrogens(oneMole);
		}
		return molecules;
		
	}
	
	
}
