package createSomDataset;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.openscience.cdk.Atom;
import org.openscience.cdk.Bond;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.inchi.InChIGenerator;
import org.openscience.cdk.inchi.InChIGeneratorFactory;
import org.openscience.cdk.interfaces.IAtom;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.interfaces.IBond;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.io.SDFWriter;
import org.openscience.cdk.io.iterator.IteratingSDFReader;
import org.openscience.cdk.io.listener.PropertiesListener;
import org.openscience.cdk.isomorphism.Mappings;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smarts.SmartsPattern;
import org.openscience.cdk.smiles.SmiFlavor;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import ambit2.smarts.SMIRKSManager;
import ambit2.smarts.SMIRKSReaction;
import sulfation.CreateDataset;

public class CollectData_convertSMIRKS_to_SDF {
	SMIRKSManager smg = new SMIRKSManager(SilentChemObjectBuilder.getInstance());
	public SmilesGenerator sg = new SmilesGenerator(SmiFlavor.Isomeric);
	public SmilesParser sp = new SmilesParser(DefaultChemObjectBuilder.getInstance());
	CreateDataset createDataset_sulfation = new CreateDataset();
	public String UGT_SMARTS = "[H][#8]-[#6](=O)C1([H])[#8]-[#6]([H])C([H])([#8][H])C([H])([#8][H])C1([H])[#8][H]";
	public static void main(String[] args) throws Exception{
		String inputPath = "G:/PhaseIIPredictor/UGT/Training_SMIRKS.txt";
		String outputPath = "G:/PhaseIIPredictor/UGT/Training.sdf";
		CollectData_convertSMIRKS_to_SDF cd = new CollectData_convertSMIRKS_to_SDF();
		//IAtomContainerSet molecules = cd.readSMRIKSFile_getSubstrates(inputPath);
		//cd.generateData(molecules, outputPath);
		cd.generateDataForValidation();
		cd.generateDataForTesting();
		//cd.generateDataWithNS();
		
	}
	
	public void generateDataWithNS() throws Exception{
		IAtomContainerSet non_substrates = getNSMolecules();
		int numInstaces = 60;
		ArrayList<Integer> valdiation_60 = this.createDataset_sulfation.getRandomList(numInstaces, non_substrates.getAtomContainerCount(), new ArrayList<Integer>());
		IAtomContainerSet validationSet = this.createDataset_sulfation.getMoleculesByIndices(non_substrates, valdiation_60);
		ArrayList<Integer> test_60 = this.createDataset_sulfation.getRandomList(numInstaces, non_substrates.getAtomContainerCount(), valdiation_60);
		IAtomContainerSet testSet = this.createDataset_sulfation.getMoleculesByIndices(non_substrates, test_60);
		ArrayList<Integer> training_num = new ArrayList<>();
		for(int i = 0; i < non_substrates.getAtomContainerCount(); i++) {
			if(!valdiation_60.contains(i) && !test_60.contains(i) && !training_num.contains(i)) training_num.add(i);
		}
		IAtomContainerSet trainingSet = this.createDataset_sulfation.getMoleculesByIndices(non_substrates, training_num);
		String inputPath_train = "G:/PhaseIIPredictor/UGT/Training.sdf";
		String inputPath_validation = "G:/PhaseIIPredictor/UGT/Validation.sdf";
		String inputPath_test = "G:/PhaseIIPredictor/UGT/Testing.sdf";
		String outputPath_validation_NS = "G:/PhaseIIPredictor/UGT/withNS/UGT_validation.sdf";
		String outputPath_test_NS = "G:/PhaseIIPredictor/UGT/withNS/UGT_test.sdf";
		String outputPath_training_NS = "G:/PhaseIIPredictor/UGT/withNS/UGT_training.sdf";
		IAtomContainerSet train = readSDFFile(inputPath_train);		
		IAtomContainerSet validation = readSDFFile(inputPath_validation);
		IAtomContainerSet test = readSDFFile(inputPath_test);
		train.add(trainingSet);
		validation.add(validationSet);
		test.add(testSet);
		writeSDF(train, outputPath_training_NS);
		writeSDF(validation, outputPath_validation_NS);
		writeSDF(test,outputPath_test_NS);
		
	}
	
	/**
	 * This function will prepare the non-substrates for UGT conjugation reaction and store them into an IAtomContainerSet
	 * @return
	 * @throws Exception
	 */
	public IAtomContainerSet getNSMolecules() throws Exception{
		String inputPath_NS = "G:/PhaseIIPredictor/UGT/withNS/non-substrates.csv";
		BufferedReader br = new BufferedReader(new FileReader(new File(inputPath_NS)));
		IAtomContainerSet non_substrates = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		String oneLine = br.readLine(); //title line
		while((oneLine = br.readLine()) != null) {
			String smiles = oneLine.split(",")[1];
			try {
				IAtomContainer oneMole = this.sp.parseSmiles(smiles);
				non_substrates.addAtomContainer(oneMole);	
			}catch (Exception e) {
				System.out.println(oneLine);
			}
		}
		br.close();
		return non_substrates;
	}
	
	public void generateDataForTesting() throws Exception{
		String inputPath = "G:/PhaseIIPredictor/UGT/Test_SMIRKS.txt";
		String outputPath = "G:/PhaseIIPredictor/UGT/Testing.sdf";
		String outputPath_metabolite = "G:/PhaseIIPredictor/UGT/Testing_metabolite.sdf";
		IAtomContainerSet molecules = readSMRIKSFile_getSubstrates(inputPath);
		IAtomContainerSet products =readSMIRKSFile_getProducts(inputPath);
		generateData(molecules, outputPath);
		generateData(products, outputPath_metabolite);
	}
	
	
	public void generateDataForValidation() throws Exception{
		String inputPath = "G:/PhaseIIPredictor/UGT/Validation_SMIRKS.txt";
		String outputPath = "G:/PhaseIIPredictor/UGT/Validation.sdf";
		String outputPath_metabolite = "G:/PhaseIIPredictor/UGT/Validation_metabolite.sdf";
		IAtomContainerSet molecules = readSMRIKSFile_getSubstrates(inputPath);
		IAtomContainerSet products =readSMIRKSFile_getProducts(inputPath);
		generateData(molecules, outputPath);
		generateData(products, outputPath_metabolite);
	}
	
	public IAtomContainerSet readSMRIKSFile_getSubstrates(String inputPath) throws Exception{
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);				
		BufferedReader br = new BufferedReader(new FileReader(new File(inputPath)));
		String oneLine;
		int counter = 0;
		while((oneLine = br.readLine()) != null) {
			counter++;
			String smirks = oneLine;
			SMIRKSReaction reaction = this.smg.parse(smirks);
			IAtomContainer product = this.sp.parseSmiles(reaction.productsSmarts);
			product = this.sp.parseSmiles(this.sg.create(product));
			IAtomContainer oneProcessed = generateMoleculeForDB(product);	
			if(oneProcessed == null || oneProcessed.isEmpty()) continue;
			resultSet.addAtomContainer(oneProcessed);			
		}
		br.close();
		return resultSet;
	}
	
	public IAtomContainerSet readSMIRKSFile_getProducts(String inputPath) throws Exception{
		String type = "UGT";
		InChIGeneratorFactory inchiFac = InChIGeneratorFactory.getInstance();
		ArrayList<String> existedMetaboliteInChiKey = new ArrayList<>();
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);				
		BufferedReader br = new BufferedReader(new FileReader(new File(inputPath)));
		String oneLine;
		int counter = 0;
		while((oneLine = br.readLine()) != null) {
			counter++;
			//System.out.println("Processing SMIRKS.txt file line: " + counter);
			String smirks = oneLine;
			SMIRKSReaction reaction = this.smg.parse(smirks);
			IAtomContainer product = this.sp.parseSmiles(reaction.productsSmarts);
			product = this.sp.parseSmiles(this.sg.create(product));
			IAtomContainer oneProcessed = generateMoleculeForDB(product);	
			if(oneProcessed == null || oneProcessed.isEmpty()) continue;
			resultSet.addAtomContainer(product);
						
		}
		br.close();
				
		for(int i = 0; i < resultSet.getAtomContainerCount(); i++) {
			IAtomContainer oneResult = resultSet.getAtomContainer(i);
			IAtomContainerSet cleanedResults = generateMetabolite_Special(oneResult, type);
			if(cleanedResults!=null && !cleanedResults.isEmpty()) resultSet.add(cleanedResults);
			
		}
		IAtomContainerSet metbaolites_noDuplicate = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		for(int i = 0; i < resultSet.getAtomContainerCount(); i++) {
			IAtomContainer product = resultSet.getAtomContainer(i);
			InChIGenerator inchiGen = inchiFac.getInChIGenerator(product);
			String inChiKey = inchiGen.getInchiKey().split("-")[0];
			if(!existedMetaboliteInChiKey.contains(inChiKey)) {
				existedMetaboliteInChiKey.add(inChiKey);
				metbaolites_noDuplicate.addAtomContainer(product);
			}
			
		}
		
		return metbaolites_noDuplicate;
	}
	

	
	public IAtomContainer generateMoleculeForDB(IAtomContainer product) throws Exception{
		IAtomContainer oneMole = product.clone();
		IAtomContainer result = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainer.class);
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(oneMole);
		AtomContainerManipulator.convertImplicitToExplicitHydrogens(oneMole);
		SmartsPattern smp = SmartsPattern.create(this.UGT_SMARTS);
		smp.setPrepare(false);
		Mappings mapped = smp.matchAll(oneMole);
		//System.out.println(this.sg.create(oneMole));
		if(mapped.uniqueAtoms().countUnique() == 0) {
			System.out.println("Invalid");
			return result;
		}
		ArrayList<Integer> matchedAtomsList = new ArrayList<>();
		for(int[] oneMap : mapped.uniqueAtoms()) {
			for(int i = 0; i < oneMap.length; i++) {
				Integer idx = oneMap[i];
				if(!matchedAtomsList.contains(idx)) matchedAtomsList.add(idx);
			}
		}
		for(int i = 0; i < oneMole.getAtomCount(); i++) {
			Integer idx = oneMole.indexOf(oneMole.getAtom(i));
			if(!matchedAtomsList.contains(idx)) {
				if(!result.contains(oneMole.getAtom(i))) result.addAtom(oneMole.getAtom(i));
			}
		}
		for(int i = 0; i < result.getAtomCount(); i++) {
			IAtom workOnAtom = result.getAtom(i);						
			for(int j = i; j < result.getAtomCount(); j++) {
				IBond oneBond = oneMole.getBond(workOnAtom, result.getAtom(j));
				if(oneBond != null && !result.contains(oneBond)) {
					result.addBond(oneBond);
				}
			}
		}
		//boolean processed = false;
		for(int i = 0; i < result.getAtomCount(); i++) {
			//if(processed) break;
			IAtom workOnAtom = result.getAtom(i);	
			if(workOnAtom.getProperty("JointAtom") != null || workOnAtom.getProperty("FakeAtom") != null) continue;
			List<IAtom> neighborList = oneMole.getConnectedAtomsList(workOnAtom);
			for(int j = 0; j < neighborList.size(); j++) {
				if(matchedAtomsList.contains(oneMole.indexOf(neighborList.get(j)))) {					
					workOnAtom.setProperty("JointAtom", true);					
					if(workOnAtom.getFormalCharge() == 0) {
						IAtom fakeAtom = new Atom("H");
						fakeAtom.setImplicitHydrogenCount(0);
						IBond fakeBond = new Bond(workOnAtom, fakeAtom, IBond.Order.SINGLE);
						fakeAtom.setProperty("FakeAtom", true);
						result.addAtom(fakeAtom);
						result.addBond(fakeBond);
						//processed = true;
						//break;
					}
					else {
						//System.out.println(oneMole.indexOf(neighborList.get(j)));
						workOnAtom.setFormalCharge(workOnAtom.getFormalCharge() - 1);
					}
				}
			}
			//int check = workOnAtom.getImplicitHydrogenCount();
			//check = 0;
		}

		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(result);
		AtomContainerManipulator.suppressHydrogens(result);

		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(result);
		String target_atom_property =  "None";
		int count = 0;
		for(int i = 0; i < result.getAtomCount(); i++) {
			if(result.getAtom(i).getProperty("JointAtom") != null) {
				count++;
				if(target_atom_property.equals("None")) target_atom_property = "<" + (i + 1) + "," + result.getAtom(i).getSymbol() + ";UGT;R" + count + ">";
				else target_atom_property = target_atom_property + "\n" + "<" + (i + 1) + "," + result.getAtom(i).getSymbol() + ";UGT;R" + count + ">";
				
				
			}
		}	
		result.setProperty("BoM", target_atom_property);
		//System.out.println(this.sg.create(result));
		return result;
	}
	
	public void generateData(IAtomContainerSet molecules, String outputPath) throws Exception{
		SDFWriter sdfw = new SDFWriter(new FileOutputStream(outputPath));
		sdfw.write(molecules);
		sdfw.close();
		IAtomContainerSet checkList = readSDFFile(outputPath);
		for(int i = 0; i < checkList.getAtomContainerCount(); i++) {
			IAtomContainer check = checkList.getAtomContainer(i);
			String bom = check.getProperty("BoM");
			if(bom == null) continue;
			//System.out.println(bom);
			String location = bom.split(";")[0].split(",")[0].replace("<", "");
			String symbol = bom.split(";")[0].split(",")[1].replace(">", "");
			Integer idx = Integer.parseInt(location);
			IAtom checkAtom = check.getAtom((idx - 1));
			if(checkAtom.getSymbol().equalsIgnoreCase("C") || checkAtom.getSymbol().equalsIgnoreCase("H") || !checkAtom.getSymbol().equalsIgnoreCase(symbol)) {
				System.out.println((i+1));
			}
			
		}
	}
		
	

	public IAtomContainerSet readSDFFile(String pathToInputFile) throws Exception {
		IChemObjectBuilder bldr = SilentChemObjectBuilder.getInstance();
		IteratingSDFReader sdfr = new IteratingSDFReader(new FileReader(pathToInputFile),
				bldr);
		Properties prop = new Properties();
		//prop.setProperty("ForceReadAs3DCoordinates", "true");
		PropertiesListener listener = new PropertiesListener(prop);
		sdfr.addChemObjectIOListener(listener);
		sdfr.customizeJob();
		IAtomContainerSet MOLS = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		while (sdfr.hasNext())MOLS.addAtomContainer(sdfr.next());
		sdfr.close();
		return MOLS;

	}
	
	public void writeSDF(IAtomContainerSet molecules, String outputPath) throws Exception{
		SDFWriter sdfw = new SDFWriter(new FileOutputStream(outputPath));
		sdfw.write(molecules);
		sdfw.close();
	}
	/**
	 * This function is used to handle some non-substrates in the NS dataset that contain nitro group but using a wrong representation 
	 * @param molecule
	 * @return
	 * @throws Exception
	 */
	public IAtomContainer processNitrogen(IAtomContainer molecule) throws Exception{
		String smirks_nitro_smirks = "[H][#8:4][N:1]([H])([*:3])=[O:2]>>[#8-:4]-[#7+:1](-[*:3])=[O:2]";
		String smirks_nitrogen_wrong_smirks = "[H][N:1]([#6:3])([#6:2])([#6:5])[#6:4]>>[#6:3][N+:1]([#6:2])([#6:5])[#6:4]";
		SMIRKSManager smk_manager = new SMIRKSManager(SilentChemObjectBuilder.getInstance());
		SMIRKSReaction reaction = smk_manager.parse(smirks_nitro_smirks);
		IAtomContainerSet converted = smk_manager.applyTransformationWithSingleCopyForEachPos(molecule, null, reaction);
		if(converted == null || converted.isEmpty()) {
			reaction = smk_manager.parse(smirks_nitrogen_wrong_smirks);
			converted = smk_manager.applyTransformationWithSingleCopyForEachPos(molecule, null, reaction);
			if(converted == null || converted.isEmpty()) {
				System.out.println("Can't handle this molecule");
				return null;
			}
		}
		
		return this.sp.parseSmiles(this.sg.create(converted.getAtomContainer(0)));
	}

	/**
	 * This function will generate multiple metabolites for a product that contains more than one target smirks string
	 * @param molecule
	 * @return
	 * @throws Exception
	 */
	public IAtomContainerSet generateMetabolite_Special(IAtomContainer molecule, String type) throws Exception{
		InChIGeneratorFactory inchiFac = InChIGeneratorFactory.getInstance();		
		SmilesGenerator sg = new SmilesGenerator(SmiFlavor.Isomeric);
		String smirks;
		String feature_smarts;
		String fix_smirks = null;
		if(type.equals("UGT")) {
			smirks = "[#8]-[#6]-1-[#6](-[#8])-[#6](-[*:1])-[#8]-[#6](-[#6]-1-[#8])-[#6](-[#8])=O>>[*:1][H]";
			fix_smirks = "[#7+:1][H]>>[#7:1]";
			feature_smarts = UGT_SMARTS;
		}
		else if(type.equals("SULT")) {
			smirks = "TBD";
			feature_smarts = "[H][#8]S([#8])(=O)=O";
		}
		else throw new Exception("Please specify the phaseII reaction type");
		SMIRKSManager smk_manager = new SMIRKSManager(SilentChemObjectBuilder.getInstance());
		SMIRKSReaction reaction = smk_manager.parse(smirks);
		IAtomContainerSet results = smk_manager.applyTransformationWithSingleCopyForEachPos(molecule, null, reaction);
		//IAtomContainerSet results_metabolites = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		ArrayList<String> existedMetabolitesInChIKey = new ArrayList<>();
		if(results == null) {
			System.out.println("molecule with no product: " + sg.create(molecule));
			return results;
		}
		if(results.getAtomContainerCount() > 1) {
			System.out.println("molecule has more than one target functional group: " + sg.create(molecule));
		}
		for(int i = 0; i < results.getAtomContainerCount(); i++) {
			IAtomContainer oneResult = results.getAtomContainer(i);
			if(fix_smirks != null) {
				SMIRKSManager smk_manager_fix = new SMIRKSManager(SilentChemObjectBuilder.getInstance());
				SMIRKSReaction reaction_fix = smk_manager.parse(fix_smirks);
				IAtomContainerSet results_fix = smk_manager_fix.applyTransformationWithSingleCopyForEachPos(oneResult, null, reaction_fix);
				if(results_fix!=null) {
					oneResult = results_fix.getAtomContainer(0);
				}
			}
			if(!matchSMARTS(oneResult, feature_smarts)) {
				results.removeAtomContainer(i);
				i--;
			}
			InChIGenerator inChIGen = inchiFac.getInChIGenerator(oneResult);
			String inChIKey = inChIGen.getInchiKey();
			if(!existedMetabolitesInChIKey.contains(inChIKey)) {
				existedMetabolitesInChIKey.add(inChIKey);
				results.addAtomContainer(oneResult);
				//System.out.println(sg.create(oneResult));
			}			
		}
		return results;
	}
	
	/**
	 * This function checks if the molecule contains the substructure described by the smarts
	 * @param product
	 * @param smarts
	 * @return
	 * @throws Exception
	 */
	public boolean matchSMARTS(IAtomContainer product, String smarts) throws Exception{
		IAtomContainer oneMole = product.clone();
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(oneMole);
		AtomContainerManipulator.convertImplicitToExplicitHydrogens(oneMole);
		SmartsPattern  smp = SmartsPattern.create(smarts);
		smp.setPrepare(false);
		if(smp.matches(oneMole)) return true;
		else {
			smp.setPrepare(true);
			return smp.matches(oneMole);
		}
	}
}
