package createSomDataset;

import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;

import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.silent.SilentChemObjectBuilder;

import ambit2.smarts.SMIRKSManager;
import ambit2.smarts.SMIRKSReaction;
import generateInstance.GenerateRawFeatures;
import generateInstance.RemoveUselessFeatures;
import phase2utils.GenerateTitleline;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.converters.ArffSaver;

public class GenerateTrainingTsv {
	int depth = 4;
	CollectData_convertSMIRKS_to_SDF cd = new CollectData_convertSMIRKS_to_SDF();
	GenerateRawFeatures generateRawFeatures = new GenerateRawFeatures(depth);
	public static void main(String[] args) throws Exception{
		//String inputPath = "G:/PhaseIIPredictor/UGT/Training.sdf";		
		//String inputPath = "G:/PhaseIIPredictor/Sulfation/Sulfation_training.sdf";
		GenerateTrainingTsv gnbl = new GenerateTrainingTsv();
		//gnbl.generate("UGT");
		gnbl.generate("SULT");
	}
	
	public void generate(String type) throws Exception{
		String inputPath;
		String outputPath_allFeatures;
		if(type.equals("UGT")) {
			inputPath = "G:/PhaseIIPredictor/UGT/withNS/UGT_training.sdf";	
			outputPath_allFeatures = "G:/PhaseIIPredictor/UGT/withNS/TrainingData/trainingFeatures.csv";
		}
		else if(type.equals("SULT")) {
			inputPath = "G:/PhaseIIPredictor/Sulfation/withNS/Sulfation_training.sdf";
			outputPath_allFeatures = "G:/PhaseIIPredictor/Sulfation/withNS/trainingFeatures.csv";
		}
		else throw new Exception("Please select the correct phase II reaction type: UGT or SULT");
		//String outputPath_allFeatures = "G:/PhaseIIPredictor/UGT/TrainingData/trainingFeatures.csv";
		//String outputPath_allFeatures = "G:/PhaseIIPredictor/Sulfation/trainingFeatures.csv";
		FileWriter fw = new FileWriter(new File(outputPath_allFeatures));
		String titleLine = GenerateTitleline.generateTitle(this.depth);	
		fw.write(titleLine + "\n");
		ArrayList<String> rawFeaturesAll = new ArrayList<>();
		IAtomContainerSet molecules = 	this.cd.readSDFFile(inputPath);
		for(int i = 0; i < molecules.getAtomContainerCount(); i++) {
			//if(i < 960) continue;
			//if(i > 3) break;
			IAtomContainer oneMole = molecules.getAtomContainer(i);
			ArrayList<String> rawFeatureList = new ArrayList<>();
			try {
				rawFeatureList = this.generateRawFeatures.generateRawFeatures(oneMole);	
			}catch (Exception e) {
				//e.printStackTrace();				
				System.out.println("Failed processing molecule " + (i-550 + 1) + ": " + cd.sg.create(oneMole));
				IAtomContainer converted = this.cd.processNitrogen(oneMole);
				if(converted == null || converted.isEmpty()) System.out.println("Failed to convert the nitrogen contained functional group");
				else{
					System.out.println("Converted: " + cd.sg.create(converted));
					try {
						rawFeatureList = this.generateRawFeatures.generateRawFeatures(converted);		
					}catch(Exception e2) {
						System.out.println("Failed to generate raw features for the converted structure");
						continue;
					}
				}
			}
			for(int j = 0; j < rawFeatureList.size(); j++) {
				if(!rawFeaturesAll.contains(rawFeatureList.get(j))) {
					rawFeaturesAll.add(rawFeatureList.get(j));
					fw.write(rawFeatureList.get(j) + "\n");
				}
				
			}
			
			
			System.out.println("processed molecule : " + (i+1) );
		}
		fw.close();
		generateNameBondListArff(rawFeaturesAll);
	}
	
	
	public void generateNameBondListArff(ArrayList<String> rawFeatures) throws Exception{
		String outputPath_name = "G:/PhaseIIPredictor/UGT/withNS/TrainingData/nameList.arff";
		String outputPath_bond = "G:/PhaseIIPredictor/UGT/withNS/TrainingData/bondList.arff";
		ArrayList<String> nameList = new ArrayList<>();
		ArrayList<String> bondList = new ArrayList<>();
		for(int i = 0; i < rawFeatures.size(); i++) {
			String name = rawFeatures.get(i).split(",")[0];
			String bond = rawFeatures.get(i).split(",")[1];
			if(!nameList.contains(name)) nameList.add(name);
			if(!bondList.contains(bond)) bondList.add(bond);		
		}
		Instances nameInstances = createNameInstances(nameList);
		Instances bondInstances = createBondInstances(bondList);
		ArffSaver arff = new ArffSaver();
		arff.setFile(new File(outputPath_name));
		arff.setInstances(nameInstances);
		arff.writeBatch();
		
		arff = new ArffSaver();
		arff.setFile(new File(outputPath_bond));
		arff.setInstances(bondInstances);
		arff.writeBatch();
	}
	
	public Instances createNameInstances(ArrayList<String> nameList) throws Exception{
		ArrayList<Attribute> attList = new ArrayList<>();
		Attribute nameAttribute = new Attribute("Name",nameList);	
		attList.add(nameAttribute);
		Instances nameInstances = new Instances("Rel", attList, 100000); 		
		for(int i = 0; i < nameList.size(); i++) {
			Instance oneInstance = new DenseInstance(attList.size()); 
			oneInstance.setValue(nameAttribute, nameList.get(i));
			nameInstances.add(oneInstance);
		}
		nameInstances.setClassIndex(0);
		//RemoveUselessFeatures.selectByInfoGain(nameInstances, 5);
		return nameInstances;
	}
	
	public Instances createBondInstances(ArrayList<String> bondList) {
		ArrayList<Attribute> attList = new ArrayList<>();
		Attribute bondAttribute = new Attribute("Bond",bondList);	
		attList.add(bondAttribute);
		Instances bondInstances = new Instances("Rel", attList, 100000); 		
		for(int i = 0; i < bondList.size(); i++) {
			Instance oneInstance = new DenseInstance(attList.size()); 
			oneInstance.setValue(bondAttribute, bondList.get(i));
			bondInstances.add(oneInstance);
		}
		bondInstances.setClassIndex(0);
		return bondInstances;
	}

}
