package createSomDataset;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;

import generateInstance.GenerateInstances;
import generateInstance.GenerateRawFeatures;
import phase2utils.Utilities;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.converters.ArffSaver;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.Remove;
import weka.filters.unsupervised.attribute.Reorder;

public class PrepareCV {
	CollectData_convertSMIRKS_to_SDF cd = new CollectData_convertSMIRKS_to_SDF();
	int depth = 4;
	//GenerateRawFeatures generateFeatures = new GenerateRawFeatures(depth);
	GenerateInstances generateInstances = new GenerateInstances(depth);
	
	
	
	public Instances getInstancesByName(Instances nameInstances, Instances allInstances) throws Exception{
		ArrayList<String> nameList = Utilities.convertNameList(nameInstances);
		ArrayList<Attribute> attList = new ArrayList<>();
		Integer nameIdx = Utilities.findIndexOfName(allInstances);
		for(int i = 0; i < allInstances.numAttributes(); i++){
			Attribute oneAtt = allInstances.attribute(i);
			attList.add(oneAtt);
		}
		Instances subInstances = new Instances("Rel", attList, 100000);
		for(int i = 0; i < allInstances.numInstances(); i++){
			Instance oneInstance = allInstances.get(i);
			String name = oneInstance.stringValue(nameIdx);//The name attribute
			if(nameList.contains(name)){
				subInstances.add(oneInstance);
			}
		}
		subInstances.setClass(allInstances.classAttribute());
		return subInstances;
	}
	
	public Instances readingInstancesFromCSV(String inputPath) throws Exception{		
		BufferedReader br = new BufferedReader(new FileReader(new File(inputPath)));
		String oneLine = br.readLine();//title line
		ArrayList<Attribute> attList = this.generateInstances.createAttributes(oneLine);	
		Instances allInstances = new Instances("Rel", attList, 100000); 
		while((oneLine = br.readLine()) != null) {
			Instance oneInstance_raw = new DenseInstance(attList.size()); 
			Instance oneInstance = this.generateInstances.createOneInstance(oneLine, oneInstance_raw, attList);
			allInstances.add(oneInstance);
		}
		br.close();
		return allInstances;
	
	}
	/**
	 * This fucntion will return the dataset after the name and bonds are removed
	 * @param dataset
	 * @return
	 * @throws Exception
	 */
	public Instances getDatasetWithoutNameBond(Instances dataset) throws Exception{
		String indices_toRemove = Utilities.findIndcesOfNameAndBond(dataset);
		return removeAttributes(dataset, indices_toRemove);
	}
	
	/**
	 * This function is designed to remove attributes with the input indices.
	 * The purpose is to remove the first and second attribute, which are name of molecule and bond.
	 * Note: We assume Filter.useFilter(data) will return a new filtered data_new without modifying the "data" file. Need to double check when debug.
	 * @param dataset
	 * @param indices
	 * @return
	 * @throws Exception
	 */
	public static Instances removeAttributes(Instances dataset, String indices) throws Exception{
        if(dataset.classIndex()==-1){
        	dataset.setClassIndex(dataset.numAttributes()-1);
        }
		Remove filter = new Remove(); 
        filter.setAttributeIndices(indices); 
        filter.setInvertSelection(false); 
        filter.setInputFormat(dataset); 
        Instances filteredData = Filter.useFilter(dataset, filter);
        if(filteredData.classIndex()==-1){
        	filteredData.setClassIndex(filteredData.numAttributes()-1);
        }
        return filteredData;
	}
	
	public void writeArffFile(Instances dataSet, String outPath) throws Exception{
		ArffSaver saver = new ArffSaver();
		saver.setInstances(dataSet);
		saver.setFile(new File(outPath));
		//saver.setDestination(new File("./data/test.arff"));   // **not** necessary in 3.5.4 and later
		saver.writeBatch();
	}
	
	/**
	 * This function read a arff file and returns the Instances stored in the file
	 * @param inputPath_name
	 * @return
	 * @throws Exception
	 */
	public Instances readArffFile(String inputPath_name) throws Exception{
		Instances names = new Instances(new FileReader(new File(inputPath_name)));
		if(names.classIndex()==-1){
			names.setClassIndex(names.numAttributes()-1);
		}
		return names;
	}
	
	public Instances readCsv(String inputPath_csv) throws Exception{
		BufferedReader br = new BufferedReader(new FileReader(new File(inputPath_csv)));
		ArrayList<String> rawList = new ArrayList<>();
		String oneLine = br.readLine();//title line
		String titleLine = oneLine;
		while((oneLine = br.readLine()) != null) {
			rawList.add(oneLine);
		}
		br.close();
		return this.generateInstances.createInstances(rawList, titleLine);
		
	}			
	
	public ArrayList<String> getAttributeList(Instances allIsntances) throws Exception{
		ArrayList<String> resultList = new ArrayList<>();
		for(int i = 0; i < allIsntances.numAttributes(); i++) {
			String att_name = allIsntances.attribute(i).name();
			resultList.add(att_name);
		}
		return resultList;
	}
	
	/**
	 * 1st. Existence of the attributes in the testDataSet; 
	 * 2nd. The orders of the attributes in the testDataSet;
	 * @param testInstances
	 * @param trainInstances
	 * @return
	 * @throws Exception
	 */
	public Instances getSameAttributes(Instances testInstances, Instances trainInstances) throws Exception{
		Instances allignedInstances = testInstances;
		ArrayList<Attribute> allAtt = new ArrayList<Attribute>();
		ArrayList<Integer> retainAttIdx = new ArrayList<Integer>();
		//This for loop stores all the attributes in the testInstances into ArrayList<Attribute> allAtt
		for(int j = 0; j < testInstances.numAttributes(); j++){
			Attribute oneAtt = testInstances.attribute(j);
			allAtt.add(oneAtt);
		}
		/* This for loop will:          Note that attributes in the trainInstances belong to a subset of attributes of testInstances
		 * 1. get the i-th attribute att of the trainInstances
		 * 2. find the index of attribute att in the testInstances
		 * 3. Store i-th as the index of attribute att in the ArrayList<Integer> retainAttIdx. This order will be used in the reorder filter 
		 */
		for(int i = 0; i < trainInstances.numAttributes(); i++){
			Attribute oneTrainAtt = trainInstances.attribute(i);
			int idx = allAtt.indexOf(oneTrainAtt);
			if(idx == -1){
				System.out.println(oneTrainAtt.toString() + "idx = " + idx);
			}
			retainAttIdx.add(idx);
		}
		int[] retainList = new int[retainAttIdx.size()];
		for(int k = 0; k < retainAttIdx.size(); k++){
			retainList[k] = retainAttIdx.get(k);
		}
		Reorder reorederFiler = new Reorder(); 
		reorederFiler.setAttributeIndicesArray(retainList);
		reorederFiler.setInputFormat(testInstances);
		allignedInstances = Filter.useFilter(allignedInstances, reorederFiler);

        if(allignedInstances.classIndex()==-1){
        	allignedInstances.setClassIndex(allignedInstances.numAttributes()-1);
        }
		return allignedInstances;
	}
	
}
