package generateInstance;

import java.util.ArrayList;
import java.util.List;

import phase2utils.GenerateTitleline;
import phase2utils.Utilities;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;

public class GenerateInstances {
	Integer depth = 4;
	public GenerateInstances(Integer depth) {
		this.depth = depth;
	}
	/**
	 * This function takes as input an ArrayList<String> for rawData and a String for titleLine
	 * @param rawData: the values for each eta-eta bond
	 * @param titleLine: the names of all attributes, including names and features, separated by comma 
	 * @return Instances: featured instances for each eta-eta bond within the the molecule. 
	 * @throws Exception
	 */
	public Instances createInstances(ArrayList<String> rawData, String titleLine) throws Exception{
		//Convert nameLsitInstances from "Instances" to "ArrayList" to make it more efficient in both computation and storage
		ArrayList<Attribute> attList = createAttributes(titleLine);		
		Instances bomInstances = new Instances("Rel", attList, 100000); 
		String oneLine;
		for(int i = 0; i < rawData.size(); i++){
			oneLine = rawData.get(i);
			Instance oneInstance = new DenseInstance(attList.size()); 
			oneInstance = createOneInstance(oneLine, oneInstance, attList);			
			bomInstances.add(oneInstance);			
		}
		
		if(bomInstances.classIndex()==-1){
			bomInstances.setClassIndex(2);
		}
		/*
		 * Match attributes
		 */
		//Instances matchedInstance = matchAttributes(bomInstances,this.attributes);
		//System.out.println("Instances of dataset have been created successfully");
		//return matchedInstance;
		return bomInstances;
	}
	

	/**
	 * Reformat the instances of the input molecules so they can be input to the learned models by applying feature selection and normalization
	 * @param testSet: The instances of the user's input molecules
	 * @param attList: The attributes that are used in the learned model
	 * @param meanList: The mean values of each attribute
	 * @param maxList: The max values of each attribute
	 * @param minList: The min values of each attribute
	 * @return
	 * @throws Exception
	 */
	public Instances matchAttributes(Instances testSet, ArrayList<String> attListToMatch) throws Exception{
		/*
		 * Create the class atrribute
		 */
		List my_nominal_values = new ArrayList(3); 
		my_nominal_values.add("F"); 
		my_nominal_values.add("T"); 
		Attribute lastAttribute  = new Attribute(attListToMatch.get(attListToMatch.size()-1), my_nominal_values);
		/*
		 * Convert the attribute name strings to attribute
		 */
		ArrayList<Attribute> attsToMatch = new ArrayList<Attribute>();
		for(int i = 0; i<attListToMatch.size();i++){
			if(i<attListToMatch.size()-1){
				Attribute attribute = new Attribute(attListToMatch.get(i));
				attsToMatch.add(attribute);
			}
			else{
				
				attsToMatch.add(lastAttribute);
			}
		}

		int numAttributes = attListToMatch.size();
		Instances matched = new Instances("Rel", attsToMatch, testSet.size());
		Instance sample = new DenseInstance(numAttributes);
		/*
		 * Assign values for each testing instance
		 */
		for(int idxInstance = 0; idxInstance< testSet.size(); idxInstance++){
			//ArrayList misFeatures = new ArrayList();
			Instance temp = testSet.get(idxInstance);
			String label;
			//System.out.println("Check Predictor temp.classValue() == " + temp.classValue());
			if(temp.classValue()==0.0){
				label = "F";
			}
			else{
				label = "T";
			}
			
			for(int idxAttribute = 0; idxAttribute < numAttributes; idxAttribute++){
				  Attribute oneAttribute = attsToMatch.get(idxAttribute);
				  if(idxAttribute<numAttributes-1){					  
					  String nameOfAtt = oneAttribute.toString();
					  String[] formatAtt = nameOfAtt.split(" ");
					  
					  if(formatAtt.length!=3){
						  String combineSpaces = "";
						  for(int kk = 1; kk<formatAtt.length-1; kk++){
							  if(kk<formatAtt.length - 2){
								  combineSpaces = combineSpaces + formatAtt[kk]+ " ";
							  }
							  else combineSpaces = combineSpaces + formatAtt[kk];
						  }
						  formatAtt[1] = combineSpaces;
					  }
					 
					  if(formatAtt[1].contains("'")){
						  formatAtt[1] = formatAtt[1].replace("'", "");
					  }
					  //System.out.println(formatAtt[1]);
					  Attribute convAtt = testSet.attribute(formatAtt[1]);
					  double vle = temp.value(convAtt);
					  if(!Double.isNaN(vle)){					
						  sample.setValue(oneAttribute, vle);
					  }

				  }
				  else{
					  //Set class with random value. Does not matter
					  sample.setValue(lastAttribute, label);
					  
				  }
			}						
			matched.add(sample);
		}
		if(matched.classIndex() == -1){
			matched.setClassIndex(matched.numAttributes()-1);
		}
		
		return matched;
		
	}
	
	/**
	 * This function generates the attribute list for the dataset. The first attribute is the name of the molecule. The last attribute is the class attribute.
	 * @param titleLine: The string contains all the attribute names
	 * @param nameList: The names of molecules. Each name string is a nominal string in the name attribute.
	 * Note that the number of attributes is 486 = 26 + 17*26 + 18
	 * @return
	 */
	public ArrayList<Attribute> createAttributes(String titleLine){
		ArrayList<Attribute> attList = new ArrayList<>();
		String[] parseLine = titleLine.split(",");
		ArrayList<String> numHydrogenList = getNumHydrogenList();
		for(int i = 0; i<11; i++){
			if(i == 2) {
				List<String> class_nominal_values = new ArrayList<String>();
				class_nominal_values.add("F");// F = not BoM. 0
				class_nominal_values.add("T");// T = BoM. 1 
				Attribute classAttribute = new Attribute(parseLine[i],class_nominal_values);
				attList.add(classAttribute);
			}
			else if(i == 0 || i == 1){
				Attribute oneAtt = new Attribute(parseLine[i],  (ArrayList<String>) null);
				attList.add(oneAtt);
			}
			else {
				Attribute oneAtt = new Attribute(parseLine[i]);
				attList.add(oneAtt);
			}
			
		}
		//Feature attributes starts at index = 3.
		for(int i = 11; i < parseLine.length; i++){
			Attribute oneAtt = new Attribute(parseLine[i]);
			if(parseLine[i].equals("atomHydrogenCount")){
				oneAtt = new Attribute(parseLine[i],numHydrogenList);
				attList.add(oneAtt);
				continue;
			}
			attList.add(oneAtt);
		}
		return attList;
	}
	
	/**
	 * Create getNumHydrogenList nominal attribute.
	 * @return
	 */
	public ArrayList<String> getNumHydrogenList(){
		ArrayList<String> numHydrogenList = new ArrayList<>();
		numHydrogenList.add("C");//new for type 3
		numHydrogenList.add("S");//new for type 3
		numHydrogenList.add("N");//new for type 3 
		numHydrogenList.add("P");//new for type 3
		numHydrogenList.add("O");
		numHydrogenList.add("OH");
		numHydrogenList.add("CH");
		numHydrogenList.add("CH2");
		numHydrogenList.add("CH3");
		numHydrogenList.add("NH");
		numHydrogenList.add("NH2");
		numHydrogenList.add("NH3");
		numHydrogenList.add("PH");
		numHydrogenList.add("PH2");
		numHydrogenList.add("PH3");
		numHydrogenList.add("SH");
		numHydrogenList.add("SH2");
		numHydrogenList.add("SH3");
		numHydrogenList.add("Others");
		return numHydrogenList;
	}
	
	/**
	 * This function assign values in the dataset file to a Instance.
	 * @param oneLine
	 * @param oneInstance
	 * @return
	 */
	public Instance createOneInstance(String oneLine, Instance oneInstance, ArrayList<Attribute> attList){
		String[] parseLine = oneLine.split(",");
		String label = parseLine[2];//need to consider cyp index
		Attribute att_Class = attList.get(2);
		oneInstance.setValue(att_Class, label);//The last attribute is the class attribute. Note that the label is "T" or "F"
		for(int i = 0; i < 11; i++){
			if(i == 2) continue;//The second one is the value for the label attribute. It has been handled so we skip it
			Attribute oneAtt = attList.get(i);
			if(oneAtt.isNumeric()) {
				oneInstance.setValue(oneAtt, Double.parseDouble(parseLine[i]));
			}
			else oneInstance.setValue(oneAtt,parseLine[i]);
		}
		
		for(int i = 11; i < oneInstance.numAttributes()-1; i++){ //Set values for the other n-2 attributes. Assume n is the number of attributes 
			Attribute oneAtt = attList.get(i);//i = 10, i+1
			//System.out.println(oneAtt.toString() + "," + parseLine[i+2]);
			//Att[0] = par[0], Att[1] = par[1], Att[2] = par[3], ... Att[last] = par[2]	
			if(oneAtt.isNumeric()){
				//String tt = parseLine[i];
				//Double db = Double.parseDouble(tt);
				//oneInstance.setValue(oneAtt, db);//parseLine[i]
				oneInstance.setValue(oneAtt, Double.parseDouble(parseLine[i]));
			}
			else{
				//oneAtt.addStringValue(parseLine[i]);
				if(oneAtt.name().equals("atomHydrogenCount")) {
					try {
						oneInstance.setValue(oneAtt, parseLine[i]);
					}catch (Exception e) {
						oneInstance.setValue(oneAtt, "Others");
					}
				}
				else oneInstance.setValue(oneAtt, parseLine[i]);
			}
			if(oneInstance.isMissing(oneAtt)){
				String tt = parseLine[i];
				Double db = Double.parseDouble(tt);
				Integer it = Integer.parseInt(tt);
				oneInstance.setValue(oneAtt, db);
				oneInstance.setValue(oneAtt, it);
				//System.out.println("gg");
			}
		}		
		return oneInstance;
		
	}
}
