package generateInstance;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;


import org.openscience.cdk.CDKConstants;
import org.openscience.cdk.aromaticity.Aromaticity;
import org.openscience.cdk.aromaticity.ElectronDonation;
import org.openscience.cdk.atomtype.SybylAtomTypeMatcher;
import org.openscience.cdk.exception.CDKException;
import org.openscience.cdk.graph.Cycles;
import org.openscience.cdk.graph.PathTools;
import org.openscience.cdk.interfaces.IAtom;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomType;
import org.openscience.cdk.interfaces.IBond;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smiles.SmiFlavor;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.tools.CDKHydrogenAdder;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import generateFeatures.FingerprintQueries;
import generateFeatures.GenerateAtomFeatures;
import generateFeatures.GenerateMolecularFeatures;
import phase2utils.UniqueIDFunctionSet;
import phase2utils.Utilities;
import weka.core.Attribute;
import weka.core.Instance;
import weka.core.DenseInstance;
import weka.core.Instances;

public class GenerateRawFeatures {
	public static final String[] cypList = {"1A2", "2A6", "2B6", "2C8", "2C9", "2C19", "2D6", "2E1", "3A4"};
	public static final String[]	str	= { "C.1", "C.2", "C.3", "C.ar", "C.cat", "N.1",
			"N.2", "N.3", "N.4", "N.ar", "N.am", "N.pl3", "O.2", "O.3", "O.co2", "S.2",
			"S.3", "S.O", "S.O2", "P.3", "F", "Cl", "Br", "I" };
	public int depth;
	String titleLine;
	ArrayList<String> attributes;
	GenerateAtomFeatures gaf_new = new GenerateAtomFeatures();
	
	public GenerateRawFeatures(int depth) {
		this.depth = depth;
	}
	
	public ArrayList<String> generateRawFeatures(IAtomContainer molecule) throws Exception {				
		AtomContainerManipulator.suppressHydrogens(molecule);
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(molecule);
		CDKHydrogenAdder adder = CDKHydrogenAdder.getInstance(molecule.getBuilder());
		Aromaticity aromaticity = new Aromaticity(ElectronDonation.cdk(), Cycles.all());
		aromaticity.apply(molecule);
		UniqueIDFunctionSet.assignUniqueID(molecule);
		adder.addImplicitHydrogens(molecule);		
		HashMap<Integer, String> atomFPMap = generateAtomFPMap(molecule);
		String molecularFeatures = GenerateMolecularFeatures.generateMolecularFeatures_new(molecule);
		ArrayList<String> allBonds = new ArrayList<String>();
		String moleName = (String) molecule.getProperties().get("cdk:Title");
		ArrayList<String> bom_list = Utilities.getBoMs(molecule);
		ArrayList<Integer> bom_idx_list= new ArrayList<>();
		String bom = "F";
		if(bom_list != null) {
			for(int i = 0; i < bom_list.size(); i++) {
				bom_idx_list.add(Integer.parseInt(bom_list.get(i)));
			}			
		}
		if(moleName==null){
			 SmilesGenerator sg = new SmilesGenerator(SmiFlavor.Absolute);
			 moleName = sg.create(molecule); //			 
		}
		/**
		 * Generate atom features
		 */
		//go through all atoms in the molecule
		for(int i = 0; i < molecule.getAtomCount(); i++){
			bom = "F";
			/**
			 * Generate atom features.
			 */
			IAtom oneAtom = molecule.getAtom(i);

			if(bom_idx_list.contains(molecule.indexOf(oneAtom) + 1)) bom = "T";
			
			String atomType = oneAtom.getSymbol(); //Check whether it's a C, B, N, P, S or O atom
			//int numberHydrogens = oneAtom.getImplicitHydrogenCount();

			//If it's a not an Oxygen atom and has more than one Hydrogen atoms attached.
			String atomType_Feature = this.gaf_new.generateAtomType(molecule, oneAtom);
			String atomHydrogenCount = this.gaf_new.generateHydrogenBondDescriptor(molecule,oneAtom);//CH,CH2,CH3, etc
			String[] atomicDescriptorFeatures = this.gaf_new.generateAtomDescriptorFeatures(molecule,oneAtom); // Atomic Features for left Atom	
			String atomicDescriptorFeaturesString = Utilities.convertListToString(atomicDescriptorFeatures);
			String atomEnvironmentDepth_1 = generateEnviromentFeatures(molecule, oneAtom, 1);
			String atomEnvironmentDepth_2 = generateEnviromentFeatures(molecule, oneAtom, 2);
			String atomEnvironmentDepth_3 = generateEnviromentFeatures(molecule, oneAtom, 3);
			String atomEnvironmentDepth_4 = generateEnviromentFeatures(molecule, oneAtom, 4);

			//Here we use depth as the numClosestAtoms.
			String atomClosestAtomFP = this.gaf_new.generateClosestAtomsFPs(molecule, oneAtom, depth);
			String atomClosestAtomType = this.gaf_new.generateClosestAtomsAtomTypes(molecule, oneAtom, depth);

			/**
			 * generateAtomFingerPrints
			 */
			//long start=System.currentTimeMillis();
			String atomFP = generateAtomFingerPrints(molecule, oneAtom, this.depth, false, atomFPMap);
			
			//long end=System.currentTimeMillis();
			//System.out.println("Time cost for atom features: " + (end - start));
			
			String atomFeatures = atomHydrogenCount + "," + atomType_Feature + "," + atomicDescriptorFeaturesString + "," + 
							      atomEnvironmentDepth_1 + "," + atomEnvironmentDepth_2 + "," + atomEnvironmentDepth_3 + "," + atomEnvironmentDepth_4 + "," + 
							      //atomFP;
							      atomClosestAtomFP + "," + atomClosestAtomType + "," + atomFP;//Merged atom feature String
			
			String instanceFinal = moleName + "," + atomType + "." + UniqueIDFunctionSet.getUniqID(oneAtom) + ","
								+ bom + "," + molecularFeatures + "," + atomFeatures;// + "," + bondFeatures + "," + mergedFeature_Left;
			allBonds.add(instanceFinal.replace("NaN", "-99.0")); //The bonds of the molecule is added into the array in the order of the bond
		}								
		return allBonds;
	}
	
	/**
	 * Generate atom fingerprint that satisfies some specific patterns
	 * @param molecule
	 * @return
	 * @throws Exception
	 */
	public String generateAtomFingerPrints(IAtomContainer molecule, IAtom oneAtom, int depth, boolean mergeFP, HashMap<Integer, String> resultMap) throws Exception{
		/**
		 * Always reset the the molecule before processsing it
		 */
		String[] resultFP = new String[(depth+1) * FingerprintQueries.newQueriesList.size()];
		//Initialize resultFP
		for(int i = 0; i < resultFP.length; i++){
			resultFP[i] = "0";
		}
		for(int i = 0; i < depth+1; i++){			
			ArrayList<IAtom> neighAtoms = getNeighAtomsDepth(molecule, oneAtom, depth);
			for(int j = 0; j < neighAtoms.size(); j++){
				IAtom neighAtom = neighAtoms.get(j);
				String[] fp = resultMap.get(neighAtom.getProperty("UniqueID")).split(",");
				//String[] fp = {};
				//String[] fp = GenerateAtomFeatures_new.generateAtomFingeprint(molecule,neighAtom).split(",");

				for(int k = 0; k < fp.length; k++){
					if(fp[k].equals("1")){
						//resultFP[k + FingerPrintQueries_new.newQueriesList.size()*i + GeneratePatterns.queriesList.size()] = "1";
						resultFP[k + FingerprintQueries.newQueriesList.size()*i] = "1";
					}
				}
			}			
		}
		StringBuffer outPut = new StringBuffer();
		for(int i = 0; i < resultFP.length; i++){
			if(i == resultFP.length-1){//If reaches the end
				outPut.append(resultFP[i]);
			}
			else outPut.append(resultFP[i] + ",");
		}
		return outPut.toString();
	}
	
	public ArrayList<IAtom> getNeighAtomsDepth(IAtomContainer molecule, IAtom center, int depth){
		ArrayList<IAtom> resultList = new ArrayList<>();
		List<List<IAtom>> pathCollection = PathTools.getPathsOfLength(molecule, center, depth);
		//System.out.println("center is " + center.getSymbol() + (molecule.indexOf(center)+1));
		for(int i = 0; i < pathCollection.size(); i++){
			IAtom endpoint = pathCollection.get(i).get(pathCollection.get(i).size()-1);
			resultList.add(endpoint);//Get the atom at the given depth
			//System.out.println("End is " + endpoint.getSymbol() + (molecule.indexOf(endpoint)+1));
		}
		return resultList;
		
	}
	/**
	 * Generate an HashMap<String, String> based on the atom's uniqueID within the molecule
	 */
	public HashMap<Integer, String> generateAtomFPMap(IAtomContainer molecule) throws Exception{
		HashMap<Integer, String> resultMap = new HashMap<>();
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(molecule);
		CDKHydrogenAdder adder = CDKHydrogenAdder.getInstance(molecule.getBuilder());
		Aromaticity aromaticity = new Aromaticity(ElectronDonation.cdk(), Cycles.all());
		aromaticity.apply(molecule);
		adder.addImplicitHydrogens(molecule);
		
		for(int i = 0; i < molecule.getAtomCount(); i++){
			IAtom oneAtom = molecule.getAtom(i);
			Integer uniqueID = oneAtom.getProperty("UniqueID");
			if(!resultMap.containsKey(uniqueID)){
				String fp = this.gaf_new.generateAtomFingeprint(molecule,oneAtom);
				resultMap.put(uniqueID, fp);
			}
		
		}
		return resultMap;		
	}
	
	/**
	 * This function is used to generate the environmental features for a atom within the molecule.
	 * @param molecule
	 * @param depth
	 *            : the bond length of an atom to the centered atom.
	 * @return String representation of the environmental features for a single
	 *         molecule where each feature is separated by a comma.
	 *         A (number of atoms) * (24 + 4 tuple vector). Each of the first 24 tuple represents a atom type. Each of the last tuple is either 0 or 1. Tuple(Bond_) 1,2,3,4 means single, double, triple and aromatic 
	 * @throws CDKException
	 */
	public String generateEnviromentFeatures(IAtomContainer molecule, IAtom oneAtom, int depth) throws CDKException {
		IChemObjectBuilder bldr = SilentChemObjectBuilder.getInstance();
		IAtomContainer mol = molecule;
		Aromaticity aromaticity = new Aromaticity(ElectronDonation.cdk(), Cycles.all());
		aromaticity.apply(mol);
		//int dep = depth;
		SybylAtomTypeMatcher satm = SybylAtomTypeMatcher.getInstance(bldr);
		int length = mol.getAtomCount();
		int o;
		int[] atomType = new int[24];
		String enviromentFeatures;
		for (int typeIdx = 0; typeIdx < 24; typeIdx++){
			atomType[typeIdx] = 0;
		}
		int singleBond = 0;
		int doubleBond = 0;
		int tripleBond = 0;
		int aromaticBond = 0;
		List<List<IAtom>> pathCollection = PathTools.getPathsOfLength(mol, oneAtom, depth);
		List<IAtom> atomsOnOnePath;
		IAtom lastSecondAtom, lastAtom;
		for (int j = 0; j < pathCollection.size(); j++) {
			int position = 0;
			atomsOnOnePath = pathCollection.get(j);
			lastSecondAtom = atomsOnOnePath.get(atomsOnOnePath.size() - 1);
			lastAtom = atomsOnOnePath.get(atomsOnOnePath.size() - 2);
			String type_1 = lastSecondAtom.getSymbol();
			try {
				IAtomType s_1 = satm.findMatchingAtomType(mol, lastSecondAtom);
				type_1 = s_1.getAtomTypeName();
			}catch(Exception e) {
				
			}
			for (int k = 0; k < 24; k++) {
				if (type_1.equalsIgnoreCase(str[k]))
					position = k;
			}
			atomType[position] = 1;
			IBond bond = mol.getBond(lastSecondAtom, lastAtom);
			IBond.Order order = bond.getOrder();
			o = order.numeric();
			boolean isAromatic = bond.getFlag(CDKConstants.ISAROMATIC);
			if (isAromatic)
				aromaticBond = 1;
			else if (o == 1)
				singleBond = 1;
			else if (o == 2)
				doubleBond = 1;
			else
				tripleBond = 1;
		}
		StringBuffer strbuf = new StringBuffer();
		for (int n = 0; n < 24; n++) {
			strbuf.append(atomType[n]).append(',');
		}
		strbuf.append(singleBond).append(",");
		strbuf.append(doubleBond).append(",");
		strbuf.append(tripleBond).append(",");
		strbuf.append(aromaticBond);
		enviromentFeatures = strbuf.toString();
			//System.out.println("enviromentFeatures's length" + enviromentFeatures.split(",").length);
		return enviromentFeatures;
	}

	
	
	
}

