package phase2filter;

import static org.junit.Assert.*;

import java.io.File;
import java.util.LinkedHashMap;

import org.apache.commons.lang3.StringUtils;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.exception.CDKException;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smiles.SmilesParser;

import biotransformerapi.BiotransformerAPI;
import biotransformerapi.BiotransformerAPI.ioFormats;
import phase2filter.utils.FileUtils;

public class BiotransformerAPITest {

	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
	}

	@AfterClass
	public static void tearDownAfterClass() throws Exception {
	}

	@Before
	public void setUp() throws Exception {
	}

	@After
	public void tearDown() throws Exception {
	}

	
	@Test
	public void test() throws CDKException, Exception {
		
		BiotransformerAPI btAPI 	= new BiotransformerAPI();
		IChemObjectBuilder 	builder = SilentChemObjectBuilder.getInstance();
		SmilesParser	smiParser	= new SmilesParser(builder);
		IAtomContainer molecule1	= smiParser.parseSmiles("CC1=CC=C(C=C1)C(=O)C2=CC=C(N2C)CC(=O)O");
		IAtomContainer molecule2	= smiParser.parseSmiles("[O-]S(=O)(=O)[O-]");
		
		IAtomContainerSet molecules = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		
		
		molecules.addAtomContainer(molecule1);
		molecules.addAtomContainer(molecule2);
				
		LinkedHashMap<String, Object> parameters = new LinkedHashMap<String, Object>();
		parameters.put("input", molecules);
		parameters.put("inputFormat", "IAtomcontainerSet");		

		
		/*
		 * "Testing: InputFormat (IAtomContainerSet); OutputFormat (IAtomContainerSet)."
		 */		
		System.out.println("Testing: InputFormat (IAtomContainerSet); OutputFormat (IAtomContainerSet).");
		
		parameters.put("outputFormat", "IATOMCONTAINERSET");		
		IAtomContainerSet results = (IAtomContainerSet) btAPI.predict(parameters);
		assertTrue("Molecule 1 must be a PhaseII Candidate.", ((String)results.getAtomContainer(0).getProperty("Phase2Filter")).contentEquals("True"));
		assertTrue("Molecule 2 must not be a PhaseII Candidate.", ((String)results.getAtomContainer(1).getProperty("Phase2Filter")).contentEquals("False"));

		/*
		 * "Testing: InputFormat (IAtomContainerSet); OutputFormat (SDFile)."
		 */
		System.out.println("Testing: InputFormat (IAtomContainerSet); OutputFormat (SDFile).");
		
		parameters.put("outputFormat", "sdfile");
		parameters.put("output", "testResults.sdf");
		String additionaResults = (String) btAPI.predict(parameters);
		File resultsSDFile = new File(additionaResults);
		assertTrue("The output with format SDFile named " + parameters.get("output") + " could not be found."
				, resultsSDFile.exists() && !resultsSDFile.isDirectory());
		results = FileUtils.parseSdf(additionaResults);
		assertTrue("Molecule 1 must be a PhaseII Candidate.", ((String)results.getAtomContainer(0).getProperty("Phase2Filter")).contentEquals("True"));
		assertTrue("Molecule 2 must not be a PhaseII Candidate.", ((String)results.getAtomContainer(1).getProperty("Phase2Filter")).contentEquals("False"));
	
		/*
		 * "Testing: InputFormat (SDFile); OutputFormat (SDFile)."
		 */		
		System.out.println("Testing: InputFormat (SDFile); OutputFormat (SDFile).");		
		
		parameters.put("inputFormat", "sdfile");
		parameters.put("outputFormat", "sdfile");
		String inputFileName = "testMolecules.sdf";
		
		FileUtils.saveAtomContainerSetToSDF(molecules, inputFileName);
		parameters.put("input",  "testMolecules.sdf");
		additionaResults = (String) btAPI.predict(parameters);
		resultsSDFile = new File(additionaResults);
		assertTrue("The output with format SDFile named " + parameters.get("output") + " could not be found."
				, resultsSDFile.exists() && !resultsSDFile.isDirectory());
		results = FileUtils.parseSdf(additionaResults);
		assertTrue("Molecule 1 must be a PhaseII Candidate.", ((String)results.getAtomContainer(0).getProperty("Phase2Filter")).contentEquals("True"));
		assertTrue("Molecule 2 must not be a PhaseII Candidate.", ((String)results.getAtomContainer(1).getProperty("Phase2Filter")).contentEquals("False"));
		
		File inputFile = new File(inputFileName);
		inputFile.delete();
		resultsSDFile.delete();
		
		
	}

}
