package phase2filter;

import static org.junit.Assert.*;

import java.util.ArrayList;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.exception.CDKException;
import org.openscience.cdk.exception.InvalidSmilesException;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;

import biotransformerapi.BiotransformerAPI.ioFormats;
import phase2filter.prediction.P2Filter;

public class P2FilterTest {

	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
	}

	@AfterClass
	public static void tearDownAfterClass() throws Exception {
	}

	@Before
	public void setUp() throws Exception {
	}

	@After
	public void tearDown() throws Exception {
	}

	@Test
	public void test() throws CDKException, Exception {
		
		
		IChemObjectBuilder 	builder = SilentChemObjectBuilder.getInstance();
		SmilesGenerator smiGen 		= new SmilesGenerator().isomeric(); 	
		SmilesParser	smiParser	= new SmilesParser(builder);
		P2Filter p2f				= new P2Filter();
		
		IAtomContainer molecule1	= smiParser.parseSmiles("CC1=CC=C(C=C1)C(=O)C2=CC=C(N2C)CC(=O)O");
		IAtomContainer molecule2	= smiParser.parseSmiles("[O-]S(=O)(=O)[O-]");
		
		IAtomContainerSet molecules = DefaultChemObjectBuilder
				.getInstance().newInstance(IAtomContainerSet.class);
		
		molecules.addAtomContainer(molecule1);
		molecules.addAtomContainer(molecule2);
		
		ArrayList<String> results = p2f.filter(molecules);
		
		assertTrue("Molecule 1 must be a PhaseII Candidate.", (results.get(0) == "T"));
		assertTrue("Molecule 2 is not a PhaseII Candidate.",  (results.get(1) == "F"));
		
		System.out.println("Molecule 1: " + p2f.filter(molecule1));
		
	}

}
