package phase2filter.prediction;


import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.ObjectInputStream;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.io.ObjectInputStream;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.exception.CDKException;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import phase2filter.utils.ChemStructureManipulator;
import phase2filter.utils.FeatureGenerator;
import phase2filter.utils.StructSearcher;
import weka.classifiers.Classifier;
import weka.classifiers.trees.RandomForest;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.converters.CSVSaver;

public class P2Filter {
	
	public LinkedHashMap<String, Object> properties =  new LinkedHashMap<String, Object>();
	
	public P2Filter() {
		// TODO Auto-generated constructor stub	
		 this.properties.put("path", (Object) "supportfiles/P2Filter");
	}
	
	public P2Filter(LinkedHashMap<String, Object> props) {
		// TODO Auto-generated constructor stub
		 this.properties = props;
	}
	
	public Instances generateWekaInputFeatures(IAtomContainer molecule, boolean preprocess) throws CDKException, Exception{
		IAtomContainerSet molecules  = DefaultChemObjectBuilder
				.getInstance().newInstance(IAtomContainerSet.class);
		molecules.addAtomContainer(molecule);
		return generateWekaInputFeatures(molecules, preprocess);
	}
	
	
	public Instances generateWekaInputFeatures(IAtomContainerSet molecules, boolean preprocess) throws CDKException, Exception{
		String[] header = ("nAcidicGroups	tpsa	aLogp	nHBA	nHBD	monoisoMass	nRotBonds	nBasicGroups	"
				+ "heteroatom	hydroxyl_acidic	unsubstituted_aromatic_carbon	carboxyl	carbon_hydrogen_bond	"
				+ "alcohol	carboxylic_acid_or_base	hydroxyl	unfunctionalized_methylen_group	hBond	"
				+ "12_diol	ewg	heterocyclic_amine	imine	amine	carboxylic_ester	peroxide").split("\t");
		
		ArrayList<Attribute> atts = new ArrayList<Attribute>();
		
		/*
		 * Adding labels
		 */
		for(String label : header){
			Attribute att = new Attribute(label.trim());
			atts.add(att);
		}
		/*
		 * Add class attribute as required by weka.
		 */
		Attribute classAttribute = new Attribute("class"); 
		atts.add(classAttribute);
		Instances inputFeatures = new Instances("Rel", atts, 100000);
		
		/*
		 * Generate features with the FeatureGenerator
		 */
		FeatureGenerator fgen =  new FeatureGenerator();
		
		for(IAtomContainer molecule : molecules.atomContainers()){
//			IAtomContainer moleculeClone = molecule.clone();
			
			if(preprocess) {
//				moleculeClone = ChemStructureManipulator.preprocessContainer(moleculeClone);
//				AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(moleculeClone);
//				AtomContainerManipulator.convertImplicitToExplicitHydrogens(moleculeClone);	
//				moleculeClone = ChemStructureManipulator.preprocessContainer(molecule);
				molecule = ChemStructureManipulator.preprocessContainer(molecule);
			}
			AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(molecule);
			AtomContainerManipulator.convertImplicitToExplicitHydrogens(molecule);
			
			LinkedHashMap<String, Double> pcprops = fgen.generateAllFeatures(molecule);
//			System.out.println("pcprops");
//			System.out.println(pcprops);
			
			/*
			 * fill in the attributes
			 */

			Instance sample = new DenseInstance(atts.size());
			for(int i = 0; i < header.length; i++){
				Attribute att = atts.get(i);
				sample.setValue(att, pcprops.get(header[i]));
			}
			sample.setValue(classAttribute, 0.0);
			inputFeatures.add(sample);
		}
		
		
		
//		CSVSaver saver = new CSVSaver();
//		saver.setInstances(inputFeatures);
//		saver.setFile(new File("data/temp.csv"));
//		saver.writeBatch();

		
		return inputFeatures;
	}
	
	
	public ArrayList<String> filter(IAtomContainer molecule) throws CDKException, Exception{
		Path modelpath 	= Paths.get((String)this.properties.get("path"), "model", FilenameUtils.getName("P2Filter.model"));
		Path supportfilepath 	= Paths.get((String)this.properties.get("path"), FilenameUtils.getName("supportfile.csv"));	
		return filter(modelpath.toString(), molecule, supportfilepath.toString());
	}
	
	
	public ArrayList<String> filter(String modelpath, IAtomContainer molecule, 
			String supportfilepath) throws CDKException, Exception{
		
		IAtomContainerSet molecules  = DefaultChemObjectBuilder
				.getInstance().newInstance(IAtomContainerSet.class);
		molecules.addAtomContainer(molecule);
		return filter(modelpath, molecules, supportfilepath);
	}
	
	public ArrayList<String> filter(IAtomContainerSet moleculeSet) throws CDKException, Exception{
		Path modelpath 	= Paths.get((String)this.properties.get("path"), "model", FilenameUtils.getName("P2Filter.model"));
		Path supportfilepath 	= Paths.get((String)this.properties.get("path"), FilenameUtils.getName("supportfile.csv"));	
		return filter(modelpath.toString(), moleculeSet, supportfilepath.toString());
	}
	
	public ArrayList<String> filter(String modelpath, IAtomContainerSet moleculeSet, 
			String supportfilepath) throws CDKException, Exception{
		ArrayList<String> results = new ArrayList<String>();
	
		LinkedHashMap<Integer, Integer> indexMappings = new LinkedHashMap<Integer, Integer>();
		
		IAtomContainerSet molecules =  DefaultChemObjectBuilder.getInstance().newInstance(
				IAtomContainerSet.class);
		
		
		for(int i = 0; i<moleculeSet.getAtomContainerCount(); i++){
			
			IAtomContainer atc = moleculeSet.getAtomContainer(i);
			AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(atc);
			AtomContainerManipulator.convertImplicitToExplicitHydrogens(atc);
			
			
			/*
			 * Can I set the mass to be less than 900 because in my phase II filter set, the largest substract (nevertoxin 2) has a mass of 894.48 Da, followed by
			 * tacrolimus with a mass of 803.49 Da ?  && FeatureGenerator.getMonoisotopicMass(atc)< 900.0 
			 */
			
			if( FeatureGenerator.getMonoisotopicMass(atc)< 900.0 && StructSearcher.isPotentialPhaseIICandidateByReactionPatternMatching(atc)){
				
				results.add("T");
				indexMappings.put(molecules.getAtomContainerCount(), i);
				molecules.addAtomContainer(atc);
			} else{
				results.add("F");				
			}
		}
		
//		System.out.println(molecules.getAtomContainerCount() + " out of " + results.size() + " compound(s) passed the first filter");
//		System.out.println(indexMappings);
		
		if(molecules.getAtomContainerCount()>0){
			Instances candidateInstances = this.generateWekaInputFeatures(molecules,true);
			ObjectInputStream iStream = new ObjectInputStream(new FileInputStream(modelpath));
			RandomForest rf = (RandomForest) (iStream).readObject();
			Classifier cls = (Classifier) weka.core.SerializationHelper.read(modelpath);
//			AbstractClassifier cls = (AbstractClassifier) weka.core.SerializationHelper.read(model);
			
			BufferedReader supFileBreader = new BufferedReader(new FileReader(supportfilepath));
		
			/*
			 * Create attribute arraylist,mean arraylist, max arraylist, min arraylist
			 */
			ArrayList<String> attList = new ArrayList<String>();
			ArrayList<String> meanList = new ArrayList<String>();
			ArrayList<String> maxList = new ArrayList<String>();
			ArrayList<String> minList = new ArrayList<String>();
			
			/*
			 * Write the output 
			 */
			int counter = 0;
			String supLine = supFileBreader.readLine();
			while((supLine = supFileBreader.readLine())!=null){
				String[] elements = supLine.split(",");
				attList.add(elements[0]);
				meanList.add(elements[1]);
				maxList.add(elements[2]);
				minList.add(elements[3]);
			}
//			int countR = 0;
//			int countT = 0;
			
			
			Instances matched = matchAttributes(candidateInstances,attList,meanList,maxList,minList);
			matched.setClassIndex(matched.numAttributes()-1);
//			System.out.println(matched.size());
			for(int i = 0; i<matched.size();i++){
				counter++;
//				System.out.println(i);
//				Instance oneSample = matched.get(i);
				Instance oneSample = matched.instance(i);
				double result = rf.classifyInstance(oneSample);
//				double result = cls.classifyInstance(oneSample);
//				System.out.println(result);
				
//				String prediction=matched.classAttribute().value((int)result);
//				System.out.println("prediction: " + prediction);
				
				if(result == 1.0){
//					results.add("F");
//					System.out.println(indexMappings.get(i));
//					System.out.println("previous value " + results.get(indexMappings.get(i)));
//					results.set(indexMappings.get(i), "F");
//					System.out.println("Now: " + results.get(indexMappings.get(i)));
				}
				else if(result == 0.0){
//					System.out.println("previous value " + results.get(indexMappings.get(i)));
//					System.out.println("Value remains the same");
				}
			}
			iStream.close();
			supFileBreader.close();
		}
		
		
//		StringUtils.join(results,"\t");
	
//		System.out.println(StringUtils.join(results,"\n"));
		
		return results;
	
	}

	
	public IAtomContainerSet returnFilteredPhaseIICandidates(IAtomContainerSet moleculeSet) throws CDKException, Exception{
		IAtomContainerSet filteredMolecules =  DefaultChemObjectBuilder.getInstance().newInstance(
				IAtomContainerSet.class);
		
		ArrayList<String> attributes = this.filter(moleculeSet);
		
		
		for(int i = 0; i < moleculeSet.getAtomContainerCount(); i++){
			
			if(attributes.get(i) == "T"){
				filteredMolecules.addAtomContainer(moleculeSet.getAtomContainer(i));
			}
		}
		
		return filteredMolecules;
	}
	
	
	public LinkedHashMap<String, IAtomContainerSet> partitionSetForPhaseIIMetabolism(IAtomContainerSet moleculeSet) throws CDKException, Exception{
		 LinkedHashMap<String, IAtomContainerSet> partitionedMolecules =  new  LinkedHashMap<String, IAtomContainerSet>();
		 
		 partitionedMolecules.put("phaseIISubstrates", DefaultChemObjectBuilder.getInstance().newInstance(
					IAtomContainerSet.class));
		 partitionedMolecules.put("phaseIINonSubstrates", DefaultChemObjectBuilder.getInstance().newInstance(
					IAtomContainerSet.class));
		 
		 
		 ArrayList<String> attributes = this.filter(moleculeSet);
		 
		 
			for(int i = 0; i < moleculeSet.getAtomContainerCount(); i++){
				
				if(attributes.get(i) == "T"){
					partitionedMolecules.get("phaseIISubstrates").addAtomContainer(moleculeSet.getAtomContainer(i));
				}
				else if(attributes.get(i) == "F"){
					partitionedMolecules.get("phaseIINonSubstrates").addAtomContainer(moleculeSet.getAtomContainer(i));
				}
			}
		 
		 
		 return partitionedMolecules;
	}
	
	
//	public IAtomContainerSet filterandReturnAnnotatedContainers(String model, IAtomContainerSet molecules, 
//			String supportfile) throws CDKException, Exception{
//		IAtomContainerSet results = molecules;
//	
//		Instances candidateInstances = this.generateWekaInputFeatures(molecules);
//		Classifier cls = (Classifier) weka.core.SerializationHelper.read(model);
//		
//		BufferedReader supFileBreader = new BufferedReader(new FileReader(supportfile));
//	
//		/*
//		 * Create attribute arraylist,mean arraylist, max arraylist, min arraylist
//		 */
//		ArrayList<String> attList = new ArrayList<String>();
//		ArrayList<String> meanList = new ArrayList<String>();
//		ArrayList<String> maxList = new ArrayList<String>();
//		ArrayList<String> minList = new ArrayList<String>();
//		
//		/*
//		 * Write the output 
//		 */
//		int counter = 0;
//		String supLine = supFileBreader.readLine();
//		while((supLine = supFileBreader.readLine())!=null){
//			String[] elements = supLine.split(",");
//			attList.add(elements[0]);
//			meanList.add(elements[1]);
//			maxList.add(elements[2]);
//			minList.add(elements[3]);
//		}
////		int countR = 0;
////		int countT = 0;
//		
//		Instances matched = matchAttributes(candidateInstances,attList,meanList,maxList,minList);
//		matched.setClassIndex(matched.numAttributes()-1);
//		
//		for(int i = 0; i<matched.size();i++){
//			counter++;
//			Instance oneSample = matched.get(i);
//			double result = cls.classifyInstance(oneSample);
//			if(result == 1){
////				results.add("T");
//			}
//			else if(result == 0){
////				results.add("F");
//			}
//		}
//		StringUtils.join(results,"\t");
//
//		
//	return results;
//	
//	}
	
	
	/**
	 * Apply feature selection and normalization on the given molecule
	 * @param Instances testSet, ArrayList<String> attList, ArrayList<String> meanList, ArrayList<String> maxList, ArrayList<String> minList
	 * @return weka Instances with all raw feature values         
	 * @throws Exception
	 */
	public Instances matchAttributes(Instances candidateInstances, ArrayList<String> attList, ArrayList<String> meanList, ArrayList<String> maxList, ArrayList<String> minList) throws Exception{
		
		List my_nominal_values = new ArrayList(3); 
		my_nominal_values.add("T"); 
		my_nominal_values.add("F"); 
		Attribute lastAttribute  = new Attribute(attList.get(attList.size()-1), my_nominal_values);
		// Create nominal attribute "position" 
		
		ArrayList<Attribute> atts = new ArrayList<Attribute>();
		for(int i = 0; i<attList.size();i++){
			if(i<attList.size()-1){
				Attribute Attribute = new Attribute(attList.get(i));
				atts.add(Attribute);
			}
			else{
				
				atts.add(lastAttribute);
			}
		}
		
		int length = candidateInstances.size();
		int numAttributes = attList.size();
		Instances matched = new Instances("Rel", atts, length);
		Instance sample = new DenseInstance(numAttributes);
		
		for(int j = 0; j< length; j++){
			ArrayList misFeatures = new ArrayList();
			Instance temp = candidateInstances.get(j);
			for(int vidx = 0; vidx < numAttributes; vidx++){
				  Attribute att = atts.get(vidx);
				  if(vidx<numAttributes-1){
					  
					  String stAtt = att.toString();
					  String[] whatever = stAtt.split(" ");
					  
					  if(whatever.length!=3){
						  String combineSpaces = "";
						  for(int kk = 1; kk<whatever.length-1; kk++){
							  if(kk<whatever.length - 2){
								  combineSpaces = combineSpaces + whatever[kk]+ " ";
							  }
							  else combineSpaces = combineSpaces + whatever[kk];
						  }
						  whatever[1] = combineSpaces;
					  }
					 
					  if(whatever[1].contains("'")){
						  whatever[1] = whatever[1].replace("'", "");
					  }
//					  System.out.println(whatever[1]);
					  Attribute convAtt = candidateInstances.attribute(whatever[1]);
					  double vle = temp.value(convAtt);
					  if(!Double.isNaN(vle)){
						  double min = Double.parseDouble(minList.get(vidx));
						  double max = Double.parseDouble(maxList.get(vidx));
						  //double normedVal = (vle-min)/(max-min);
						  double normedVal = vle;
						  /*
						  if(normedVal > max){
							  throw new Exception("Attribute value error");
						  }
						  */
						  sample.setValue(att, normedVal);
					  }
					  else{
						  //Add the name of the missing feature into the misFeatures ArrayList
						  misFeatures.add(whatever[1]);
						  //System.out.println("Molecule--" + j + "contains missing feature, synthesized");
						  double syn = Double.parseDouble(meanList.get(vidx));
						  sample.setValue(att, syn);

					  }
				  }
				  else{
					  //Set class with random value. Does not matter
					  sample.setValue(lastAttribute, "T");
					  
				  }
			}
			//Output those missing features
			if(misFeatures.size()>0){
				int numMisFeatures = misFeatures.size();
				String misFeatureString = "Molecule" + j + ": " +  numMisFeatures + 
						" features could not be computed from the structure representation, and were thus synthesized: " + misFeatures.get(0);
				for(int i = 1; i < numMisFeatures; i++){
					misFeatureString = misFeatureString + "," + misFeatures.get(i);
				}
				System.out.println(misFeatureString);
			}
			
			
			matched.add(sample);
		}
		
		return matched;
		
	}
	
	
}
