package phase2filter.utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import org.openscience.cdk.CDKConstants;
import org.openscience.cdk.exception.CDKException;
import org.openscience.cdk.exception.InvalidSmilesException;
import org.openscience.cdk.inchi.InChIGenerator;
import org.openscience.cdk.inchi.InChIGeneratorFactory;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.io.SDFWriter;
import org.openscience.cdk.libio.jena.CDK;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import phase2filter.prediction.P2Filter;
import weka.core.Instances;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;

public class DataSetManipulator {

	StructSearcher csearcher = new StructSearcher();
	FeatureGenerator fgen =  new FeatureGenerator();
	IChemObjectBuilder builder = SilentChemObjectBuilder.getInstance();
	SmilesGenerator sg = new SmilesGenerator().isomeric();
	SmilesParser sp = new SmilesParser(builder);
	InChIGeneratorFactory inchiFactory = InChIGeneratorFactory.getInstance();
	
	public DataSetManipulator() throws Exception {
		// TODO Auto-generated constructor stub
	}

	public void convertOriginalToSDF(String fileName) throws Exception{
		BufferedReader bReader = new BufferedReader(new FileReader(fileName));
		String line;
		
		int line_nr = 0;
		
		LinkedHashMap<String, IAtomContainer> phaseIISubs = new LinkedHashMap<String, IAtomContainer>();
		LinkedHashMap<String, IAtomContainer> phaseIINonSubs = new LinkedHashMap<String, IAtomContainer>();
		String references;
		int withPhaseIIGroup = 0;
		
		while((line = bReader.readLine()) != null ){
			line_nr++;
			
			if(line_nr>=2 && line_nr <1171){
				System.out.println("line nr. " + line_nr);
				String[] sline = line.trim().split("\t");
				if(sline[3] != null){
					IAtomContainer atc = this.sp.parseSmiles(sline[3]);
					atc.setProperty(CDKConstants.TITLE, sline[0].trim());
					AtomContainerManipulator.convertImplicitToExplicitHydrogens(atc);
					atc = ChemStructureManipulator.preprocessContainer(atc);
					
					references = "";
					if(sline.length>9 && sline[9]!= null){
						references = references + sline[9].trim()+"\n";
					}
					if(sline.length>12 && sline[12]!= null){
						references = references + sline[12].trim()+"\n";
					}
					if(sline.length>15 && sline[15]!= null){
						references = references + sline[15].trim()+"\n";
					}
					if(sline.length>18 && sline[18]!= null){
						references = references + sline[18].trim()+"\n";
					}
					if(sline.length>21 && sline[21]!= null){
						references = references + sline[21].trim()+"\n";
					}					
					if(sline.length>24 && sline[24]!= null){
						references = references + sline[24].trim()+"\n";
					}						
					if(sline.length>27 && sline[27]!= null){
						references = references + sline[27].trim()+"\n";
					}
					
					
					if(! (sline[5].contentEquals("NO") || sline[5].contentEquals("YES"))){
						System.err.println("make sure that " + sline[5] + " = YES or NO.");
						break;
					}
					
//					if(sline[5].contentEquals("NO")){
//						if(sline.length>6 && sline[5]!= null){
//							references = sline[5];
//						}
//						else{
//							references = "";
//						}
//						
//					}
//					else if(sline[5].contentEquals("YES")){
//						references = "";
//						if(sline.length>9 && sline[9]!= null){
//							references = references + sline[9].trim()+"\n";
//						}
//						if(sline.length>12 && sline[12]!= null){
//							references = references + sline[12].trim()+"\n";
//						}
//						if(sline.length>15 && sline[15]!= null){
//							references = references + sline[15].trim()+"\n";
//						}
//						if(sline.length>18 && sline[18]!= null){
//							references = references + sline[18].trim()+"\n";
//						}
//						if(sline.length>21 && sline[21]!= null){
//							references = references + sline[21].trim()+"\n";
//						}					
//						if(sline.length>24 && sline[24]!= null){
//							references = references + sline[24].trim()+"\n";
//						}						
//						if(sline.length>27 && sline[27]!= null){
//							references = references + sline[27].trim()+"\n";
//						}					
//					}
//					else{
//						System.err.println("make sure that " + sline[5] + " = YES or NO.");
//						break;
//					}
					
					System.out.println(references.trim());

					InChIGenerator iGen = this.inchiFactory.getInChIGenerator(atc);					
					
					if(sline[5].contentEquals("NO") && phaseIISubs.containsKey(atc.getProperty("InChIKey")) || 
							sline[5].contentEquals("YES") && phaseIINonSubs.containsKey(atc.getProperty("InChIKey"))) {
						System.err.println(sline[0] + "(" + phaseIISubs.get(atc.getProperty("InChIKey")).getProperty(CDKConstants.TITLE) +  ") cannot be both a sub and non-sub.");
						break;
					}
					else if (sline[5].contentEquals("YES")){
						if(phaseIISubs.containsKey(atc.getProperty("InChIKey"))){
							System.err.println(sline[0] + "(" + phaseIISubs.get(atc.getProperty("InChIKey")).getProperty(CDKConstants.TITLE) +  ") are duplicates");
							break;
						}else{
							atc.setProperty("InChIKey", iGen.getInchiKey().toString());
							atc.setProperty("PhaseIISub","Y");
							atc.setProperty("References", references.trim());
							phaseIISubs.put(atc.getProperty("InChIKey").toString(), atc);
						}
						
					}
					else if (sline[5].contentEquals("NO")){
						if(phaseIINonSubs.containsKey(atc.getProperty("InChIKey"))){
							System.err.println(sline[0] + "(" + phaseIINonSubs.get(atc.getProperty("InChIKey")).getProperty(CDKConstants.TITLE) +  ") are duplicates");
							break;
						}else{
							atc.setProperty("InChIKey", iGen.getInchiKey().toString());
							atc.setProperty("PhaseIISub","N");
							atc.setProperty("References", references.trim());
							phaseIINonSubs.put(atc.getProperty("InChIKey").toString(), atc);
						}
					}					
				}
				else {
					System.err.println(sline[0] + " is missing a SMILES string.");
					break;
				}
			}	
		}
		System.out.println("Phase II substrates: " + phaseIISubs.size());
		System.out.println("Phase II non-substrates: " + phaseIINonSubs.size());
		
		
		SDFWriter sdfWr = new SDFWriter(new FileOutputStream("data/PhaseIIFilterDataSet-with-phase2-group.sdf"));
		
		for(Map.Entry<String, IAtomContainer> sub : phaseIISubs.entrySet()){
			sdfWr.write(sub.getValue());
		}
		for(Map.Entry<String, IAtomContainer> nonSub : phaseIINonSubs.entrySet()){
			
			if(StructSearcher.isPotentialPhaseIICandidateByReactionPatternMatching(nonSub.getValue())){
				sdfWr.write(nonSub.getValue());
				withPhaseIIGroup++;
			}
			
			
		}	
		System.out.println("Non-substrates with phase II groups: " + withPhaseIIGroup);
		sdfWr.close();
		bReader.close();
	}
	
	
	public void generatePropertiesAndSaveToTSV(String SdfFileName) throws Exception{
		
		IAtomContainerSet containers = FileUtils.parseSdf(SdfFileName);
		FeatureGenerator fgen = new FeatureGenerator();
		System.out.println(SdfFileName);
		System.out.println(SdfFileName.split(".sdf").length);
		BufferedWriter bWriter = new BufferedWriter(new FileWriter(SdfFileName.split(".sdf")[0] + "_all_features.tsv" ));
		
		LinkedHashMap<String, Double> first = fgen.generateAllFeatures(containers.getAtomContainer(0));
		
		ArrayList<String> headers = new ArrayList<String>();
		headers.add("Compound");
		for(String feat : first.keySet()) {
			headers.add(feat);
		}
		bWriter.write(StringUtils.join(headers, "\t") + "\t" + "phase2sub");
		bWriter.newLine();
		for(IAtomContainer atc : containers.atomContainers()){

			ArrayList<Double> allFeatures = fgen.generateAllFeaturesToArrayList(atc);
			bWriter.write(atc.getProperty(CDKConstants.TITLE).toString() + "\t" + StringUtils.join(allFeatures, "\t") + "\t" + atc.getProperty("PhaseIISub"));
			bWriter.newLine();
		}
		
		bWriter.close();
	}
	
	public void removePhaseIIMetabolitesAndSaveToSDF(String tsvFileName) throws Exception{
		
		BufferedReader bReader = new BufferedReader (new FileReader(tsvFileName));
		SDFWriter sdfWriter = new SDFWriter(new FileOutputStream(FilenameUtils.getFullPathNoEndSeparator(tsvFileName) + 
				"/" + FilenameUtils.getBaseName(tsvFileName)  + ".sdf" ));
		String bioTline;
		
		int btCounter = 0;
		int removed = 0;
		
		while((bioTline = bReader.readLine()) != null){
			String[] sline = bioTline.split("\t");
			btCounter++;
			
			LinkedHashMap<String, IAtomContainer> containersByInChIKey = new LinkedHashMap<String, IAtomContainer>();
			LinkedHashMap<String, IAtomContainer> containers = new LinkedHashMap<String, IAtomContainer>();


			System.out.println(sline[0]);
			IAtomContainer container = sp.parseSmiles(sline[2]);
			InChIGeneratorFactory inchiGenFactory = InChIGeneratorFactory.getInstance();
			InChIGenerator gen0 = inchiGenFactory.getInChIGenerator(container);
			String inchikey = gen0.getInchiKey();
//			System.out.println(inchikey);
			if(!containersByInChIKey.containsKey(inchikey)){
				if(!StructSearcher.isPotentialPhaseIIMetaboliteByReactionPatternMatching(container)){
//					System.out.println("Adding molecule");
					container.setProperty("Database ID", sline[0]);
					container.setProperty(CDKConstants.TITLE, sline[1]);
					containersByInChIKey.put(inchikey, container);					
					sdfWriter.write(container);
				}
				else{
					System.err.println(sline [0] + "( "+ sline[1] + " ) is already a phase II metabolite\n\n");
					removed++;
				}
			}

		}
		sdfWriter.close();
		
		System.out.print("Removed " + removed  + " molecules.");
	}
	
	public static void main(String[] args) throws Exception{
		DataSetManipulator dm = new DataSetManipulator();
		
//		dm.convertOriginalToSDF("data/PhaseIIFilterDataSet.tsv");
//		dm.generatePropertiesAndSaveToTSV("data/PhaseIIFilterDataSet-with-phase2-group.sdf");
//		dm.generatePropertiesAndSaveToTSV("data/phaseIIFilter-testSet-3D.sdf");
//		dm.removePhaseIIMetabolitesAndSaveToSDF("/Users/yandj/Projects/DB_data/Drugbank/Drugbank-5.0/drugMetabolites-DrugBank40-sep092017_mini.tsv");
//
		IAtomContainer  mol = dm.sp.parseSmiles("CC1=CC(NS(=O)(=O)C2=CC=C(N)C=C2)=NO1");
////		System.out.print(mol);
		P2Filter p2f = new P2Filter();
//		
////		Instances i = p2f.generateWekaInputFeatures(mol);
//		IAtomContainerSet containers = FileUtils.parseSdf("data/PhaseIIFilterDataSet-with-phase2-group.sdf");
		ArrayList<String> res = p2f.filter("supportfiles/P2Filter/phaseIIFilterRForest-Sep5.model", mol, "supportfiles/P2Filter/supportfile.csv");
//		ArrayList<String> res = p2f.filter("supportfiles/P2Filter/phaseIIFilterRForest-Sep5.model", containers, "supportfiles/P2Filter/supportfile.csv");
//////		p2f.filter("supportfiles/phaseIIFilterRForest-Sep5.model", containers, "supportfiles/supportfile.csv");
////		ArrayList<String> res = p2f.filter(mol);
////		System.out.println(res.size());
		System.out.println("Phase 2 prediction: " +  res.get(0));
//		int pos = 0;
//		int neg = 0;
//		for(int i  =0; i < res.size(); i++){
//			System.out.println(i + " : " + res.get(i));
//			if (res.get(i) == "T"){
//				pos++;
//			}
//			else{
//				neg++;
//			}
//		}
//		System.out.println("Phase 2 substrates: 	" + pos);
//		System.out.println("Phase 2 non-substrates: " + neg);
	}

	

}

