package phase2filter.utils;


import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.openscience.cdk.exception.CDKException;
import org.openscience.cdk.exception.InvalidSmilesException;
import org.openscience.cdk.geometry.surface.NumericalSurface;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.qsar.DescriptorValue;
import org.openscience.cdk.qsar.descriptors.molecular.ALOGPDescriptor;
import org.openscience.cdk.qsar.descriptors.molecular.AcidicGroupCountDescriptor;
import org.openscience.cdk.qsar.descriptors.molecular.BasicGroupCountDescriptor;
import org.openscience.cdk.qsar.descriptors.molecular.CPSADescriptor;
import org.openscience.cdk.qsar.descriptors.molecular.HBondAcceptorCountDescriptor;
import org.openscience.cdk.qsar.descriptors.molecular.HBondDonorCountDescriptor;
import org.openscience.cdk.qsar.descriptors.molecular.RotatableBondsCountDescriptor;
import org.openscience.cdk.qsar.descriptors.molecular.TPSADescriptor;
import org.openscience.cdk.qsar.descriptors.molecular.WeightDescriptor;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;
import org.openscience.cdk.tools.manipulator.MolecularFormulaManipulator;




public class FeatureGenerator {
	
	ALOGPDescriptor aLogp = new ALOGPDescriptor();
	CPSADescriptor cpsad = new CPSADescriptor();
	TPSADescriptor tpsa = new TPSADescriptor();
	WeightDescriptor weightd = new WeightDescriptor();
	AcidicGroupCountDescriptor agcd = new AcidicGroupCountDescriptor();
	BasicGroupCountDescriptor bgcd = new BasicGroupCountDescriptor();
	HBondAcceptorCountDescriptor hbca = new HBondAcceptorCountDescriptor();
	HBondDonorCountDescriptor hbcd = new HBondDonorCountDescriptor();
	RotatableBondsCountDescriptor rbcd = new RotatableBondsCountDescriptor();
	IChemObjectBuilder builder = SilentChemObjectBuilder.getInstance();
	StructSearcher csearcher =  new StructSearcher();

	public static void main(String[] args) throws Exception{
		FeatureGenerator fgen =  new FeatureGenerator();
		SmilesGenerator sg = new SmilesGenerator();
		SmilesParser sp = new SmilesParser(fgen.builder);
		IAtomContainer molecule = sp
				.parseSmiles("CC(C)CC(=O)C1=C(O)C(CC=C(C)C)=C(O)C2=C1OC(=O)C=C2");
		
		IAtomContainer pmolecule = ChemStructureManipulator.preprocessContainer(molecule);
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(pmolecule);
		
//		LinkedHashMap<String, Double> pcprops = fgen.generatePhysChemProperties(pmolecule);
		LinkedHashMap<String, Double> pcprops = fgen.generateAllFeatures(pmolecule);
		System.out.println(StringUtils.join(pcprops.keySet(), "\t") );
		
		System.out.println(pcprops);
		
	}
	
	public FeatureGenerator()  throws CDKException, Exception{
		// TODO Auto-generated constructor stub
		
		this.agcd.initialise(this.builder);
		this.bgcd.initialise(this.builder);
	}
	
	
	public LinkedHashMap<String, Double> generatePhysChemProperties(IAtomContainer molecule){
		LinkedHashMap<String, Double> pcprops = new LinkedHashMap<String, Double>();
		NumericalSurface nsd = new NumericalSurface(molecule);	
		DescriptorValue cpsa = this.cpsad.calculate(molecule);
				
//		nsd.calculateSurface();
//		pcprops.put("ssa", Double.valueOf(nsd.getTotalSurfaceArea()));
//		pcprops.put("totalFormalCharge", Double.valueOf(AtomContainerManipulator.getTotalFormalCharge(molecule)));
		
		pcprops.put("nAcidicGroups", Double.valueOf(agcd.calculate(molecule).getValue().toString()));
		pcprops.put("tpsa", Double.valueOf(tpsa.calculate(molecule).getValue().toString()));
		pcprops.put("aLogp", Double.valueOf(aLogp.calculate(molecule).getValue().toString().split(",")[0]));
		pcprops.put("nHBA", Double.valueOf(hbca.calculate(molecule).getValue().toString()));
		pcprops.put("nHBD", Double.valueOf(hbcd.calculate(molecule).getValue().toString()));		
		pcprops.put("monoisoMass", MolecularFormulaManipulator.getMajorIsotopeMass(MolecularFormulaManipulator.getMolecularFormula(molecule)));
		pcprops.put("nRotBonds", Double.valueOf(rbcd.calculate(molecule).getValue().toString()));		
		pcprops.put("nBasicGroups", Double.valueOf(bgcd.calculate(molecule).getValue().toString()));

		return pcprops;
	}
	
	public LinkedHashMap<String, Double> generateAllFeatures(IAtomContainer molecule) throws Exception{
		LinkedHashMap<String, Double> allFeatures = new LinkedHashMap<String, Double>();
		
		LinkedHashMap<String, Double> physchem  = this.generatePhysChemProperties(molecule);
		LinkedHashMap<String, Double> p2f = this.csearcher.generateP2FFingerprint(molecule);
		
		
		for(Map.Entry<String, Double> pc: physchem.entrySet()){
			allFeatures.put(pc.getKey(),pc.getValue());
		}
		for(Map.Entry<String, Double> pf: p2f.entrySet()){
			allFeatures.put(pf.getKey(),pf.getValue());
		}		
		
		return allFeatures;
	}
	
	public ArrayList<Double> generateAllFeaturesToArrayList(IAtomContainer molecule) throws Exception{
		ArrayList<Double> allFeatures = new ArrayList<Double>();
		
		LinkedHashMap<String, Double> physchem  = this.generatePhysChemProperties(molecule);
		LinkedHashMap<String, Double> p2f = this.csearcher.generateP2FFingerprint(molecule);
		
		
		for(Map.Entry<String, Double> pc: physchem.entrySet()){
			allFeatures.add(pc.getValue());
		}
		for(Map.Entry<String, Double> pf: p2f.entrySet()){
			allFeatures.add(pf.getValue());
		}		
		
		return allFeatures;
	}
	
	
	public static Double getMonoisotopicMass(IAtomContainer molecule){
		Double  mass = null;
		
		try {
			mass = MolecularFormulaManipulator.getMajorIsotopeMass(MolecularFormulaManipulator.getMolecularFormula(molecule));
		}
		catch (Exception e){
			
			System.out.println(e.getLocalizedMessage());
		}
		
		
		return mass;
	}
}

