package phase2filter.utils;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;

import org.apache.commons.io.FilenameUtils;
import org.openscience.cdk.CDKConstants;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.exception.CDKException;
import org.openscience.cdk.inchi.InChIGenerator;
import org.openscience.cdk.inchi.InChIGeneratorFactory;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.io.SDFWriter;
import org.openscience.cdk.io.iterator.IteratingSDFReader;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;

public class FileUtils {

	public FileUtils() {
		// TODO Auto-generated constructor stub
	}

	public static IAtomContainerSet parseSdf(String sdfFileName) throws FileNotFoundException {
		IAtomContainerSet containers = DefaultChemObjectBuilder.getInstance().newInstance(
				IAtomContainerSet.class);
		
		IChemObjectBuilder bldr = SilentChemObjectBuilder.getInstance();
		
		IteratingSDFReader sdfr = new IteratingSDFReader(new FileReader(sdfFileName), bldr);
		
		while (sdfr.hasNext()){
			IAtomContainer mol = sdfr.next();
			containers.addAtomContainer(mol);	
		}
			
		return containers;		
	}

	public static void saveAtomContainerSetToSDF(IAtomContainerSet containers, String outputFileName) throws CDKException, IOException{
		SDFWriter sdfWriter = new SDFWriter(new FileOutputStream(outputFileName));		
		sdfWriter.write(containers);
		sdfWriter.close();
	}
	
	public static void buildSdfFromTSV(String tsvFileName) throws IOException, CDKException{
		
		BufferedReader bRead = new BufferedReader(new FileReader(tsvFileName));
		SDFWriter sdfWriter = new SDFWriter(new FileOutputStream(FilenameUtils.getFullPathNoEndSeparator(tsvFileName) + 
				"/" + FilenameUtils.getBaseName(tsvFileName)  + ".sdf" ));
		
		int counter = 0;
		String line = null;
		
		IChemObjectBuilder 	builder = SilentChemObjectBuilder.getInstance();
		SmilesGenerator smiGen 		= new SmilesGenerator().isomeric();
		SmilesParser	smiParser		= new SmilesParser(builder);
		InChIGeneratorFactory inchiGenFactory = InChIGeneratorFactory.getInstance();
		
		
		while((line = bRead.readLine()) !=null){
			String[] sline = line.split("\t");
			IAtomContainer atc= smiParser.parseSmiles(sline[2]);
			InChIGenerator gen = inchiGenFactory.getInChIGenerator(atc);
			atc.setProperty(CDKConstants.TITLE, sline[0]);
			atc.setProperty("Name", sline[1]);
			
			if(gen.getInchiKey() == sline[3].trim()){
				atc.setProperty("InChIKey", sline[3]);
			}
			else{
				System.err.println("Issue of inchikey incompatibility with " + sline[0]);
			}
			atc.setProperty("Origin", sline[4]);
			
			sdfWriter.write(atc);

		}
		
		sdfWriter.close();
		bRead.close();
	}

}

