package phase2filter.utils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.openscience.cdk.CDKConstants;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.aromaticity.Aromaticity;
import org.openscience.cdk.aromaticity.ElectronDonation;
import org.openscience.cdk.graph.CycleFinder;
import org.openscience.cdk.graph.Cycles;
import org.openscience.cdk.interfaces.IAtom;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.io.iterator.IteratingSDFReader;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;
import org.openscience.cdk.smiles.smarts.SMARTSQueryTool;
import org.openscience.cdk.smiles.smarts.SmartsPattern;
import org.openscience.cdk.tools.CDKHydrogenAdder;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import ambit2.smarts.query.SMARTSException;
import ambit2.smarts.query.SmartsPatternCDK;
import phase2filter.utils.StructuralClassFinder.StructuralClassName;


/**
 * @author Yannick Djoumbou Feunang; 2017
 * 
 *
 */

public class StructSearcher {

	LinkedHashMap<String, String> queries;
	static LinkedHashMap<String, String> phaseIIsubstratePatterns;
	static LinkedHashMap<String, String> phaseIIMetabolitePatterns;
	IChemObjectBuilder builder = SilentChemObjectBuilder.getInstance();
	
	public StructSearcher() throws Exception {
		// TODO Auto-generated constructor stub
		queries = this.getP2FFingerprintPatterns();
		phaseIIMetabolitePatterns = this.getPhaseIIfingerpint();
		phaseIIsubstratePatterns = this.getPhase2CandidateFingerprintPatterns();
		
	}

	public static void main(String args[]) throws Exception {

		IChemObjectBuilder builder = SilentChemObjectBuilder.getInstance();
		SmilesGenerator sg = new SmilesGenerator();
		SmilesParser sp = new SmilesParser(builder);
		IAtomContainer molecule = sp
				.parseSmiles("COC1=C(C=NC(CC2=CC=C(O)C(O)=C2)=C1NCO)C([O-])=O");
		
		IAtomContainer pmolecule = ChemStructureManipulator.preprocessContainer(molecule);

		StructSearcher csearcher = new StructSearcher();
//		LinkedHashMap<String, String> queries = csearcher.getP2FFingerprintPatterns();
//		
//		SMARTSQueryTool querytool = new SMARTSQueryTool(queries.get("H-bond"), builder);
//		
//		List<List<Integer>> matchingAtoms = csearcher.getMatchingAtoms(molecule, querytool);
		
		LinkedHashMap<String, Double> fp = csearcher.generateP2FFingerprint(pmolecule);
		System.out.println(fp);
		
	}
	public static LinkedHashMap<String, String> getPhase2CandidateFingerprintPatterns() throws Exception {
		LinkedHashMap<String, String> queries = new LinkedHashMap<String, String>();

		queries.put("ALKYL_OH_GLUCURONIDATION" ,  "[H][#8;X2][#6;A;X4]([#6])([#6,#1])[#6,#1]");
		queries.put("AROMATIC_OH_GLUCURONIDATION" ,  "[#8;H1X2]-[#6,#7,#16;a]");
		queries.put("ALIPHATIC_ACYL_O_GLUCURONIDATION" ,  "[#6;A;!$(C(-[NX3H2])-[CX3](=O)[OX2H1,OX1-])][#6;A;X3]([#8;H1X2])=O");
		queries.put("AROMATIC_ACYL_O_GLUCURONIDATION" ,  "[#6;a]-[#6]([#8;A;X2H1,X1-])=O");
		queries.put("HYDROXYLAMINE_N_GLUCURONIDATION_PATTERN1" ,  "[H][#7;A;X3R0;!$(NC=[O,N,S])](!@-[#6])[#8;A;X2R0;$([H1]),$(O[#6;!$(C=[N,O,S])])]");
		queries.put("HYDROXYLAMINE_N_GLUCURONIDATION_PATTERN2" ,  "[#6]!@-[#7;A;H0X3;$([H0]([#6])[#6])!$(NC=[O,N,S])](!@-[#6])[#8;A;X2;$([H1]),$(O[#6;!$(C=[N,O,S])])]");
		queries.put("SULFONAMIDE_N_GLUCURONIDATION" ,  "[#6]-[#16;X4](=[#8;X1])(=[#8;X1])-[#7;X3;H1,H2]");
		queries.put("GLUCURONIDATION_OF_PRIMARY_AROMATIC_AMINE" ,  "[#6;a]-[N;X3;H2]");
		queries.put("PYRIDINE_GLUCURONIDATION" ,  "[#6]-[#6;R1]=,:1[#6]=,:[#6][#6;R1]=,:[#7;X2][#6;R1]=,:1");
		queries.put("THIOL_GLUCURONIDATION" ,  "[#6]-[#16;H1X2]");
		queries.put("THIOPHENOL_GLUCURONIDATION" ,  "[#16;H1X2]-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]=,:1");
		queries.put("_13_DICARBONYL_C_GLUCURONIDATION" ,  "[O;X1]=[#6;X3]-[#6;X4H2]-[#6;X3]=[O;X1]");
		queries.put("ISOFLAVONE_4P_OH_GLUCURONIDATION" ,  "[H][#8;X2]-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1](=,:[#6;R1][#6;R1]=,:1)-[#6;R1]1=,:[#6;R1][#8;R1][#6;R2]2=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R2]2[#6;R1]1=[O;X1]");
		queries.put("ISOFLAVONE_7_OH_GLUCURONIDATION" ,  "[H][#8;X2]-[#6;R1]1=,:[#6;R1][#6;R1]=,:[#6;R2]2[#6;R2](=,:[#6;R1]1)[#8;R1][#6;R1]=,:[#6;R1](-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]=,:1)[#6;R1]2=[O;X1]");
		queries.put("AMIDE_N_GLUCURONIDATION" ,  "[#7;X3;$([H2]),$([H1][#6;!$(C=[O,N,S])])]-[#6](-[#6X4,#7X3])=[O;X1]");
		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN1" ,  "[#6]=,:1[#7;H1v3X3R1][#7;R1]=,:[#6,#7][#6,#7]=,:1");
		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN5" ,  "*!@-[#6]-1=[#7]-[#7]=[#7;X2R1]-[#7;H1v3X3R1]-1");
		queries.put("_3_OH_GLUCURONIDATION_OF_STEROL" ,  "[#8;A;H1X2][#6]~1~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~2~[#6,#8,#7,#16;A;R2](~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R1]~[#6,#7]-,=3-,=[#6,#7]-,=4-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6]-,=[#6,#7]-,=4-,=[#6,#8,#7,#16][#6;A;X4;H2,H1][#6,#7]~2-,=3)~[#6,#8,#7,#16]~1");
		queries.put("_3_OH_GLUCURONIDATION_OF_STEROL_PATTERN2" ,  "[#8;A;H1X2][#6]1=,:[#6][#6]=,:[#6]-2[#6](=,:[#6]1)-[#6,#7]~[#6,#8,#7,#16;A;R1]~[#6,#7]~1-,=[#6,#7]-,=3-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6]-,=[#6,#7]-,=3-,=[#6,#8,#7,#16]-,=[#6,#7,#8,#16]~[#6,#7]-2~1");
		queries.put("_6_OH_GLUCURONIDATION_OF_STEROL" ,  "[#8;A;H1X2][#6]~1~[#6,#8,#7,#16;A;R1]~[#6,#7]~2-,=[#6,#7]-,=3-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6]-,=[#6,#7]-,=3-,=[#6,#7][#6;A;X4;H2,H1][#6,#7]~2~[#6,#7]~2~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#7]~1~2");
		queries.put("_6_OH_GLUCURONIDATION_OF_STEROL_PATTERN2" ,  "[#8;A;H1X2][#6]-1-,=[#6,#8,#7,#16;A;R1]-,=[#6,#7]-,=2-,=[#6,#7]-,=3-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6]-,=[#6,#7]-,=3-,=[#6,#7][#6;A;X4;H2,H1][#6,#7]-,=2-[#6]2=,:[#6][#6]=,:[#6][#6]=,:[#6]-12");
		queries.put("_7_OH_GLUCURONIDATION_OF_STEROL" ,  "[H][#8;A;X2]C1([H])[#6,#8,#7,#16;A;R1]~[#6]~2~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~2~[#6,#7]~2~[#6;A;X4;H2,H1]~[#6,#8,#7,#16]-,=[#6,#7]-,=3-,=[#6]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#7]-,=3-,=[#6,#7]1~2");
		queries.put("_7_OH_GLUCURONIDATION_OF_STEROL_PATTERN2" ,  "[H][#8;A;X2]C1([H])[#6,#8,#7,#16;A;R1][#6]2=,:[#6][#6]=,:[#6][#6]=,:[#6]2-[#6,#7]~2-,=[#6;A;X4;H2,H1]-,=[#6,#8,#7,#16]-,=[#6,#7]-,=3-,=[#6]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#7]-,=3-,=[#6,#7]1~2");
		queries.put("_17_OH_GLUCURONIDATION_OF_STEROL" ,  "[#8;A;H1X2][#6;A;X4]1([#1,#6;R0])[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#7]-,=2-,=[#6,#7]~3~[#6,#8,#7,#16;A;R1]~[#6,#8,#7,#16]~[#6,#7]~4~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#7]~4~[#6,#7]~3[#6;A;X4;H2,H1][#6,#8,#7,#16]-,=[#6,#7]1-,=2");
		queries.put("_17_OH_GLUCURONIDATION_OF_STEROL_PATTERN2" ,  "[H][#8;A;X2][#6;A;X4]1([#1,#6;R0])[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#7]-,=2-,=[#6,#7]~3-,=[#6,#8,#7,#16;A;R1]-,=[#6,#8,#7,#16]-[#6]4=,:[#6][#6]=,:[#6][#6]=,:[#6]4-[#6,#7]~3[#6;A;X4;H2,H1][#6,#8,#7,#16]-,=[#6,#7]1-,=2");
		queries.put("GLUCURONIDATION_OF_AMINE_OXIDE" ,  "[$([#6][N;X4+]([#6])([#6])[#8;X1-]),$([#6][#7;X3+](=,:[#6])-[#8;X1-])]");
		queries.put("TERTIARY_ALIPHATIC_AMINE_N_GLUCURONIDATION_PATTERN1" ,  "[#6;A][#7;A;H0X3R0;+0!$([N][#6]=[!#6;!#1])!$([N]#[!#6,!#1])]([#6;A])[#6;A]");
		queries.put("TERTIARY_ALIPHATIC_AMINE_N_GLUCURONIDATION_PATTERN2" ,  "[H][#7;A;H1X4+1;!$([N][#6]=[!#6;!#1])!$([N]#[!#6,!#1])]([#6;A])([#6;A])[#6;A]");
		queries.put("_3_OH_GLUCURONIDATION_OF_PHENOLIC_COMPOUND" ,  "[#6]!@-[#6;R1]1=,:[#6;R1][#6;R1](!@-[#8;A;X1-,X2H1])=,:[#6;R1][#6;H1X3R1]=,:[#6;H1X3R1]1");
		queries.put("C24_BILE_ACID_C24_COOH_GLUCURONIDATION" ,  "[#6;A;H3X4][#6;A;H1X4]([#6;A;H2X4][#6;A;H2X4][#6]([#8;A;X1-,X2H1])=O)[#6]1-[#6]-,=[#6]-[#6]2-[#6]-,=3-[#6;CX4H2,$([CX4H1]-[OX2H1])]-[#6]-[#6]4-,=[#6]-,=[#6]([#8;A;X2H1,X1-])-[#6]-[#6]C4([#6;A;H3X4])[#6]-,=3-[#6]-[#6;CX4H2,$([CX4H1]-[OX2H1])]C12[#6;A;H3X4]");
		queries.put("_3_O_SULFATION_OF_PHENOLIC_COMPOUND" ,  "[H][#6;R1]1=,:[#6;R1][#6;R1]([#8;A;H1X2])=,:[#6;R1][#6;R1](!@-[*,#1;$([#6;R1]1=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]([OX2H1,OX1-])=,:[#6;R1]1),$([#6;R1]1=,:[#6;R1]([OX2H1,OX1-])[#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R1]1),$([#6;R1]1=,:[#6;R1][#6;R1]=,:[#6;R1]([OX2H1,OX1-])[#6;R1]=,:[#6;R1]1),$([NX3H0+0,NX4H1+;$([N]([#6])([#6])[#6]);!$([N]*~[#7,#8,#15,#16])]),$([NX3H1+0,NX4H2+;$([N]([#6])[#6]);!$([N]*~[#7,#8,#15,#16])]),$([NX3H2+0,NX4H3+;!$([N][!C]);!$([N]*~[#7,#8,#15,#16])]),$([NX3H2+0,NX4H3+]c),$([OX2](c)[CX4;!$(C([OX2])[O,S,#7,#15,F,Cl,Br,I])]),$([OX2]([CX4;!$(C([OX2])[O,S,#7,#15,F,Cl,Br,I])])[CX4;!$(C([OX2])[O,S,#7,#15,F,Cl,Br,I])]),$([#7X3;$([H2]),$([H1][#6;!$(C=[O,N,S])]),$([#7]([#6;!$(C=[O,N,S])])[#6;!$(C=[O,N,S])])]-[CX3;$([R0][#6]),$([H1R0])]=[OX1]),$([OX2][CX3;$([R0][#6]),$([H1R0])](=[OX1])[#6;!$(C=[O,N,S])]),$([CX4]),$([CX3]),$([#6;R1]1:,=[#6;R1]:,-[#6;R1]:,=[#6;R1]:,-[#6;R1]:,=[#6;R1]1),$([#6;X3](=,:[#6;X3])-,:[#1,*]),$([#6;A;X3](=[#6;X3])-[#1,*]),$(C([Br])([Br])[Br]),$(C([Cl])([Cl])[Cl]),$(C([F])([F])[F]),$(C([I])([I])[I]),$(C#[NX1]),$([NX1+]),$([#16;A](=O)(=O)[OX2H1,OX1-]),$([NX4H3+1]),$([NX4H0+;$([N]([#6])([#6])[#6][#6]);!$([N]*~[#7,#8,#15,#16])]),$([$([NX3](=O)=O),$([NX3+](=O)[O-])][!#8]),$([$([CX3H][#6]),$([CX3H2])]=[OX1]),$([#6][CX3](=[OX1])[#6]),$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[$([OX2H]),$([OX1-])]),$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[ClX1]),$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[OX2][#6;!$(C=[O,N,S])]),$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[#7X3;$([H2]),$([H1][#6;!$(C=[O,N,S])]),$([#7]([#6;!$(C=[O,N,S])])[#6;!$(C=[O,N,S])])])])=,:[#6;R1]1[H]");
		queries.put("_4_O_SULFATION_OF_PHENOLIC_COMPOUND" ,  "[H][#6;R1]1=,:[#6;R1][#6;R1](!@-[#8;H1X2R0])=,:[#6;R1][#6;R1]([H])=,:[#6;R1]1!@-[*,#1;$([#6;R1]1=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]([OX2H1,OX1-])=,:[#6;R1]1),$([#6;R1]1=,:[#6;R1]([OX2H1,OX1-])[#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R1]1),$([#6;R1]1=,:[#6;R1][#6;R1]=,:[#6;R1]([OX2H1,OX1-])[#6;R1]=,:[#6;R1]1),$([NX3H0+0,NX4H1+;$([N]([#6])([#6])[#6]);!$([N]*~[#7,#8,#15,#16])]),$([NX3H1+0,NX4H2+;$([N]([#6])[#6]);!$([N]*~[#7,#8,#15,#16])]),$([NX3H2+0,NX4H3+;!$([N][!C]);!$([N]*~[#7,#8,#15,#16])]),$([NX3H2+0,NX4H3+]c),$([OX2](c)[CX4;!$(C([OX2])[O,S,#7,#15,F,Cl,Br,I])]),$([OX2]([CX4;!$(C([OX2])[O,S,#7,#15,F,Cl,Br,I])])[CX4;!$(C([OX2])[O,S,#7,#15,F,Cl,Br,I])]),$([#7X3;$([H2]),$([H1][#6;!$(C=[O,N,S])]),$([#7]([#6;!$(C=[O,N,S])])[#6;!$(C=[O,N,S])])]-[CX3;$([R0][#6]),$([H1R0])]=[OX1]),$([OX2][CX3;$([R0][#6]),$([H1R0])](=[OX1])[#6;!$(C=[O,N,S])]),$([CX4]),$([CX3]),$([#6;R1]1:,=[#6;R1]:,-[#6;R1]:,=[#6;R1]:,-[#6;R1]:,=[#6;R1]1),$([#6;X3](=,:[#6;X3])-,:[#1,*]),$([#6;A;X3](=[#6;X3])-[#1,*]),$(C([Br])([Br])[Br]),$(C([Cl])([Cl])[Cl]),$(C([F])([F])[F]),$(C([I])([I])[I]),$(C#[NX1]),$([NX1+]),$([#16;A](=O)(=O)[OX2H1,OX1-]),$([NX4H3+1]),$([NX4H0+;$([N]([#6])([#6])[#6][#6]);!$([N]*~[#7,#8,#15,#16])]),$([$([NX3](=O)=O),$([NX3+](=O)[O-])][!#8]),$([$([CX3H][#6]),$([CX3H2])]=[OX1]),$([#6][CX3](=[OX1])[#6]),$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[$([OX2H]),$([OX1-])]),$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[ClX1]),$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[OX2][#6;!$(C=[O,N,S])]),$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[#7X3;$([H2]),$([H1][#6;!$(C=[O,N,S])]),$([#7]([#6;!$(C=[O,N,S])])[#6;!$(C=[O,N,S])])])]");
		queries.put("_4_O_SULFATION_OF_HYDROXYCINNAMATE_PATTERN1" ,  "[#8;A;H1X2][#6;R1]1=,:[#6;R1][#6;R1]=,:[#6;R1]([#6;R1]=,:[#6;R1]1-[#1,$([O][H]),$([O][C](-[H])(-[H])[H])])[#6;A;H1X3]=[#6;A;H1X3][#6]([#8;A;X1-,X2H1])=[O;X1]");
		queries.put("_4_O_SULFATION_OF_HYDROXYCINNAMATE_PATTERN2" ,  "[#8;A;H1X2][#6;R1]1=,:[#6;R1][#6;R1]=,:[#6;R1]([#6;R1]=,:[#6;R1]1-[#1,$([O][H]),$([O][C](-[H])(-[H])[H])])[#6;A;H2X4][#6;A;H2X4][#6]([#8;A;X1-,X2H1])=[O;X1]");
		queries.put("PHENOLIC_STEROID_SULFONATION" ,  "[H][#8]-[#6]=,:1[#6]=,:[#6][#6]~2=,:[#6]([#6]=,:1)~[#6,#7]~[#6,#7]~[#6,#7]~1-,=[#6,#7]-,=3-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A]-,=[#6]-,=3-,=[#6]-,=[#6]~[#6,#7]~2~1");
		queries.put("_7_OH_HYDROXYLATION_OF_FLAVONOL" ,  "[#8;A;X1-,X2H1][#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R2]2=,:[#6;R2]([#6;R1]=,:1)[#8][#6](-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]=,:1)=,:[#6]([#8;A;X1-,X2H1])[#6]2=O");
		queries.put("SULFATION_OF_PRIMARY_ALCOHOL" ,  "[#6;A;H2X4;!$(C([OX2H])[O,S,#7,#15])][#8;H1X2]");
		queries.put("SULFATION_OF_SECONDARY_ALCOHOL_PATTERN1" ,  "[H][#8][#6;A;H1X4;!$(C([OX2H])[O,S,#7,#15])]([H])([#6])[#6]");
		queries.put("STEROL_3_OH_SULFATION" ,  "[H][#8]-[#6]~1~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~2~[#6,#8,#7,#16;A;R2](~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R1]~[#6,#8,#7,#16;A;R2]~3-,=[#6,#8,#7,#16;A;R2]-,=4-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A]-,=[#6,#8,#7,#16]-,=4-,=[#6,#8,#7,#16]~[#6,#8,#7,#16;A]~[#6,#8,#7,#16]~2~3)~[#6,#8,#7,#16]~1");
		queries.put("STEROL_7_OH_SULFATION" ,  "[H][#8]C1([H])[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R2]~2~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~2~[#6,#8,#7,#16]~2~[#6,#8,#7,#16;A]~[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=3-,=[#6,#8,#7,#16;A]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A;R2]-,=3-,=[#6]1~2");
		queries.put("STEROL_7_OH_SULFATION_PATTERN2" ,  "[H][#8]C1([H])[#6,#7]~[#6]=,:2[#6]=,:[#6][#6]=,:[#6][#6]=,:2~[#6,#7]~2~[#6,#8,#7,#16;A]~[#6,#8,#7,#16]-,=[#6,#7]-,=3-,=[#6,#8,#7,#16;A]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#7]-,=3-,=[#6,#7]1~2");
		queries.put("STEROL_12_OH_SULFATION" ,  "[H][#8]C1([H])[#6]~[#6,#8,#7,#16]~2~[#6,#8,#7,#16]~3~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R2]~3~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R1]~[#6,#8,#7,#16;A;R2]~2-,=[#6,#8,#7,#16;A;R2]-,=2-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A]-,=[#6]1-,=2");
		queries.put("STEROL_12_OH_SULFATION_PATTERN2" ,  "[H][#8]C1([H])[#6]~[#6,#7]~2~[#6]=,:3[#6]=,:[#6][#6]=,:[#6][#6]=,:3~[#6,#7]~[#6,#7]~[#6,#7]~2-,=[#6,#7]-,=2-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A]-,=[#6]1-,=2");
		queries.put("GALACTOSYLCERAMIDE_3P_SULFATION" ,  "[H][#8]C([H])([#6]([H])-,=[#6](-[*,#1;#1,OX2H1])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])[#1,#6;A])C([H])([#7]([H])-[#6]([#6,#1;A])=O)C([H])([H])[#8][#6;A@@H;H1X4]1[#8][#6;A@@H;H1X4]([#6;A@H;H1X4]([#8;A;H1X2])[#6;A;H1X4]([#8;A;H1X2])[#6;A;H1X4]1[#8;A;H1X2])C([H])([H])[#8;A;H1X2]");
		queries.put("DIGALACTOSYLCERAMIDE_3P_SULFATION" ,  "[H][#8][#6;A@@H;H1X4]1[#6;A;H1X4]([#8])[#6;A;H1X4]([#8])[C@@]([H])([#8][#6;A@@;H1X4]2([H])[#6;A@H;H1X4]([#8]-[#6@@H](-[#8]C([H])([H])C([H])([#7]([H])-[#6](-[#6,#1])=O)C([H])([#8][H])[#6]([H])-,=[#6](-[*,#1;#1,OX2H1])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])[#1,#6;A])[#6;A@;H1X4]([H])([#8][H])[#6;A;H1X4]2[#8][H])C([H])([H])[#8][H])[#8][#6;A@@H;H1X4]1C([H])([H])[#8][H]");
		queries.put("GALACTOSYL_1_ALKYL_2_ACYL_SULFATION" ,  "[H][#8][#6;A;H1X4]1[#6;A@H;H1X4]([#8]C([H])([H])C([H])([#8]-[#6]([#6,#1;A])=O)C([H])([H])[#8][#6;A;X4]([H])([H])[#6])[#8][#6;A@@H;H1X4]([#6;A@H;H1X4]([#8][H])[#6;A;H1X4]1[#8][H])C([H])([H])[#8][H]");
		queries.put("ARYLAMINE_N_ACETYLATION_PATTERN1" ,  "[#7;A;H2X3]!@-[#6]=,:1[#6;R1]=,:[#6][#6]=,:[#6][#6;R1]=,:1");
		queries.put("ARYLAMINE_N_ACETYLATION_PATTERN2" ,  "[H][#7;A;X3+0,X4+;!$([N]~[!#6])][#7;A;X3]([H])[#6;a]");
		queries.put("ARYLAMINE_N_ACETYLATION_PATTERN3" ,  "[#7;A;H2X3]!@-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1]=,:[#6][#6]=,:1");
		queries.put("L_CYSTEINE_S_CONJUGATE_N_ACETYLATION" ,  "[#6]-[#16;X2][#6;A;H2X4][#6;A@H;X4]([#7;A;H2X3])[#6](-[#8-])=O");
		queries.put("ARALKYLAMINE_N_ACETYLATION" ,  "[H][#7;A;X3;+0!$([N]~[!#6])!$([N]*~[#7,#8,#15,#16])][#6;A;H2X4]!@-[#6;A;H2X4]!@-[#6;a]");
		queries.put("O_ACETYLATION_OF_AROMATIC_HYDROXYLAMINE" ,  "[#6;a]!@-[#7;A;X3][#8;A;X2H1,X1-]");
		queries.put("N_HYDROXYL_N_ARYLACETAMIDE_O_ACETYLATION" ,  "[H][#8]-[#7;X3](-[#6;X3]([#6;A;H3X4])=[O;X1])!@-[#6;X3]-1=[#6]-[#6]=[#6]-[#6]=[#6]-1");
		queries.put("N_ACETYLATION_OF_PRIMARY_ARYLAMINE" ,  "[#6;a]!@-[#7;A;H2X3]");
//		queries.put("DIACYLGLYCEROL_O_ACYLATION" , "[$([#8]-,=[#6;R0](-,=[#7;R0]-[#6;R0]-[#6;R0]-[#16;X2R0]-[#6;R0](-[#6;H2X4R0])=[O;R0])-[#6;R0]-[#6;R0]-[#7;R0]-,=[#6;R0](-,=[#8;R0])[#6;A;H1X4]([#8])[C;R0]([#6;R0])([#6;R0])[#6;H2X4R0]-[#8;R0][P;R0]([!#1!#6;O,$([O-])])(=[O;R0])[#8;R0][P;R0]([!#1!#6;O,$([O-])])(=[O;R0])[#8]-[#6]-[#6]-1-[#8]-[#6](-[#6](-[#8])-[#6]-1-[#8]P([!#1!#6;O,$([O-])])([!#1!#6;O,$([O-])])=O)-n1cnc2c(-[#7])ncnc12), $([H][#8][#6;A;H2X4][#6;A;H1X4]([#6;A;H2X4][#8]-[#6]([#6,#1;A])=O)[#8]-[#6]([#6,#1;A])=O)]";)
		queries.put("ARALKYLAMINE_N_ACETYLATION_PATTERN2" ,  "[H][#7;A;X3]([H])[#6;A;H2X4]!@-[#6;A;H2X4]!@-[#6;a]");
		queries.put("PEPTIDE_ALPHA_N_ACETYLATION" ,  "[#7;A;X3;H2,H1][#6](-[$([#1,*])])-[#6](=O)-[#7]-[#6](-[$([#1,*])])-[#6](-[#8])=O");
		queries.put("N_O_ACYLTRANSFER_OF_ARYLHYROXAMIC_ACID" ,  "[#6;A;H3X4][#6;X3](=[O;X1])-[#7;X3]([#8;A;X1-,X2H1])-[#1,$([#6;$(C=[O,N,S])])]");
		queries.put("CATECHOL_O_METHYLATION_PATTERN1" ,  "[H][#8;A;X2][#6;R1]1=,:[#6;R1](-[*,#1;!$([OX2H1])])[#6;R1](-[*,#1])=,:[#6;R1](-[*,#1])[#6;R1](-[*,#1;!$([OX2H1])])=,:[#6;R1]1[#8;A;X2][H]");
		queries.put("CATECHOL_O_METHYLATION_PATTERN2" ,  "[H][#8;A;X2][#6;R1]1=,:[#6;R1]([H])[#6;R1]([H])=,:[#6;R1](-[!#8])[#6;R1]([H])=,:[#6;R1]1[#8;A;X2][H]");
		queries.put("CATECHOL_O_METHYLATION_PATTERN3" ,  "[H][#8;A;X2][#6;R1]=,:1[#6;R1]([H])=,:[#6;R1]([H])[#6;R1](-[#6;R0])=,:[#6;R1]([H])[#6;R1]=,:1[#8;A;X2][H]");
		queries.put("PHOSPHATIDYLETHANOLAMINE_N_METHYLATION" ,  "[#7;H2X3R0]-[#6;H2X4R0]-[#6;H2X4R0]-[#8;R0]P([!#1!#6;O,$([O-])])(=O)[#8;R0]-[#6;H2X4R0][#6;A;H1X4]([#6;H2X4R0]-[#8;R0]-[#6;R0]([#6,#1;A])=[O;R0])[#8;R0]-[#6;R0]([#6,#1;A])=[O;R0]");
		queries.put("FLAVONE_3P_O_METHYLATION" ,  "[#8;A;H1X2][c;R1]1[c;R1][c;R1][c;R1][c;R1]([c;R1]1)-[#6;R1]1=,:[#6;R1][#6;R1](=[O;X1])c2ccccc2[#8]1");
		queries.put("THIOPURINE_S_METHYLATION" ,  "[#16;A;H1X2][c;R1]1[n;R1][c;R1][n;R1][c;R2]2[n;R1][c;R1][n;R1][c;R2]12");
//		queries.put("PHOSPHATIDYL_N_MONOMETHYLETHANOLAMINE_N_METHYLATION" ,  "[H][#7;A;H1X3]([#6;A;H3X4])[#6;A;H2X4R0][#6;A;H2X4R0][#8]P([#8;A;X1-,X2H1])(=O)[#8][#6;A;H2X4][#6;A;H1X4]([#6;A;H2X4][#8]-[#6]([#6,#1;A])=O)[#8]-[#6]([#6,#1;A])=O");
//		queries.put("PHOSPHATIDYL_N_DIMETHYLETHANOLAMINE_N_METHYLATION" ,  "[#6;A;H3X4][#7;A;X3]([#6;A;H3X4])[#6;A;H2X4R0][#6;A;H2X4R0][#8]P([#8;A;X1-,X2H1])(=O)[#8][#6;A;H2X4][#6;A;H1X4]([#6;A;H2X4][#8]-[#6]([#6,#1;A])=O)[#8]-[#6]([#6,#1;A])=O");
		queries.put("THIOETHER_S_METHYLATION" ,  "[#6]-[#16;X2+0;R0]-[#6]");
		queries.put("SPHINGOLIPID_C9_METHYLATION" ,  "[#6;A;H3X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H1X3]=[#6;A;H1X3][#6;A;H2X4][#6;A;H2X4][#6;A;H1X3]=[#6;A;H1X3][#6;A;H1X4]([#8;A;H1X2])[#6;A;H1X4]([#6;A;H2X4][#8;A;H1X2])[#7;A;H1X3][#6](-[#6])=[O;X1]");
		queries.put("N_METHYLATION_OF_AZAHETEROCYCLE_PATTERN1" ,  "[#6]~[#7;a;X2R1;r5,r6]~[#6]");
		queries.put("N_METHYLATION_OF_AZAHETEROCYCLE_PATTERN2" ,  "[#6]@-[#7;A;H1X3R1;r5,r6]@-[#6]");
		queries.put("METHYLATION_OF_AROMATIC_CARBOXYLIC_ACID" ,  "[#6;a]-[#6;X3]([#8;A;H1X2])=[O;X1]");
		queries.put("THIOL_S_METHYLATION" ,  "[#6]-[#16;R0;X2H1+0,X1-1]");
		queries.put("GSH_CONJUGATION_OF_EPOXIDE" ,  "[#6]-[#6;X4]-1-[#8]-[#6;X4]-1-[#1,#6,#7,#8]");
		queries.put("GSH_CONJUGATION_OF_SULFHYDRYL" ,  "[#6][S;X2H1]");
		queries.put("GSH_CONJUGATION_OF_ARENE_EPOXIDE" ,  "[#8;X2]-1-[#6;X3]-2=[#6;X3]-1-[#6]=[#6]-[#6]=[#6]-2");
		queries.put("GSH_CONJUGATION_OF_ORGANOHALIDE_PATTERN1" ,  "[H][C;X4]([#1,#6])([F,Cl,Br,I])[*,#1]");
		queries.put("GSH_CONJUGATION_OF_ORGANOHALIDE_PATTERN2" ,  "[H][#6;X3](-[F,Cl,Br,I])=[#6;X3]([*,#1])-[*,#1]");
		queries.put("GSH_CONJUGATION_OF_ACYL_CHLORIDE" ,  "[#6]-[#6;X3](Cl)=[O;X1]");
		queries.put("GSH_CONJUGATION_OF_CARBOXIMIDIC_ACID" ,  "[#6]-[#7]=[#6]-[H1X2;O,S]");
		queries.put("GSH_CONJUGATION_OF_ALPHA_BETA_UNSATURATED_CARBONYL_PATTERN1" ,  "[H][#6;X3R1]-1=[#6;X3R1]-[#6;R1](=[!#1!#6;R0;OX1,NX2])-[#6;X3]=[#6;X3]-[#6;R1]-1=[O;X1]");
		queries.put("GSH_CONJUGATION_OF_ALPHA_BETA_UNSATURATED_CARBONYL_PATTERN2" ,  "[H][#6;X3](-*)!@=[#6;X3](!@[$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[OX2][#6;!$(C=[O,N,S])]),$([$([CX3H][#6]),$([CX3H2])]=[OX1]),$([#6][CX3](=[OX1])[#6])])-[*,#1]");
		queries.put("GSH_CONJUGATION_OF_ISOCYANATE" ,  "[#6]-[#7;X2]=[C;X2]=[O;X1]");
		queries.put("GSH_CONJUGATION_OF_ISOTHIOCYANATE" ,  "[#6]-[#7;X2]=[C;X2]=[S;X1]");
		queries.put("GSH_CONJUGATION_OF_ARYL_HALIDE_PATTERN1" ,  "[!#1!#6;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]~[#6]1=,:[#6][#6]=,:[#6](-[F,Cl,Br,I])[#6](~[#1,!#6;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6]1");
		queries.put("GSH_CONJUGATION_OF_ARYL_HALIDE_PATTERN2" ,  "[#1,!#6;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]-[#6]=,:1[#6]=,:[#6][#6](-[F,Cl,Br,I])=,:[#6]([!#1!#6;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6]=,:1");
		queries.put("GSH_CONJUGATION_OF_ARYL_HALIDE_PATTERN3" ,  "[$([H][#8]-[#6]=,:1[#6]=,:[#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6]=,:[#6][#6]=,:1-[F,Cl,Br,I]),$([H][#8]-[#6]=,:1[#6]=,:[#6][#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6][#6]=,:1-[F,Cl,Br,I]),$([H][#8]-[#6]1=,:[#6][#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6][#6](-[F,Cl,Br,I])=,:[#6]1),$([H][#8]-[#6]1=,:[#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6]=,:[#6][#6]=,:[#6]1-[F,Cl,Br,I]),$([H][#8]-[#6]=,:1[#6]=,:[#6][#6](-[F,Cl,Br,I])=,:[#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6]=,:1),$([H][#8]-[#6]1=,:[#6][#6]=,:[#6][#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6]1-[F,Cl,Br,I])]");
		queries.put("GSH_CONJUGATION_OF_QUINONE_PATTERN1" ,  "[H][#6;R1]-1=[#6]-[#6]=[#6;R1]-[#6;R1](=[O;X1R0])-[#6;R1]-1=[O;X1R0]");
		queries.put("GSH_CONJUGATION_OF_QUINONE_PATTERN2" ,  "[H][#6]-1=[#6]-[#6;R1](=[O;X1R0])-[#6;R1]=[#6;R1]-[#6;R1]-1=[O;X1R0]");
		queries.put("GSH_CONJUGATION_OF_HALO_S_TRIAZINE" ,  "[F,Cl,Br,I]-[#6;R1]=,:1[#7;R1]=,:[#6;R1][#7;R1]=,:[#6;R1][#7;R1]=,:1");
		queries.put("GSH_CONJUGATION_OF_EWG_SUBSTITUTED_BENZENE_PATTERN1" ,  "[H][#6;R1]1=,:[#6;R1][#6;R1](-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6;R1](-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6;R1]=,:[#6;R1]1-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]");
		queries.put("GSH_CONJUGATION_OF_EWG_SUBSTITUTED_BENZENE_PATTERN2" ,  "[H][#6;R1]=,:1[#6;R1](-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6;R1][#6;R1](-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6;R1][#6;R1]=,:1-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]");
		queries.put("GSH_CONJUGATION_OF_ALKYLARYL_PHOSPHATE_OR_PHOSPHOROTHIOATE" ,  "[#1,#6]-[#8;X2]!@-P(=[#8,#16])(!@-[#8]-[CX4H3,$([CX4H2]-[CX4H3])])!@-[#8]-[#6;R1]1=,:[#6,#7;R1][#6;R1](-[F,Cl,Br,I,#1])=,:[$([#7]),$([#6]-[H,F,Cl,Br,I])][#6;R1]=,:[#6;R1]1-[F,Cl,Br,I,#1]");
		queries.put("GSH_CONJUGATION_OF_4_HYDROXYAZAPHOSPHORINE" ,  "[#7;X3R0;NX3H1+0,NX3H2+0,$([NX3+0;!$([N]~[!#6]);!$([N]*~[#7,#8,#15,#16])])][P;X4]1(=[O;X1])[#7;NX3H1,$([NX3]-[#6])][#6;A;H1X4R1]([#8;A;X2H1,X1-])[#6;R1]-[#6;R1]-[#8]1");
		queries.put("GLUTATHIONE_MEDIATED_NITRITE_RELEASE" ,  "[#6,#7][#6](=,:[#6,#7])!@-[#7;X3+](-[#8;X1-])=[O;X1]");
		queries.put("GLYCINATION_OF_ALIPHATIC_ACID" ,  "[H][#8;X2]-[#6;X3](-[#6;$([#7;A;R0]-,=[#6;A;R0]-,=[#6;A;R0]-,=[#6;A;R0]),$([#7]-[#6;A;X3]=[#6;A]-[#6]=,:1[#6]=,:[#6][#6]=,:[#6][#6]=,:1)])=[O;X1]");
		queries.put("GLYCINATION_OF_ARYL_ACID" ,  "[#8;A;X2H1,X1-][#6;X3](=O)-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]=,:1");
		queries.put("C24_BILE_ACID_COA_GLYCINATION_PATTERN1" ,  "[#6;A;H3X4][#6;A;H1X4]([#6;A;H2X4][#6;A;H2X4][#6](=O)-[#16;X2]-[#6]-[#6]-[#7]-[#6](=O)-[#6]-[#6]-[#7]-[#6](=O)-[#6@H]([#8;A;H1X2])C([#6])([#6])[#6]-[#8]P([#8;A;X1-,X2H1])(=O)[#8]P([#8;A;X1-,X2H1])(=O)[#8]-[#6]-[#6@H]-1-[#8]-[#6@@H](-[#7]2[#6]=,:[#7][#6]=,:3[#6]2=,:[#7][#6]=,:[#7][#6]=,:3[#7;A;H2X3])-[#6@H]([#8;A;H1X2])-[#6@@H]-1-[#8]P([#8])([#8;A;X1-,X2H1])=O)[#6]1-[#6]-,=[#6]-[#6]2-[#6]-,=3-[#6;CX4H2,$([CX4H1]-[OX2H1])]-[#6]-,=[#6]4-,=[#6]-,=[#6]([#8;A;X2H1,X1-])-[#6]-[#6]C4([#6;A;H3X4])[#6]-,=3-[#6]-[#6;CX4H2,$([CX4H1]-[OX2H1])]C12[#6;A;H3X4]");
		queries.put("C23_BILE_ACID_COA_GLYCINATION_PATTERN2" ,  "[#6;A;H3X4][#6;A;H1X4]([#6;A;H2X4][#6](=O)-[#16;X2]-[#6]-[#6]-[#7]-[#6](=O)-[#6]-[#6]-[#7]-[#6](=O)-[#6@H]([#8;A;H1X2])C([#6])([#6])[#6]-[#8]P([#8;A;X1-,X2H1])(=O)[#8]P([#8;A;X1-,X2H1])(=O)[#8]-[#6]-[#6@H]-1-[#8]-[#6@@H](-[#7]2[#6]=,:[#7][#6]=,:3[#6]2=,:[#7][#6]=,:[#7][#6]=,:3[#7;A;H2X3])-[#6@H]([#8;A;H1X2])-[#6@@H]-1-[#8]P([#8])([#8;A;X1-,X2H1])=O)[#6]1-[#6]-,=[#6]-[#6]2-[#6]-,=3-[#6;CX4H2,$([CX4H1]-[OX2H1])]-[#6]-,=[#6]4-,=[#6]-,=[#6]([#8;A;X2H1,X1-])-[#6]-[#6]C4([#6;A;H3X4])[#6]-,=3-[#6]-[#6;CX4H2,$([CX4H1]-[OX2H1])]C12[#6;A;H3X4]");
		queries.put("C24_BILE_ACID_COA_TAURINATION_PATTERN1" ,  "[#6;A;H3X4][#6;A;H1X4]([#6;A;H2X4][#6;A;H2X4][#6](=O)-[#16;X2]-[#6]-[#6]-[#7]-[#6](=O)-[#6]-[#6]-[#7]-[#6](=O)-[#6@H]([#8;A;H1X2])C([#6])([#6])[#6]-[#8]P([#8;A;X1-,X2H1])(=O)[#8]P([#8;A;X1-,X2H1])(=O)[#8]-[#6]-[#6@H]-1-[#8]-[#6@@H](-[#7]2[#6]=,:[#7][#6]=,:3[#6]2=,:[#7][#6]=,:[#7][#6]=,:3[#7;A;H2X3])-[#6@H]([#8;A;H1X2])-[#6@@H]-1-[#8]P([#8])([#8;A;X1-,X2H1])=O)[#6]1-[#6]-,=[#6]-[#6]2-[#6]-,=3-[#6;CX4H2,$([CX4H1]-[OX2H1])]-[#6]-,=[#6]4-,=[#6]-,=[#6]([#8;A;X2H1,X1-])-[#6]-[#6]C4([#6;A;H3X4])[#6]-,=3-[#6]-[#6;CX4H2,$([CX4H1]-[OX2H1])]C12[#6;A;H3X4]");
		queries.put("C23_BILE_ACID_COA_TAURINATION_PATTERN2" ,  "[#6;A;H3X4][#6;A;H1X4]([#6;A;H2X4][#6](=O)-[#16;X2]-[#6]-[#6]-[#7]-[#6](=O)-[#6]-[#6]-[#7]-[#6](=O)-[#6@H]([#8;A;H1X2])C([#6])([#6])[#6]-[#8]P([#8;A;X1-,X2H1])(=O)[#8]P([#8;A;X1-,X2H1])(=O)[#8]-[#6]-[#6@H]-1-[#8]-[#6@@H](-[#7]2[#6]=,:[#7][#6]=,:3[#6]2=,:[#7][#6]=,:[#7][#6]=,:3[#7;A;H2X3])-[#6@H]([#8;A;H1X2])-[#6@@H]-1-[#8]P([#8])([#8;A;X1-,X2H1])=O)[#6]1-[#6]-,=[#6]-[#6]2-[#6]-,=3-[#6;CX4H2,$([CX4H1]-[OX2H1])]-[#6]-,=[#6]4-,=[#6]-,=[#6]([#8;A;X2H1,X1-])-[#6]-[#6]C4([#6;A;H3X4])[#6]-,=3-[#6]-[#6;CX4H2,$([CX4H1]-[OX2H1])]C12[#6;A;H3X4]");
		
		
		
//		queries.put("ALKYL_OH_GLUCURONIDATION" , "[H][#8;X2][#6;A;X4]([#6])([#6,#1])[#6,#1]");
//		queries.put("AROMATIC_OH_GLUCURONIDATION" , "[#8;H1X2]-[#6,#7,#16;a]");
//		queries.put("ALIPHATIC_ACYL_O_GLUCURONIDATION" , "[#6;A;!$(C(-[NX3H2])-[CX3](=O)[OX2H1,OX1-])][#6;A;X3]([#8;H1X2])=O");
//		queries.put("AROMATIC_ACYL_O_GLUCURONIDATION" , "[#6;a]-[#6]([#8;A;X2H1,X1-])=O");
//		queries.put("HYDROXYLAMINE_N_GLUCURONIDATION_PATTERN1" , "[NX3H1;!$(NC=[O,N,S])](-[#6])[OX2;$([H1]),$(O[#6;!$(C=[N,O,S])])]");
//		queries.put("HYDROXYLAMINE_N_GLUCURONIDATION_PATTERN2" , "[#6][NX3;$([H0]([#6])[#6]);!$(NC=[O,N,S])](-[#6])[OX2;$([H1]),$(O[#6;!$(C=[N,O,S])])]");
//		queries.put("SULFONAMIDE_N_GLUCURONIDATION" , "[#6]-[#16;X4](=[#8;X1])(=[#8;X1])-[#7;X3;H1,H2]");
//		queries.put("GLUCURONIDATION_OF_PRIMARY_AROMATIC_AMINE" , "[#6;a]-[N;X3;H2]");
//		queries.put("PYRIDINE_GLUCURONIDATION" , "[#6]=,:1[#6]=,:[#6;R1][#7;H1v4X3]=,:[#6;R1][#6]=,:1");
//		queries.put("THIOL_GLUCURONIDATION" , "[#6]-[#16;H1X2]");
//		queries.put("THIOPHENOL_GLUCURONIDATION" , "[#16;H1X2]-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]=,:1");
//		queries.put("_13_DICARBONYL_C_GLUCURONIDATION" , "[O;X1]=[#6;X3]-[#6;X4H2]-[#6;X3]=[O;X1]");
//		queries.put("ISOFLAVONE_4P_OH_GLUCURONIDATION" , "[H][#8;X2]-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1](=,:[#6;R1][#6;R1]=,:1)-[#6;R1]1=,:[#6;R1][#8;R1][#6;R2]2=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R2]2[#6;R1]1=[O;X1]");
//		queries.put("ISOFLAVONE_7_OH_GLUCURONIDATION" , "[H][#8;X2]-[#6;R1]1=,:[#6;R1][#6;R1]=,:[#6;R2]2[#6;R2](=,:[#6;R1]1)[#8;R1][#6;R1]=,:[#6;R1](-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6;R1]=,:[#6;R1][#6;R1]=,:1)[#6;R1]2=[O;X1]");
//		queries.put("GLUCURONIDATION_OF_ALIPHATIC_TERTIARY_AMINE" , "[#6]!@-[#7;A;H1X3;!$([N]*~[#7,#8,#15,#16])]([#6])[#6]");
//		queries.put("AMIDE_N_GLUCURONIDATION" , "[#7;X3;$([H2]),$([H1][#6;!$(C=[O,N,S])])]-[#6](-[#6X4,#7X3])=[O;X1]");
//		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN1" , "[#6]=,:1[#7;H1v3X3R1][#7;R1]=,:[#6,#7][#6,#7]=,:1");
//		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN2" , "[*,#1]-[#7;v3X3R1]1[#6]=,:[#6,#7][#6,#7]=,:[#7;v3R1]1");
//		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN3" , "*-[#7]1[#6]=,:[#6;R1][#7;v3R1]=,:[#6;R1]1");
//		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN4" , "[#6]=,:1[#7;R1][#7;R1]=,:[#6][#7;v3R1]=,:1");
//		queries.put("_3_OH_GLUCURONIDATION_OF_STEROL" , "[#8;A;H1X2][#6]~1~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~2~[#6,#8,#7,#16;A;R2](~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R1]~[#6,#8,#7,#16;A;R2]-,=3-,=[#6,#8,#7,#16;A;R2]-,=4-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16][#6;A;H1X4][#6,#8,#7,#16]-,=4-,=[#6,#8,#7,#16][#6;A;X4;H2,H1][#6,#8,#7,#16]~2-,=3)~[#6,#8,#7,#16]~1");
//		queries.put("_6_OH_GLUCURONIDATION_OF_STEROL" , "[#8;A;H1X2][#6]~1~[#6,#8,#7,#16;A;R1]~[#6,#8,#7,#16;A;R2]~2-,=[#6,#8,#7,#16;A;R2]-,=3-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16][#6;A;H1X4][#6,#8,#7,#16]-,=3-,=[#6,#8,#7,#16][#6;A;X4;H2,H1][#6,#8,#7,#16]~2~[#6,#8,#7,#16]~2~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R2]~1~2");
//		queries.put("_17_OH_GLUCURONIDATION_OF_STEROL" , "[#8;A;H1X2][#6;A;X4]1([#1,CX4H3R0])[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A;R2]-,=2-,=[#6,#8,#7,#16;A;R2]~3~[#6,#8,#7,#16;A;R1]~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R2]~4~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~4~[#6,#8,#7,#16]~3[#6;A;X4;H2,H1][#6,#8,#7,#16]-,=[#6,#8,#7,#16]1-,=2");
//		queries.put("GLUCURONIDATION_OF_AMINE_OXIDE" , "[$([#6][N;X4+]([#6])([#6])[#8;X1-]),$([#6][#7;X3+](=,:[#6])-[#8;X1-])]");
//		queries.put("PHENOLIC_STEROID_SULFONATION" , "[H][#8]-[#6;R1]=,:1[#6;R1]=,:[#6;R1][#6]-2=,:[#6]([#6;R1]=,:1)-[#6;R1]-[#6;R1]-[#6]-1-[#6]-3-[#6;R1]-[#6;R1]-[#6;R1]-[#6]-3-[#6;R1]-[#6;R1]-[#6]-2-1");
//		queries.put("SULFATION_OF_PRIMARY_ALCOHOL" , "[#6;A;H2X4;!$(C([OX2H])[O,S,#7,#15])][#8;H1X2]");
//		queries.put("SULFATION_OF_SECONDARY_ALCOHOL" , "[H][#8][#6;A;H1X4;!$(C([OX2H])[O,S,#7,#15])]([H])([#6])[#6]");
//		queries.put("_3_O_SULFATION_OF_PHENOLIC_COMPOUND" , "[H][#8]-[#6;R1]=,:1[#6;R1]=,:[#6;R1]([H])[#6;R1]([H])=,:[#6;R1](!@-[#6])[#6;R1]=,:1");
//		queries.put("_4_O_SULFATION_OF_PHENOLIC_COMPOUND" , "[H][#6;R1]1=,:[#6;R1][#6;R1](!@-[#8;H1X2R0])=,:[#6;R1][#6;R1]([H])=,:[#6;R1]1!@-[#6]");
//		queries.put("STEROL_3_OH_SULFATION" , "[H][#8]-[#6]~1~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~2~[#6,#8,#7,#16;A;R2](~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R1]~[#6,#8,#7,#16;A;R2]~3-,=[#6,#8,#7,#16;A;R2]-,=4-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A]-,=[#6,#8,#7,#16]-,=4-,=[#6,#8,#7,#16]~[#6,#8,#7,#16;A]~[#6,#8,#7,#16]~2~3)~[#6,#8,#7,#16]~1");
//		queries.put("STEROL_7_OH_SULFATION" , "[H][#8]C1([H])[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R2]~2~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~2~[#6,#8,#7,#16]~2~[#6,#8,#7,#16;A]~[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=3-,=[#6,#8,#7,#16;A]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A;R2]-,=3-,=[#6]1~2");
//		queries.put("STEROL_12_OH_SULFATION" , "[H][#8]C1([H])[#6]~[#6,#8,#7,#16]~2~[#6,#8,#7,#16]~3~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R2]~3~[#6,#8,#7,#16]~[#6,#8,#7,#16;A;R1]~[#6,#8,#7,#16;A;R2]~2-,=[#6,#8,#7,#16;A;R2]-,=2-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16]-,=[#6,#8,#7,#16;A]-,=[#6]1-,=2");
//		queries.put("ARYLAMINE_N_ACETYLATION_PATTERN1" , "[H][#7;X3+0,X4+;!$([N]~[!#6]);!$([N]*~[#7,#8,#15,#16])]-[#6;a]");
//		queries.put("ARYLAMINE_N_ACETYLATION_PATTERN2" , "[H][#7;X3+0,X4+;!$([N]~[!#6])][#7;A;X3]([H])[#6;a]");
//		queries.put("L_CYSTEINE_S_CONJUGATE_N_ACETYLATION" , "[#6]-[#16;X2][#6;A;H2X4][#6;A@H;X4]([#7;A;H2X3])[#6](-[#8-])=O");
//		queries.put("ARALKYLAMINE_N_ACETYLATION" , "[#6;a][#6;A;H2X4][#6;A;H2X4][#7;A;X3+0,X4+;!$([N]~[!#6]);!$([N]*!@[#7,#8,#15,#16])]");
//		queries.put("N_HYDROXYARYLAMINE_O_ACETYLATION" , "[#8;A;X2H1,X1-][#7;A;H1X3]!@-[#6]=,:1[#6]=,:[#6][#6]=,:[#6][#6]=,:1");
//		queries.put("N_HYDROXYL_N_ARYLACETAMIDE_O_ACETYLATION" , "[H][#8]-[#7;X3](-[#6;X3]([#6;A;H3X4])=[O;X1])!@-[#6;X3]-1=[#6]-[#6]=[#6]-[#6]=[#6]-1");
//		queries.put("GSH_CONJUGATION_OF_EPOXIDE" , "[#6]-[#6;X4]-1-[#8]-[#6;X4]-1-[#1,#6,#7,#8]");
//		queries.put("GSH_CONJUGATION_OF_SULFHYDRYL" , "[#6][S;X2H1]");
//		queries.put("GSH_CONJUGATION_OF_ARENE_EPOXIDE" , "[#8;X2]-1-[#6;X3]-2=[#6;X3]-1-[#6]=[#6]-[#6]=[#6]-2");
//		queries.put("GSH_CONJUGATION_OF_ORGANOHALIDE_PATTERN1" , "[H][C;X4]([#1,#6])([F,Cl,Br,I])[*,#1]");
//		queries.put("GSH_CONJUGATION_OF_ORGANOHALIDE_PATTERN2" , "[H][#6;X3](-[F,Cl,Br,I])=[#6;X3]([*,#1])-[*,#1]");
//		queries.put("GSH_CONJUGATION_OF_ACYL_CHLORIDE" , "[#6]-[#6;X3](Cl)=[O;X1]");
//		queries.put("GSH_CONJUGATION_OF_CARBOXIMIDIC_ACID" , "[#6]-[#7]=[#6]-[H1X2;O,S]");
//		queries.put("GSH_CONJUGATION_OF_ALPHA_BETA_UNSATURATED_CARBONYL_PATTERN1" , "[H][#6;X3R1]-1=[#6;X3R1]-[#6;R1](=[!#1!#6;R0;OX1,NX2])-[#6;X3]=[#6;X3]-[#6;R1]-1=[O;X1]");
//		queries.put("GSH_CONJUGATION_OF_ALPHA_BETA_UNSATURATED_CARBONYL_PATTERN2" , "[H][#6;X3](-*)=[#6;X3]([$([CX3;$([R0][#6]),$([H1R0])](=[OX1])[OX2][#6;!$(C=[O,N,S])]),$([$([CX3H][#6]),$([CX3H2])]=[OX1]),$([#6][CX3](=[OX1])[#6])])-[*,#1]");
//		queries.put("GSH_CONJUGATION_OF_ISOCYANATE" , "[#6]-[#7;X2]=[C;X2]=[O;X1]");
//		queries.put("GSH_CONJUGATION_OF_ISOTHIOCYANATE" , "[#6]-[#7;X2]=[C;X2]=[S;X1]");
//		queries.put("GSH_CONJUGATION_OF_ARYL_HALIDE_PATTERN1" , "[!#1!#6;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]~[#6]1=,:[#6][#6]=,:[#6](-[F,Cl,Br,I])[#6](~[#1,!#6;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6]1");
//		queries.put("GSH_CONJUGATION_OF_ARYL_HALIDE_PATTERN2" , "[#1,!#6;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]-[#6]=,:1[#6]=,:[#6][#6](-[F,Cl,Br,I])=,:[#6]([!#1!#6;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6]=,:1");
//		queries.put("GSH_CONJUGATION_OF_ARYL_HALIDE_PATTERN3" , "[$([H][#8]-[#6]=,:1[#6]=,:[#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6]=,:[#6][#6]=,:1-[F,Cl,Br,I]),$([H][#8]-[#6]=,:1[#6]=,:[#6][#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6][#6]=,:1-[F,Cl,Br,I]),$([H][#8]-[#6]1=,:[#6][#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6][#6](-[F,Cl,Br,I])=,:[#6]1),$([H][#8]-[#6]1=,:[#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6]=,:[#6][#6]=,:[#6]1-[F,Cl,Br,I]),$([H][#8]-[#6]=,:1[#6]=,:[#6][#6](-[F,Cl,Br,I])=,:[#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6]=,:1),$([H][#8]-[#6]1=,:[#6][#6]=,:[#6][#6](-[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6]1-[F,Cl,Br,I])]");
//		queries.put("GSH_CONJUGATION_OF_QUINONE_PATTERN1" , "[H][#6;R1]-1=[#6]-[#6]=[#6;R1]-[#6;R1](=[O;X1R0])-[#6;R1]-1=[O;X1R0]");
//		queries.put("GSH_CONJUGATION_OF_QUINONE_PATTERN2" , "[H][#6]-1=[#6]-[#6;R1](=[O;X1R0])-[#6;R1]=[#6;R1]-[#6;R1]-1=[O;X1R0]");
//		queries.put("GSH_CONJUGATION_OF_HALO_S_TRIAZINE" , "[F,Cl,Br,I]-[#6;R1]=,:1[#7;R1]=,:[#6;R1][#7;R1]=,:[#6;R1][#7;R1]=,:1");
//		queries.put("GSH_CONJUGATION_OF_EWG_SUBSTITUTED_BENZENE_PATTERN1" , "[H][#6;R1]1=,:[#6;R1][#6;R1](-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6;R1](-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])[#6;R1]=,:[#6;R1]1-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]");
//		queries.put("GSH_CONJUGATION_OF_EWG_SUBSTITUTED_BENZENE_PATTERN2" , "[H][#6;R1]=,:1[#6;R1](-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6;R1][#6;R1](-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)])=,:[#6;R1][#6;R1]=,:1-[R0;F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]");
//		queries.put("GLYCINATION_OF_ALIPHATIC_ACID" , "[H][#8;X2]-[#6;X3](-[#6;$([#7;A;R0]-,=[#6;A;R0]-,=[#6;A;R0]-,=[#6;A;R0]),$([#7]-[#6;A;X3]=[#6;A]-[#6]=,:1[#6]=,:[#6][#6]=,:[#6][#6]=,:1)])=[O;X1]");
//		queries.put("GLYCINATION_OF_ARYL_ACID" , "[c;a;R1]-[#6;X3](-[OX2H1,OX1-])=O");
//		queries.put("CATECHOL_O_METHYLATION_PATTERN1" , "[H][#8;A;X2][#6;R1]1=,:[#6;R1](-[*,#1;!$([OX2H1])])[#6;R1](-[*,#1])=,:[#6;R1](-[*,#1])[#6;R1](-[*,#1;!$([OX2H1])])=,:[#6;R1]1[#8;A;X2][H]");
//		queries.put("CATECHOL_O_METHYLATION_PATTERN2" , "[H][#8;A;X2][#6;R1]1=,:[#6;R1]([H])[#6;R1]([H])=,:[#6;R1](-[!#8])[#6;R1]([H])=,:[#6;R1]1[#8;A;X2][H]");
//		queries.put("CATECHOL_O_METHYLATION_PATTERN3" , "[H][#8]-[#6]1=,:[#6]([H])[#6]([H])=,:[#6](-[!#8])[#6]([H])=,:[#6]1-[#8][H]");
//		queries.put("THIOL_S_METHYLATION" , "[#6]-[#16;R0;X2H1+0,X1-1]");
//		queries.put("THIOPURINE_S_METHYLATION" , "[#16;A;H1X2][c;R1]1[n;R1][c;R1][n;R1][c;R2]2[n;R1][c;R1][n;R1][c;R2]12");
//		queries.put("THIOETHER_S_METHYLATION" , "[#6]-[#16;X2+0;R0]-[#6]");		

		return queries;
	}
	
	public static LinkedHashMap<String, String> getP2FFingerprintPatterns() throws Exception {
		LinkedHashMap<String, String> queries = new LinkedHashMap<String, String>();
		
		queries.put("heteroatom","[!#1!#6]");//		queries.put("cation","[*;+]");
		queries.put("hydroxyl_acidic","[$([OH]-*=[!#6])]"); // http://www.daylight.com/dayhtml_tutorials/languages/smarts/smarts_examples.html
		queries.put("unsubstituted_aromatic_carbon","[c;H1]");
		queries.put("carboxyl", "[#8;A;X2H1,X1-][#6]([#6,#1;A])=O");
		queries.put("carbon_hydrogen_bond", "[#6][H]");
		queries.put("alcohol","[OX2H1][c,$([CX4;!$(C([OX2H])[O,S,#7,#15])])]");		
		queries.put("carboxylic_acid_or_base","[CX3](=O)[OX1H0-,OX2H1]");
		queries.put("hydroxyl","[OX2H1]-*");
		queries.put("unfunctionalized_methylen_group","[CX4H2]");
		queries.put("hBond", "*[H]");
		queries.put("12_diol","[OX2H][CX4;!$(C([OX2H])[O,S,#7,#15])][CX4;!$(C([OX2H])[O,S,#7,#15])][OX2H]");
		queries.put("ewg","[F,Cl,Br,I,$([N+]),$([#7;X3+](-[#8;X1-])=[O;X1]),$([#6]~[C;X4](F)(F)F),$([C;X4](Cl)(Cl)Cl),$([S;X4]([#8;A;X1-,X2H1])(=O)=O),$([#6]=O),$(C#N)]");
		queries.put("heterocyclic_amine","[#7;a;X3+0,X4+;!$([N]~[!#6]);!$([N]*~[#7,#8,#15,#16])]");
		queries.put("imine","[$([CX3]([#6])[#6]),$([CX3H][#6])]=[$([NX2][#6]),$([NX2H])]");
		queries.put("amine","[NX3+0,NX4+;!$([N]~[!#6]);!$([N]*~[#7,#8,#15,#16])]");
		queries.put("carboxylic_ester","[CX3;$([R0][#6]),$([H1R0])](=[OX1])[OX2][#6;!$(C=[O,N,S])]");
		queries.put("peroxide","[#6]!@-[#8]!@-[#8;A;X2H1,X1-]");
		
		queries.put("hexose","[#8]-[#6]-[#6;R1]-1-[#8;R1]-[#6;R1](-[#8])-[#6;R1](-[#8])-[#6;R1](-[#8])-[#6;R1]-1-[#8]");
		queries.put("ribose","[$([#8]-[#6;R0]-[#6;R0](-[#8])-[#6;R0](-[#8])-[#6;R0](-[#8])-[#6]=O),$([#8]-[#6]-[#6;R1]-1-[#8]-[#6;R1](-[#8])-[#6;R1](-[#8])-[#6;R1]-1-[#8])]");
		queries.put("macrolide","[#6]@-[#8;Rr12]@-[#6;R1r12](@-[#6])=[O;R0]");
		queries.put("coa","[#6]C([#6])([#6]-[#8]P([#8;A;X2H1,X1-])(=O)[#8]P([#8;A;X2H1,X1-])(=O)[#8]-[#6]-[#6]-1-[#8]-[#6](-[#6]([#8;A;X2H1,X1-])-[#6]-1-[#8]P([#8;A;X2H1,X1-])"
				+ "([#8;A;X2H1,X1-])=O)-[#7]-1-[#6]=[#7]-[#6]-2=[#6]-1-[#7]=[#6]-[#7]=[#6]-2-[#7])[#6]([#8;A;X2H1,X1-])-[#6](=O)-[#7]-[#6;R0]-[#6;R0]-[#6;R0](=O)-[#7;R0]-[#6]-[#6]-[#16]");
		queries.put("pyrrole","[#6]=,:1[#6]=,:[#6][#7][#6]=,:1");
		
		
		
//		queries.put("anion","[*;-]");
		
		// aliphatic atom covalently bonded to the positively charged nitrogen of a nitroso group
//		queries.put("aaposn","*-[#7;X2+]=[O;X1]");
		// [$([OX2H][CX4H2;!$(C([OX2H])[O,S,#7,#15])]),$([OX2H][CX4H;!$(C([OX2H])[O,S,#7,#15])]),$([OX2H][CX4D4;!$(C([OX2H])[O,S,#7,#15])])]
		

//		queries.put("epoxide","[#1,#6]!@-[#6]-1-[#8;R1]-[#6]-1!@-[#1,#6]");
//		queries.put("disulfide","[$([#6]-[#16X2]-[#16;X2]-[#1]),$([#6]-[#16X2]-[#16;X2]-[#6])]");
//		queries.put("aryl_halide","[#6;a]!@-[F,Cl,Br,I]");
//		queries.put("acyl_halide","[F,Cl,Br,I]!@-[#6]=[O;X1]");
//		queries.put("n_metyl","[#6;A;H3X4][#7;A;X3+0,X4+;!$([N]~[!#6]);!$([N]*~[#7,#8,#15,#16])]");
//		queries.put("n_metyl","[#6;A;H3X4][#7]");
//		queries.put("o_metyl","[OX2]([CX4H3])[c,$([CX4;!$(C([OX2])[O,S,#7,#15,F,Cl,Br,I])])]");
//		queries.put("aryl_nitro","[#6;a]-[#7;X3+](-[#8;X1-])=[O;X1]");
//		
//		queries.put("CATECHOL_O_METHYLATION_PATTERN1","");
//		queries.put("CATECHOL_O_METHYLATION_PATTERN2","");
//		queries.put("CATECHOL_O_METHYLATION_PATTERN3","");
//		queries.put("THIOL_S_METHYLATION","");
//		queries.put("THIOETHER_S_METHYLATION","");
//		queries.put("GLYCINATION_OF_ALIPHATIC_ACID","");
//		queries.put("GLYCINATION_OF_ARYL_ACID","");
//		queries.put("L_CYSTEINE_S_CONJUGATE_N_ACETYLATION","");
//		queries.put("ARALKYLAMINE_N_ACETYLATION","");
//		queries.put("N_HYDROXYARYLAMINE_O_ACETYLATION","");
//		queries.put("ALKYL_OH_GLUCURONIDATION","");
//		queries.put("AROMATIC_OH_GLUCURONIDATION","");
//		queries.put("ALIPHATIC_ACYL_O_GLUCURONIDATION","");
//		queries.put("AROMATIC_ACYL_O_GLUCURONIDATION","");
//		queries.put("HYDROXYLAMINE_N_GLUCURONIDATION","");
//		queries.put("SULFONAMIDE_N_GLUCURONIDATION","");
//		queries.put("PRIMARY_AROMATIC_AMINE","");
//		queries.put("PYRIDINE_GLUCURONIDATION","");
//		queries.put("THIOL_GLUCURONIDATION","");
//		queries.put("THIOPHENOL_GLUCURONIDATION","");
//		queries.put("_13_DICARBONYL_C_GLUCURONIDATION","");
//		queries.put("ALIPHATIC_TERTIARY_AMINE","");
//		queries.put("AMIDE_N_GLUCURONIDATION","");
//		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN1","");
//		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN2","");
//		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN3","");
//		queries.put("AZOLE_N_GLUCURONIDATION_PATTERN4","");
//		queries.put("EPOXIDE","");
//		queries.put("GSH_CONJUGATION_OF_SULFHYDRYL","");
//		queries.put("GSH_CONJUGATION_OF_ARENE_EPOXIDE","");
//		queries.put("GSH_CONJUGATION_OF_ORGANOHALIDE_PATTERN1","");
//		queries.put("GSH_CONJUGATION_OF_ORGANOHALIDE_PATTERN2","");
//		queries.put("GSH_CONJUGATION_OF_ACYL_CHLORIDE","");
//		queries.put("CARBOXIMIDIC_ACID","");
//		queries.put("ALPHA_BETA_UNSATURATED_CARBONYL_PATTERN1","");
//		queries.put("ALPHA_BETA_UNSATURATED_CARBONYL_PATTERN2","");
//		queries.put("ISOCYANATE","");
//		queries.put("ISOTHIOCYANATE","");
//		queries.put("ARYLAMINE_N_ACETYLATION_PATTERN1","");
//		queries.put("ARYLAMINE_N_ACETYLATION_PATTERN2","");
//		queries.put("ARALKYLAMINE_N_ACETYLATION","");
//		queries.put("N_HYDROXYARYLAMINE_O_ACETYLATION","");
//		queries.put("N_HYDROXYL_N_ARYLACETAMIDE_O_ACETYLATION","");
//		queries.put("SULFATION_OF_PRIMARY_ALCOHOL","");
//		queries.put("SULFATION_OF_SECONDARY_ALCOHOL","");
//		queries.put("HYDRAZINE","");
//		queries.put("","");
//		queries.put("","");
//		queries.put("","");
//		queries.put("","");
//		queries.put("","");
//		queries.put("","");
//		queries.put("","");
//		queries.put("","");
		
		return queries;
	}
	
	public static LinkedHashMap<String, String> getPhaseIIfingerpint(){
		LinkedHashMap<String, String> phaseIIPatterns = new LinkedHashMap<String, String>();
		
//		phaseIIPatterns.put("CATECHOL_O_METHYLATION_PATTERN1","[H][#8;A;X2][#6;R1]1=,:[#6;R1](-[*,#1;!$([OX2H1])])[#6;R1](-[*,#1])=,:[#6;R1](-[*,#1])[#6;R1](-[*,#1;!$([OX2H1])])=,:[#6;R1]1[#8;A;X2]C([H])([H])[H]");
//		phaseIIPatterns.put("CATECHOL_O_METHYLATION_PATTERN2","[H][#8;A;X2][#6;R1]1=,:[#6;R1]([H])[#6;R1]([H])=,:[#6;R1](-[!#8])[#6;R1]([H])=,:[#6;R1]1[#8;A;X2]C([H])([H])[H]");
//		phaseIIPatterns.put("CATECHOL_O_METHYLATION_PATTERN3","[H][#8]-[#6]1=,:[#6]([H])[#6]([H])=,:[#6](-[!#8])[#6]([H])=,:[#6]1-[#8]C([H])([H])[H]");
		phaseIIPatterns.put("GLUCURONIDATION","[#8]-[#6]-1-[#6]-[#8]-[#6;R1](-[#6](-[#8])-[#6]-1-[#8])-[#6;R0](-[#8])=[O;R0]");
		phaseIIPatterns.put("GLUTATHIONE","[$(NC(CCC(=O)NC(CS)C(=O)NCC(O)=O)C(O)=O),$([N[C@@H](CCC(O)=N[C@@H](CS)C(O)=NCC(O)=O)C(O)=O]),$([N[C@@H](CCC(O)=N[C@@H](CS)C(=O)NCC(O)=O)C(O)=O]),$([N[C@@H](CCC(=O)N[C@@H](CS)C(O)=NCC(O)=O)C(O)=O])]");
		phaseIIPatterns.put("ARYLAMINE_N_ACETYLATION_PATTERN1","[H]C([H])([H])[#6](=O)-[#8][#7;A;X3+0,X4+;!$([N]~[!#6]);!$([N]*~[#7,#8,#15,#16])][#6;a]");
		phaseIIPatterns.put("ARYLAMINE_N_ACETYLATION_PATTERN2","[H][#7;A;X3]([#6;a])[#7;A;X3+0,X4+;!$([N]~[!#6])][#8]-[#6](=O)C([H])([H])[H]");
		phaseIIPatterns.put("ARALKYLAMINE_N_ACETYLATION","[#6;a][#6;A;H2X4][#6;A;H2X4][#7;A;X3+0,X4+;!$([N]~[!#6]);!$([N]*!@[#7,#8,#15,#16])][#6](-[#6])=O");
		phaseIIPatterns.put("N_HYDROXYARYLAMINE_O_ACETYLATION","[#6]-[#6](=O)-[#8][#7;A;H1X3]!@-[#6]=,:1[#6]=,:[#6][#6]=,:[#6][#6]=,:1");
//		phaseIIPatterns.put("N_HYDROXYL_N_ARYLACETAMIDE_O_ACETYLATION","");
		phaseIIPatterns.put("SULFATE","[#6]-[#8;X2][S;X4]([#8;A;X2H1,X1-])(=[O;X1])=[O;X1]");
//		phaseIIPatterns.put("SULFATION_OF_SECONDARY_ALCOHOL","");
//		phaseIIPatterns.put("HYDRAZINE","");
		phaseIIPatterns.put("GLYCINATION_OF_ALIPHATIC_ACID","[H][#8]-[#6](=O)-[#6]-[#7]([H])-[#6;X3](-[#6;$([#7;A;R0]-,=[#6;A;R0]-,=[#6;A;R0]-,=[#6;A;R0]),$([#7]-[#6;A;X3]=[#6;A]-[#6]=,:1[#6]=,:[#6][#6]=,:[#6][#6]=,:1)])=[O;X1]");
		phaseIIPatterns.put("GLYCINATION_OF_ARYL_ACID","[H][#8;A][#6;X3](=O)[#6;A;X4]([H])([H])[#7;X3]([H])-[#6;X3](-[#6;a;R1])=O");
//		phaseIIPatterns.put("","");
//		phaseIIPatterns.put("","");
//		phaseIIPatterns.put("","");
//		phaseIIPatterns.put("","");
//		phaseIIPatterns.put("","");

		
		return phaseIIPatterns;
		
	}

	public static boolean isPotentialPhaseIICandidateByReactionPatternMatching(IAtomContainer substrate) throws Exception{
		boolean phaseII = false;
		
		if(containsCarbon(substrate)){
			IChemObjectBuilder builder = SilentChemObjectBuilder.getInstance();
//			LinkedHashMap<String, String> sfeatures = phaseIIsubstratePatterns;
			StructuralClassName className = StructuralClassFinder.findStructuralClass(substrate);
			
			if(className == StructuralClassName.ETHER_LIPID || className == StructuralClassName.GLYCEROLIPID ||
					className == StructuralClassName.GLYCEROPHOSPHOLIPID || className == StructuralClassName.SPHINGOLIPID
					|| isAcylCoAConjugate(substrate)){
				phaseII = false;
			}
			else{
//				System.out.println(phaseIIsubstratePatterns);
				for(Map.Entry<String, String> s : getPhase2CandidateFingerprintPatterns().entrySet()){
//					System.out.println(s.getKey());
//				for(String s : sfeatures.values()){
					SmartsPattern pattern = SmartsPattern.create(s.getValue(), builder);
					if(pattern.matches(substrate)){
						phaseII = true;
//						System.err.println(s);
						break;
					}			
				}	
			}
			
		
		}

		return phaseII;
	}
	
	
	public static boolean isPotentialPhaseIIMetaboliteByReactionPatternMatching(IAtomContainer substrate) throws Exception{
		boolean phaseII = false;
		
		if(containsCarbon(substrate)){
			IChemObjectBuilder builder = SilentChemObjectBuilder.getInstance();
			LinkedHashMap<String, String> sfeatures = phaseIIMetabolitePatterns;

			for(Map.Entry<String, String> s : sfeatures.entrySet()){
				SmartsPattern pattern = SmartsPattern.create(s.getValue(), builder);
				if(pattern.matches(substrate)){
					phaseII = true;
					System.out.println(s.getKey());
					break;
				}			
			}			
		}

		return phaseII;
	}
	
	
	public static boolean isAcylCoAConjugate(IAtomContainer molecule) throws SMARTSException{
		SmartsPatternCDK glutathioneConjugatePattern = new SmartsPatternCDK("[#6;R0]-[#6;R0](=O)-[#16]-[#6]-[#6]-[#7]-[#6](=O)-[#6]-[#6]-[#7]-[#6](=O)-[#6](-[#8])C([#6])([#6])[#6]-[#8]P([#8])(=O)[#8]P([#8])(=O)[#8]-[#6]-[#6]-1-[#8]-[#6](-[#6](-[#8])-[#6]-1-[#8]P([#8])([#8])=O)-[#7]-1-[#6]=[#7]-[#6]-2=[#6]-1-[#7]=[#6]-[#7]=[#6]-2-[#7]");	
		return glutathioneConjugatePattern.hasSMARTSPattern(molecule)>0;	
	}
	
	public static boolean containsCarbon(IAtomContainer molecule) {
		boolean carbon = false;
		for(IAtom at : molecule.atoms()){
			if(at.getAtomicNumber() == 6){
				carbon = true;
				break;
			}
		}	
		return carbon;
	}
	
	/**
	 * 
	 * @param mole
	 *            : A molecule of interest
	 * @param queries
	 *            : A HashMap with the functional groups and patterns to detect,
	 *            with their SMARTS patterns
	 * @return : A list containing atom-based list of the bits (1 or 0) that
	 *         describe for every functional group/pattern, whether the atom of
	 *         interest is part of a match.
	 * @throws Exception
	 */

	public ArrayList<ArrayList<Integer>> generateClassyfireAtomFingeprint(
			IAtomContainer mole, LinkedHashMap<String, String> queries) throws Exception {
		ArrayList<ArrayList<Integer>> results = new ArrayList<ArrayList<Integer>>();
		ArrayList<ArrayList<Integer>> allBits = new ArrayList<ArrayList<Integer>>();
		LinkedHashMap<Integer, int[]> atomFingerprints = new LinkedHashMap<Integer, int[]>();

		for (Map.Entry<String, String> item : queries.entrySet()) {
			// System.out.println(item.getKey());
			IChemObjectBuilder builder = DefaultChemObjectBuilder.getInstance();
			SMARTSQueryTool smartsPattern = new SMARTSQueryTool(item.getValue(), builder);
			List<List<Integer>> matchingAtoms = getMatchingAtoms(mole, smartsPattern);
			ArrayList<Integer> bits = new ArrayList<Integer>();

			if (matchingAtoms.size() > 0) {
				System.out.println(item.getKey());
				System.out.println("matching_atoms: " + matchingAtoms);
				for (int i = 0; i < matchingAtoms.size(); i++) {
					// System.out.println(matching_atoms.get(i));
					List<Integer> indexes = matchingAtoms.get(i);
					for (int j = 0; j < indexes.size(); j++) {
						bits.add(indexes.get(j));
						// results.add(bits);
					}
				}
			}
			if (bits.size() > 0) {
				System.out.println("Bits: " + bits);
			}

			allBits.add(new ArrayList<Integer>(new HashSet<Integer>((bits)))); // to
																				// remove
																				// duplicates
		}

		System.out.println(allBits);
		System.out.println(allBits.size());

		for (int k = 0; k < mole.getAtomCount(); k++) {
			ArrayList<Integer> nilArraylist = new ArrayList<Integer>();
			for (int a = 0; a < queries.size(); a++) {
				nilArraylist.add(0);
			}
			results.add(nilArraylist);
		}

		for (int l = 0; l < allBits.size(); l++) {
			if (allBits.get(l).size() > 0) {
				for (int m = 0; m < allBits.get(l).size(); m++) {
					int index = allBits.get(l).get(m);
					results.get(index).set(l, 1);
				}
			}
		}

		return results;
	}

	/**
	 * This function generates a unique fingerprint for a molecule.
	 * 
	 * @param molecule
	 *            : The molecule of interest
	 * @return : An array list with the fingerprint of the molecule
	 * @throws Exception
	 */

	public ArrayList[] generateClassyfireFingerprint(IAtomContainer molecule)
			throws Exception {

		StructSearcher cs = new StructSearcher();
//		LinkedHashMap<String, String> queries = cs.getFingerprintPatterns();
		LinkedHashMap<String, String> queries = cs.getP2FFingerprintPatterns();
		
		LinkedHashMap<String, Integer> bits = new LinkedHashMap<String, Integer>();
		int nr_of_present_groups = 0;

		ArrayList<String> fingerprint_bits = new ArrayList<String>();
		ArrayList<String> features = new ArrayList<String>();
		ArrayList[] results = new ArrayList[2];

		IChemObjectBuilder builder = DefaultChemObjectBuilder.getInstance();
		for (Map.Entry<String, String> item : queries.entrySet()) {
			// iterate over the items of a LinkedHashMap
			// System.out.println(item.getKey());
			SMARTSQueryTool smartsPattern = new SMARTSQueryTool(item.getValue(), builder);

			if (getMatchingAtoms(molecule, smartsPattern).size() > 0) {
				bits.put(item.getKey(), 1);
				nr_of_present_groups = nr_of_present_groups + 1;
				fingerprint_bits.add("1");
				features.add(item.getKey());
			}
			else {
				fingerprint_bits.add("0");
			}
		}
		// for (Map.Entry<String,Integer> bit : bits.entrySet()) {
		// fingerprint = fingerprint + bit.getValue().toString();
		// }
		// fingerprint = bits.get("carboxyl").toString() +
		// bits.get("hydroxyl").toString() + bits.get("aromatic").toString() +
		// bits.get("sulfuric acid").toString() +
		// bits.get("carbonyl").toString() + bits.get("aldehyde").toString() +
		// bits.get("indole").toString();
		// System.out.println("Fingerprint: " + fingerprint);
		// System.out.println("Fingerprint size: " + fingerprint.length());
		// System.out.println("Number of on bits: " + nr_of_present_groups);

		results[0] = features;
		results[1] = fingerprint_bits;
		return results;
	}


	public LinkedHashMap<String, Double> generateP2FFingerprint(IAtomContainer molecule) throws Exception{
		LinkedHashMap<String, Double> fingerprint =  new LinkedHashMap<String, Double>();
		AtomContainerManipulator.convertImplicitToExplicitHydrogens(molecule);
		for(Map.Entry<String, String> q : this.queries.entrySet()){
			SmartsPattern pattern = SmartsPattern.create(q.getValue(), this.builder);
			int nrOfUniqueMatches = pattern.matchAll(molecule).countUnique();
//			System.out.println(q.getKey() + " : " + nrOfUniqueMatches);
			fingerprint.put(q.getKey(), Double.valueOf(nrOfUniqueMatches));
		}
		
		
		
		return fingerprint;
		
		
	}
	
	
	/**
	 * 
	 * @param molecule
	 *            : The molecule of interest
	 * @param smartsPattern
	 *            : The SMARTS pattern to be search for in the molecule
	 * @return : A list containing a list with the indexes of each corresponding
	 *         atom match for each occurrence of the pattern
	 * @throws Exception
	 */

	public List<List<Integer>> getMatchingAtoms(IAtomContainer molecule,
			SMARTSQueryTool smartsPattern) throws Exception {
		IChemObjectBuilder builder = SilentChemObjectBuilder.getInstance();
		boolean occurrences = smartsPattern.matches(molecule);
		List<List<Integer>> indexes = smartsPattern.getMatchingAtoms();
		for (int i = 0; i < indexes.size(); i++) {
			List<Integer> currentAtomIndexes = indexes.get(i);
			System.out.println("\nMatch nr." + (i + 1) + ": " + (currentAtomIndexes));
			for (int j = 0; j < currentAtomIndexes.size(); j++) {
				System.out.println(currentAtomIndexes.get(j) + ": "
						+ molecule.getAtom((int) currentAtomIndexes.get(j)));
			}
		}
		return indexes;
	}

	/**
	 * This function creates a tsv file with the atom fingerprints
	 * 
	 * @param sdfInput
	 *            : a SDF file
	 * @param queries
	 *            : a HashMap with fingerprint patterns and their SMARTS
	 *            expressions
	 * @throws Exception
	 */
	public void serialAtomfingerprinterToSdf(File sdfInput,
			LinkedHashMap<String, String> queries) throws Exception {
		String absolutePath = sdfInput.getAbsolutePath();
		String destinationDirectory = absolutePath.substring(0,
				absolutePath.lastIndexOf(File.separator));

		String[] substrings = sdfInput.getName().split("/");
		String name = substrings[substrings.length - 1].split(".sdf")[0];
		File sdfOutput = new File(destinationDirectory + "/atom_fingerprint_" + name
				+ ".tsv");
		FileOutputStream sdfOutputOs = new FileOutputStream(sdfOutput);
		BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(sdfOutputOs));

		StringBuilder patternNames = new StringBuilder();
		for (Map.Entry<String, String> item : queries.entrySet()) {
			patternNames.append(item.getKey()).append("\t");
		}
		System.out.println(patternNames);
		bw.write("molecule ID\tatom ID\t" + patternNames + "\n"); // add header

		IteratingSDFReader reader = new IteratingSDFReader(new FileInputStream(sdfInput),
				DefaultChemObjectBuilder.getInstance());
		ArrayList<ArrayList<ArrayList<Integer>>> totalBits = new ArrayList<ArrayList<ArrayList<Integer>>>();
		ElectronDonation model = ElectronDonation.cdk();
		CycleFinder cycles = Cycles.cdkAromaticSet();
		Aromaticity aromaticity = new Aromaticity(model, cycles);

		int i = 0;
		while (reader.hasNext()) {
			i = i + 1;
			IAtomContainer molecule = (IAtomContainer) reader.next();
			AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(molecule); // 1
			aromaticity.apply(molecule); // 1
			// Kekulization.kekulize(molecule); // 2

			// CDKHydrogenAdder adder =
			// CDKHydrogenAdder.getInstance(molecule.getBuilder()); // 3
			// adder.addImplicitHydrogens(molecule); // 3
			// CDKHydrogenAdder.getInstance(builder).addImplicitHydrogens(molecule);
			// // 4 -
			// //http://efficientbits.blogspot.ca/2013/12/new-smiles-behaviour-parsing-cdk-154.html

			ArrayList<ArrayList<Integer>> atom_fingerprint = generateClassyfireAtomFingeprint(
					molecule, queries);
			totalBits.add(atom_fingerprint);
		}

		for (int j = 0; j < totalBits.size(); j++) {
			System.out.println("Number of atoms of molecule " + (j + 1) + ": "
					+ totalBits.get(j).size());
			for (int k = 0; k < totalBits.get(j).size(); k++) {
				ArrayList<Integer> bits = totalBits.get(j).get(k);
				// System.out.println("Number of patterns" + "\t" +
				// bits.size());
				StringBuilder fingerprintBits = new StringBuilder();

				for (int bit : bits) {
					fingerprintBits.append(bit).append("\t"); // separating
																// contents
																// using commas
				}
				bw.write((j + 1)
						+ "\t"
						+ (k + 1)
						+ "\t"
						+ fingerprintBits.deleteCharAt((fingerprintBits.length() - 1))
								.toString() + "\n");
				System.out.println((j + 1) + "\t" + (k + 1) + "\t" + fingerprintBits);
				// sdf_output.

			}
		}
		// return total_bits
		bw.close();
	}


	/**
	 * 
	 * @param sdfInput
	 *            : a SDF file
	 * @param queries
	 *            : a HashMap with fingerprint patterns and their SMARTS
	 *            expressions
	 * @return An array list of fingerprints for every atom of each molecule in
	 *         the SDF file
	 * @throws Exception
	 */

	public ArrayList<ArrayList<ArrayList<Integer>>> geterateSerialAtomfingerprintToArraylist(
			File sdfInput, LinkedHashMap<String, String> queries) throws Exception {
		String absolutePath = sdfInput.getAbsolutePath();
		String destinationDirectory = absolutePath.substring(0,
				absolutePath.lastIndexOf(File.separator));

		String[] substrings = sdfInput.getName().split("/");
		String name = substrings[substrings.length - 1].split(".sdf")[0];
		File sdf_output = new File(destinationDirectory + "/atom_fingerprint_" + name
				+ ".tsv");
		FileOutputStream sdfOutputOs = new FileOutputStream(sdf_output);
		BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(sdfOutputOs));

		StringBuilder patternNames = new StringBuilder();
		for (Map.Entry<String, String> item : queries.entrySet()) {
			patternNames.append(item.getKey()).append("\t");
		}
		System.out.println(patternNames);
		bw.write("molecule ID\tatom ID\t" + patternNames + "\n"); // add header

		IteratingSDFReader reader = new IteratingSDFReader(new FileInputStream(sdfInput),
				DefaultChemObjectBuilder.getInstance());
		ArrayList<ArrayList<ArrayList<Integer>>> totalBits = new ArrayList<ArrayList<ArrayList<Integer>>>();
		ElectronDonation model = ElectronDonation.cdk();
		CycleFinder cycles = Cycles.cdkAromaticSet();
		Aromaticity aromaticity = new Aromaticity(model, cycles);

		int i = 0;
		while (reader.hasNext()) {
			i = i + 1;
			IAtomContainer molecule = (IAtomContainer) reader.next();
			AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(molecule); // 1
			aromaticity.apply(molecule); // 1
			// Kekulization.kekulize(molecule); // 2

			 CDKHydrogenAdder adder =
			 CDKHydrogenAdder.getInstance(molecule.getBuilder()); // 3
			 adder.addImplicitHydrogens(molecule); // 3
			// CDKHydrogenAdder.getInstance(builder).addImplicitHydrogens(molecule);
			// // 4 -
			// //http://efficientbits.blogspot.ca/2013/12/new-smiles-behaviour-parsing-cdk-154.html
			
			System.out.println("\n\n\n\n----------------------> " + molecule.getProperty(CDKConstants.TITLE));
			ArrayList<ArrayList<Integer>> atom_fingerprint = generateClassyfireAtomFingeprint(
					molecule, queries);
			totalBits.add(atom_fingerprint);
		}

		for (int j = 0; j < totalBits.size(); j++) {
			System.out.println("Number of atoms of molecule " + (j + 1) + ": "
					+ totalBits.get(j).size());
			for (int k = 0; k < totalBits.get(j).size(); k++) {
				ArrayList<Integer> bits = totalBits.get(j).get(k);
				// System.out.println("Number of patterns" + "\t" +
				// bits.size());
				StringBuilder fingerprintBits = new StringBuilder();

				for (int bit : bits) {
					fingerprintBits.append(bit).append("\t"); // separating
																// contents
																// using commas
				}
				bw.write((j + 1)
						+ "\t"
						+ (k + 1)
						+ "\t"
						+ fingerprintBits.deleteCharAt((fingerprintBits.length() - 1))
								.toString() + "\n");
				System.out.println((j + 1) + "\t" + (k + 1) + "\t" + fingerprintBits);
				// sdf_output.

			}
		}
		bw.close();
		return totalBits;

	}



	/**
	 * 
	 * @param sdfInput
	 *            : a SDF file
	 * @param queries
	 *            : a HashMap with fingerprint patterns and their SMARTS
	 *            expressions
	 * @return : A string with the atom-based fingerprints of every molecule. A
	 *         .csv file is also saved.
	 * @throws Exception
	 */

	public String saveSerialAtomFingerprinterToCSV(File sdfInput,
			LinkedHashMap<String, String> queries) throws Exception {
		String absolutePath = sdfInput.getAbsolutePath();
		String destinationDirectory = absolutePath.substring(0,
				absolutePath.lastIndexOf(File.separator));
		String[] substrings = sdfInput.getName().split("/");
		String name = substrings[substrings.length - 1].split(".sdf")[0];
		String output = destinationDirectory + "/" + name + "_atom_fingerprint" + ".csv";
		File sdfOutput = new File(output);
		FileOutputStream sdfOutputOs = new FileOutputStream(sdfOutput);
		BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(sdfOutputOs));

		StringBuilder patternNames = new StringBuilder();
		for (Map.Entry<String, String> item : queries.entrySet()) {
			patternNames.append(item.getKey()).append(",");
		}
		// System.out.println(pattern_names);
		// bw.write("molecule ID\tatom ID\t" + pattern_names + "\n"); //add
		// header

		IteratingSDFReader reader = new IteratingSDFReader(new FileInputStream(sdfInput),
				DefaultChemObjectBuilder.getInstance());
		ArrayList<ArrayList<ArrayList<Integer>>> totalBits = new ArrayList<ArrayList<ArrayList<Integer>>>();
		ElectronDonation model = ElectronDonation.cdk();
		CycleFinder cycles = Cycles.cdkAromaticSet();
		Aromaticity aromaticity = new Aromaticity(model, cycles);

		int i = 0;
		while (reader.hasNext()) {
			i = i + 1;
			IAtomContainer molecule = (IAtomContainer) reader.next();
			AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(molecule); // 1
			aromaticity.apply(molecule); // 1
			// Kekulization.kekulize(molecule); // 2

			// CDKHydrogenAdder adder =
			// CDKHydrogenAdder.getInstance(molecule.getBuilder()); // 3
			// adder.addImplicitHydrogens(molecule); // 3
			// CDKHydrogenAdder.getInstance(builder).addImplicitHydrogens(molecule);
			// // 4 -
			// //http://efficientbits.blogspot.ca/2013/12/new-smiles-behaviour-parsing-cdk-154.html

			ArrayList<ArrayList<Integer>> atomFingerprint = generateClassyfireAtomFingeprint(
					molecule, queries);
			totalBits.add(atomFingerprint);
		}

		for (int j = 0; j < totalBits.size(); j++) {
			// System.out.println("Number of atoms of molecule " + (j+1) + ": "
			// + total_bits.get(j).size());
			for (int k = 0; k < totalBits.get(j).size(); k++) {
				ArrayList<Integer> bits = totalBits.get(j).get(k);
				// System.out.println("Number of patterns" + "\t" +
				// bits.size());
				StringBuilder fingerprintBits = new StringBuilder();

				for (int bit : bits) {
					fingerprintBits.append(bit).append(","); // separating
																// contents
																// using commas
				}
				bw.write((j + 1)
						+ ","
						+ (k + 1)
						+ ","
						+ fingerprintBits.deleteCharAt((fingerprintBits.length() - 1))
								.toString() + "\n");
				// System.out.println( (j+1) + "," +(k+1) + "," +
				// fingerprint_bits);
				// sdf_output.

			}
		}
		bw.close();
		return output;
		// return total_bits;

	}
	
}

