package phase2filter.utils;



import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.Map;

import org.openscience.cdk.exception.CDKException;
import org.openscience.cdk.exception.InvalidSmilesException;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smiles.SmilesParser;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import ambit2.smarts.query.SMARTSException;
import ambit2.smarts.query.SmartsPatternCDK;



public class StructuralClassFinder {
	
	public StructuralClassFinder() {
		// TODO Auto-generated constructor stub
	}

	public enum StructuralClassName {
		
		ALPHA_HYDROXYLATED_FATTY_ACID, BETA_HYDROXYLATED_FATTY_ACID,		
		BILE_ACID, ETHER_LIPID, 
		FATTY_ACID,
		GLYCEROLIPID, GLYCEROPHOSPHOLIPID, 
		OMEGA_HYDROXYLATED_FATTY_ACID,
		SPHINGOLIPID, 
		UNSATURATED_FATTY_ACID,
		UNFUNCTIONALIZED_UNSATURATED_FATTY_ACID,
		SATURATED_FATTY_ACID,
		UNFUNCTIONALIZED_SATURATED_FATTY_ACID
		
		
	}
	
	public static LinkedHashMap<StructuralClassName, LinkedHashMap<String, String[]>>			chemicalClassDefinitions;
	static {
		chemicalClassDefinitions = new LinkedHashMap<StructuralClassName, LinkedHashMap<String, String[]>>();
		
		chemicalClassDefinitions.put(StructuralClassName.ALPHA_HYDROXYLATED_FATTY_ACID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.ALPHA_HYDROXYLATED_FATTY_ACID).put("smarts", new String[]{

		});
		chemicalClassDefinitions.get(StructuralClassName.ALPHA_HYDROXYLATED_FATTY_ACID).put("negativeSmarts", new String[]{

		});
		chemicalClassDefinitions.put(StructuralClassName.BETA_HYDROXYLATED_FATTY_ACID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.BETA_HYDROXYLATED_FATTY_ACID).put("smarts", new String[]{

		});
		chemicalClassDefinitions.get(StructuralClassName.BETA_HYDROXYLATED_FATTY_ACID).put("negativeSmarts", new String[]{

		});
		chemicalClassDefinitions.put(StructuralClassName.BILE_ACID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.BILE_ACID).put("smarts", new String[]{

		});
		chemicalClassDefinitions.get(StructuralClassName.BILE_ACID).put("negativeSmarts", new String[]{

		});
		chemicalClassDefinitions.put(StructuralClassName.ETHER_LIPID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.ETHER_LIPID).put("smarts", new String[]{

		});
		chemicalClassDefinitions.get(StructuralClassName.ETHER_LIPID).put("negativeSmarts", new String[]{

		});
		chemicalClassDefinitions.put(StructuralClassName.FATTY_ACID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.FATTY_ACID).put("smarts", new String[]{
				"[#6]!@-,!@=[#6]!@-,!@=[#6]-[#6;X3]([#8;A;X1-,X2H1])=[O;X1]",
				"C=C"

		});
		chemicalClassDefinitions.get(StructuralClassName.FATTY_ACID).put("negativeSmarts", new String[]{

		});
		
		chemicalClassDefinitions.put(StructuralClassName.SATURATED_FATTY_ACID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.SATURATED_FATTY_ACID).put("smarts", new String[]{
				"[#6;A;X4;H2,H3][#6;A;H2X4][#6;A;H2X4][#6;X3]([#8;A;X2H1,X1-])=[O;X1]"
		});
		chemicalClassDefinitions.get(StructuralClassName.SATURATED_FATTY_ACID).put("negativeSmarts", new String[]{
				"C=C",
				"C#C"
		});

		chemicalClassDefinitions.put(StructuralClassName.UNFUNCTIONALIZED_SATURATED_FATTY_ACID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.UNFUNCTIONALIZED_SATURATED_FATTY_ACID).put("smarts", new String[]{
				"[#6;A;X4;H2,H3][#6;A;H2X4][#6;A;H2X4][#6;X3]([#8;A;X2H1,X1-])=[O;X1]"
		});
		chemicalClassDefinitions.get(StructuralClassName.UNFUNCTIONALIZED_SATURATED_FATTY_ACID).put("negativeSmarts", new String[]{
				"C=C",
				"C#C",
				"[#6;X4][#8;A;H1X2]"
		});
		
		chemicalClassDefinitions.put(StructuralClassName.GLYCEROLIPID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.GLYCEROLIPID).put("smarts", new String[]{
				"[$([OX1-,OX2H1,$([OX2](-[CX4])[CX4;R0;H2]),$([OX2]-[CX3]=[CX3]),$([OX2]-[CX3](=O)-[#1,#6])][#6;A;H2X4R0][#6;A;H1X4R0]([OX1-,OX2H1,$([OX2](-[CX4])[CX4;R0;H2]),$([OX2]-[CX3]=[CX3]),$([OX2]-[CX3](=O)-[#1,#6])])[#6;A;H2X4R0][#8]-[CX4,$([CX3]=[CX3]),$([CX3](=O)-[#1,#6])]),$([OX1-,OX2H1,$([OX2](-[CX4])[CX4;R0;H2]),$([OX2]-[CX3]=[CX3]),$([OX2]-[CX3](=O)-[#1,#6])][#6;A;H2X4R0][#6;A;H1X4R0]([#6;A;H2X4R0][OX1-,OX2H1,$([OX2](-[CX4])[CX4;R0;H2]),$([OX2]-[CX3]=[CX3]),$([OX2]-[CX3](=O)-[#1,#6])])[#8;X2]-[CX4,$([CX3]=[CX3]),$([CX3](=O)-[#1,#6])])]"	
		});
		chemicalClassDefinitions.get(StructuralClassName.GLYCEROLIPID).put("negativeSmarts", new String[]{
		});
		
		chemicalClassDefinitions.put(StructuralClassName.GLYCEROPHOSPHOLIPID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.GLYCEROPHOSPHOLIPID).put("smarts", new String[]{
				"["
				+ "$([#8]P([!#1!#6;OX1-,OX2H1,$([O]-[#6])])(=O)[#8;R0][#6;A;H2X4R0][#6;A;H1X4R0]([R0;OX1-,OX2H1,$([OX2]-[#1,CX4,$(C(=O)-[#1,#6])])])[#6;A;H2X4R0][R0;OX1-,OX2H1,$([OX2]-[#1,CX4,$(C(=O)-[#1,#6])])]),"
				+ "$([#8]P([!#1!#6;OX1-,OX2H1,$([O]-[#6])])(=O)[#8;R0][#6;A;H2X4R0][#6;A;X3R0](=O)[#6;A;H2X4R0][R0;OX1-,OX2H1,$([OX2]-[#1,CX4,$(C(=O)-[#1,#6])])])]"
		});
		chemicalClassDefinitions.get(StructuralClassName.GLYCEROPHOSPHOLIPID).put("negativeSmarts", new String[]{

		});
		
		chemicalClassDefinitions.put(StructuralClassName.SPHINGOLIPID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.SPHINGOLIPID).put("smarts", new String[]{
				"["
//				+ "$([#6;A;H3X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H1X4]([#8;A;X2H1,X1-])[#6;A;H1X4]([!#1!#6;NX4H3+,NX3H2,$([#7X3]C(=O)C)])[#6;A;H2X4][#8]),"
//				+ "$([#6;A;H3X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;H1X3]=[#6;H1X3][#6;A;H1X4]([#8;A;X2H1,X1-])[#6;A;H1X4]([!#1!#6;NX4H3+,NX3H2,$([#7X3]C(=O)C)])[#6;A;H2X4][#8])"

				+ "$([#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H1X4]([#8;A;X2H1,X1-])[#6;A;H1X4]([!#1!#6;NX4H3+,NX3H2,$([#7X3]C(=O)C)])[#6;A;H2X4][#8]),"
//				+ "$([#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;H1X3]=[#6;H1X3][#6;A;H1X4]([#8;A;X2H1,X1-])[#6;A;H1X4]([!#1!#6;NX4H3+,NX3H2,$([#7X3]C(=O)C)])[#6;A;H2X4][#8])"
				+ "$([H][#6]([#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4])-,=[#6]([H])[#6;A;H2X4]C([H])([#1,OX2H1])[#6;H1X3]=[#6;H1X3][#6;A;H1X4]([#8;A;X2H1,X1-])[#6;A;H1X4]([!#1!#6;NX4H3+,NX3H2,$([#7X3]C(=O)C)])[#6;A;H2X4][#8]),"
				// 	5-hydroxy,3E-sphingosine (LMSP01080004)
				+ "$([H][#8]C([H])([#6;A;H2X4][#6]([H])-,=[#6]([H])[#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4][#6;A;H2X4])[#6;H1X3]=[#6;H1X3][#6;A;H1X4]([!#1!#6;NX4H3+,NX3H2,$([#7X3]C(=O)C)])[#6;A;H2X4][#8])"
				+ "]"
		});
		chemicalClassDefinitions.get(StructuralClassName.SPHINGOLIPID).put("negativeSmarts", new String[]{
		});		
		
		chemicalClassDefinitions.put(StructuralClassName.ETHER_LIPID, new LinkedHashMap<String, String[]>());		
		chemicalClassDefinitions.get(StructuralClassName.ETHER_LIPID).put("smarts", new String[]{
				"[$([#8;X2][#6;A;H2X4]!@-[#6;A;X4](!@-[!#1!#6;$([OX2H1]),$([OX2]-[CX4H2]-[#6;A])])!@-[#6;A;H2X4][!#1!#6;$([OX2H1]),$([OX2]-[CX4H2]-[#6;A])]),"
				+ "$([#8]!@-[#6;A;X4](!@-[#6;A;H2X4][!#1!#6;$([OX2H1]),$([OX2]-[CX4H2]-[#6;A])])!@-[#6;A;H2X4][!#1!#6;$([OX2H1]),$([OX2]-[CX4H2]-[#6;A])]),"
				+ "$([!#1!#6;$([OX2H1]),$([OX2]-[CX4H2]-[#6;A])][#6;A;H2X4]!@-[#6;A;X3](!@=[O;X1])!@-[#6;A;H2X4][!#1!#6;$([OX2H1]),$([OX2]-[CX4H2]-[#6;A])]),"
				+ "$([#8;X2][#6;A;H2X4]!@-[#6;A;X3](!@=[O;X1])!@-[#6;A;H2X4][!#1!#6;$([OX2H1]),$([OX2]-[CX4H2]-[#6;A])])"
				+ "]"
		});
		chemicalClassDefinitions.get(StructuralClassName.ETHER_LIPID).put("negativeSmarts", new String[]{
		});
		
		
		
//		chemicalClassDefinitions.put(ChemicalClassName, new LinkedHashMap<String, String[]>());		
//		chemicalClassDefinitions.get(ChemicalClassName).put("smarts", new String[]{
//				""
//		});
//		chemicalClassDefinitions.get(ChemicalClassName.).put("negativeSmarts", new String[]{
//				""
//		});
		
		
//		chemicalClassDefinitions.put(ChemicalClassName.GLYCEROPHOSPHOLIPID, "");
//		chemicalClassDefinitions.put(ChemicalClassName.OMEGA_HYDROXYLATED_FATTY_ACID, "");
//		chemicalClassDefinitions.put(ChemicalClassName.SPHINGOLIPID, "");
//		chemicalClassDefinitions.put(ChemicalClassName.UNSATURATED_FATTY_ACID, "");
	
	}
	
	
	public static StructuralClassName findStructuralClass(IAtomContainer molecule) throws SMARTSException, CloneNotSupportedException, CDKException{
		StructuralClassName chemClass = null;
//		IAtomContainer molecule = mol.clone();
//		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(molecule);
//		AtomContainerManipulator.convertImplicitToExplicitHydrogens(molecule);
		
		if(isAlphaHydroxyFattyAcid(molecule)){
			chemClass = StructuralClassName.ALPHA_HYDROXYLATED_FATTY_ACID;
		} else if(isBetaHydroxyFattyAcid(molecule)){
			chemClass = StructuralClassName.BETA_HYDROXYLATED_FATTY_ACID;
		} else if(isOmegaHydroxyFattyAcid(molecule)){
			chemClass = StructuralClassName.OMEGA_HYDROXYLATED_FATTY_ACID;
		} else if(isUnsaturatedFattyAcid(molecule)){
			chemClass = StructuralClassName.UNSATURATED_FATTY_ACID;
		} else if(isFattyAcid(molecule)){
			chemClass = StructuralClassName.FATTY_ACID;
		}  		
		else {
			for(Map.Entry<StructuralClassName, LinkedHashMap<String, String[]>> cc : chemicalClassDefinitions.entrySet()){
				if( 
						cc.getKey() == StructuralClassName.GLYCEROLIPID || 
						cc.getKey() == StructuralClassName.GLYCEROPHOSPHOLIPID ||
						cc.getKey() == StructuralClassName.ETHER_LIPID || 
						cc.getKey() == StructuralClassName.SPHINGOLIPID ){
					
					boolean b = true;
					
					for(String smart : chemicalClassDefinitions.get(cc.getKey()).get("smarts")){
						SmartsPatternCDK pattern = new SmartsPatternCDK(smart);	
						if(!(pattern.hasSMARTSPattern(molecule)>0)){
							b = false;
							break;
						}
					}					
					for(String negativeSmart : chemicalClassDefinitions.get(cc.getKey()).get("negativeSmarts")){
						SmartsPatternCDK npattern = new SmartsPatternCDK(negativeSmart);	
						if(npattern.hasSMARTSPattern(molecule)>0){
							b = false;
							break;
						}
					}
					
					if(b == true){
						chemClass = cc.getKey();
					}					
				}
			}
		}
		

		
		
		
		return chemClass;
	}
	
	
	public static boolean isUnsubstitutedSatudatedFattyAcid(IAtomContainer molecule){	
		
		return false;
	}
	
	
	public static boolean isFattyAcid(IAtomContainer molecule){	
		
		return false;
	}
	
	
	
	
	public static boolean isUnsaturatedFattyAcid(IAtomContainer molecule){	
		
		return false;
	}
	
	public static boolean isSaturatedFattyAcid(IAtomContainer molecule) throws SMARTSException{	
		boolean sfa = true;

//		for(String smart : chemicalClassDefinitions.get(ChemicalClassName.SATURATED_FATTY_ACID).get("smarts")){
//			SmartsPatternCDK pattern = new SmartsPatternCDK(smart);	
//			if(!(pattern.hasSMARTSPattern(molecule)>0)){
//				sfa = false;
//				break;
//			}
//		}	
//		
//		for(String negativeSmart : chemicalClassDefinitions.get(ChemicalClassName.ETHER_LIPID).get("negativeSmarts")){
//			SmartsPatternCDK npattern = new SmartsPatternCDK(negativeSmart);	
//			if(npattern.hasSMARTSPattern(molecule)>0){
//				sfa = false;
//				break;
//			}
//		}
//		
//		// As per observation, 
		
		
		
		return sfa;
	}
	
	public static boolean isAlphaHydroxyFattyAcid(IAtomContainer molecule){	
		
		return false;
	}
	public static boolean isBetaHydroxyFattyAcid(IAtomContainer molecule){	
		
		return false;
	}
	public static boolean isOmegaHydroxyFattyAcid(IAtomContainer molecule){	
		
		return false;
	}
	public static boolean isEtherLipid(IAtomContainer molecule) throws SMARTSException{	
		
		boolean b = true;
		
		for(String smart : chemicalClassDefinitions.get(StructuralClassName.ETHER_LIPID).get("smarts")){
			SmartsPatternCDK pattern = new SmartsPatternCDK(smart);	
			if(!(pattern.hasSMARTSPattern(molecule)>0)){
				b = false;
				break;
			}
		}					
		for(String negativeSmart : chemicalClassDefinitions.get(StructuralClassName.ETHER_LIPID).get("negativeSmarts")){
			SmartsPatternCDK npattern = new SmartsPatternCDK(negativeSmart);	
			if(npattern.hasSMARTSPattern(molecule)>0){
				b = false;
				break;
			}
		}
		
		return b;

	}	

	public static boolean isGlyceroLipid(IAtomContainer molecule) throws SMARTSException{	
		
		boolean b = true;
		
		for(String smart : chemicalClassDefinitions.get(StructuralClassName.GLYCEROLIPID).get("smarts")){
			SmartsPatternCDK pattern = new SmartsPatternCDK(smart);	
			if(!(pattern.hasSMARTSPattern(molecule)>0)){
				b = false;
				break;
			}
		}					
		for(String negativeSmart : chemicalClassDefinitions.get(StructuralClassName.GLYCEROLIPID).get("negativeSmarts")){
			SmartsPatternCDK npattern = new SmartsPatternCDK(negativeSmart);	
			if(npattern.hasSMARTSPattern(molecule)>0){
				b = false;
				break;
			}
		}
		
		return b;

	}
	
	public static boolean isGlycerophosphoLipid(IAtomContainer molecule) throws SMARTSException{	
		
		boolean b = true;
		
		for(String smart : chemicalClassDefinitions.get(StructuralClassName.GLYCEROPHOSPHOLIPID).get("smarts")){
			SmartsPatternCDK pattern = new SmartsPatternCDK(smart);	
			if(!(pattern.hasSMARTSPattern(molecule)>0)){
				b = false;
				break;
			}
		}					
		for(String negativeSmart : chemicalClassDefinitions.get(StructuralClassName.GLYCEROPHOSPHOLIPID).get("negativeSmarts")){
			SmartsPatternCDK npattern = new SmartsPatternCDK(negativeSmart);	
			if(npattern.hasSMARTSPattern(molecule)>0){
				b = false;
				break;
			}
		}
		
		return b;

	}	
	
	public static boolean isSphingoLipid(IAtomContainer molecule) throws SMARTSException{	
		
		boolean b = true;
		
		for(String smart : chemicalClassDefinitions.get(StructuralClassName.SPHINGOLIPID).get("smarts")){
			SmartsPatternCDK pattern = new SmartsPatternCDK(smart);	
			if(!(pattern.hasSMARTSPattern(molecule)>0)){
				b = false;
				break;
			}
		}					
		for(String negativeSmart : chemicalClassDefinitions.get(StructuralClassName.SPHINGOLIPID).get("negativeSmarts")){
			SmartsPatternCDK npattern = new SmartsPatternCDK(negativeSmart);	
			if(npattern.hasSMARTSPattern(molecule)>0){
				b = false;
				break;
			}
		}
		
		return b;

	}	

	
	
	
	
	public static void main(String[] args) throws SMARTSException, CloneNotSupportedException, CDKException{
		StructuralClassFinder ccf = new StructuralClassFinder();
		IChemObjectBuilder 	builder = SilentChemObjectBuilder.getInstance();
		SmilesParser	smiParser		= new SmilesParser(builder);
		IAtomContainer molecule = smiParser.parseSmiles("CCCCCCCCCCCCCC(O)[C@H](COP(O)(O)=O)NC(=O)CCCCCCCCCCC");
		System.out.println(ccf.findStructuralClass(molecule));
	}
	
}
