package phase2utils;

import org.openscience.cdk.interfaces.IAtom;
import org.openscience.cdk.interfaces.IAtomContainer;

public class UniqueIDFunctionSet {
	/**
	 * This function will assign an uniqueID for each atom within the molecule
	 * @param oneMole
	 */
	public static void assignUniqueID(IAtomContainer oneMole){
		int uniqueID = 1;
		for(int i = 0; i < oneMole.getAtomCount(); i++){
			oneMole.getAtom(i).setProperty("UniqueID", uniqueID);
			uniqueID++;
		}
	}	
	
	/**
	 * This function will assign the given uid to the given atom within the structure
	 */
	public static void assingUID_toAtom(IAtom oneAtom, Integer uid){
		oneAtom.setProperty("UniqueID", uid);
	}
	
	
	/**
	 * This function will return the uniqueID of the input atom, based on its UniqueID property
	 * @param oneAtom
	 * @return
	 * @throws Exception
	 */
	public static Integer getUniqID(IAtom oneAtom) throws Exception{
		Integer uniqueID = oneAtom.getProperty("UniqueID");
		//if(uniqueID==null) throw new Exception("The Atom doesn't have the UniqueID property");
		return uniqueID;
	}
	/**
	 * This function will get the IAtom based having the input uniqueID within the input molecule
	 * @param uniqueID
	 * @param oneMole
	 * @return
	 * @throws Exception
	 */
	public static IAtom getAtomByUniqueID(Integer uniqueID, IAtomContainer oneMole) throws Exception{
		for(int i = 0; i < oneMole.getAtomCount(); i++){
			IAtom oneAtom = oneMole.getAtom(i);
			Integer tempID = oneAtom.getProperty("UniqueID");
			if(tempID == uniqueID) return oneAtom;
		}
		throw new Exception("Cannot find the Atom having the input ID");
	}
}
