package sulfation;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.Random;

import org.openscience.cdk.Atom;
import org.openscience.cdk.Bond;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.interfaces.IAtom;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;
import org.openscience.cdk.interfaces.IBond;
import org.openscience.cdk.interfaces.IChemObjectBuilder;
import org.openscience.cdk.io.SDFWriter;
import org.openscience.cdk.io.iterator.IteratingSDFReader;
import org.openscience.cdk.io.listener.PropertiesListener;
import org.openscience.cdk.isomorphism.Mappings;
import org.openscience.cdk.silent.SilentChemObjectBuilder;
import org.openscience.cdk.smarts.Smarts;
import org.openscience.cdk.smarts.SmartsPattern;
import org.openscience.cdk.smiles.SmiFlavor;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import createSomDataset.CollectData_convertSMIRKS_to_SDF;

public class CreateDataset {
	SmilesParser sp = new SmilesParser(DefaultChemObjectBuilder.getInstance());
	SmilesGenerator sg = new SmilesGenerator(SmiFlavor.Isomeric);
	String SULFATE_SMARTS = "[H][#8]S([#8])(=O)=O";
	public static void main(String[] args) throws Exception{
		String inputPath = "G:/PhaseIIPredictor/Sulfation/Sulfation_HMDB.tsv";
		String outputPath = "G:/PhaseIIPredictor/Sulfation/Sulfation_HMDB.sdf";
		CreateDataset cd = new CreateDataset();
		IAtomContainerSet molecules = cd.readTsv(inputPath);
		//System.out.println(molecules.getAtomContainerCount());
		//IAtomContainerSet substrates = cd.getData(molecules);
		//cd.writeSDF(substrates, outputPath);
		cd.generateThreeDatasets(molecules);
		cd.generateDataWithNS();
	}
	
	public void generateDataWithNS() throws Exception{
		IAtomContainerSet non_substrates = getNSMolecules();
		int numInstaces = 20;
		ArrayList<Integer> valdiation_20 = getRandomList(numInstaces, non_substrates.getAtomContainerCount(), new ArrayList<Integer>());
		IAtomContainerSet validationSet = getMoleculesByIndices(non_substrates, valdiation_20);
		ArrayList<Integer> test_20 = getRandomList(numInstaces, non_substrates.getAtomContainerCount(), valdiation_20);
		IAtomContainerSet testSet = getMoleculesByIndices(non_substrates, test_20);
		ArrayList<Integer> training_num = new ArrayList<>();
		for(int i = 0; i < non_substrates.getAtomContainerCount(); i++) {
			if(!valdiation_20.contains(i) && !test_20.contains(i) && !training_num.contains(i)) training_num.add(i);
		}
		IAtomContainerSet trainingSet = getMoleculesByIndices(non_substrates, training_num);
		String inputPath_train = "G:/PhaseIIPredictor/Sulfation/Sulfation_training.sdf";
		String inputPath_validation = "G:/PhaseIIPredictor/Sulfation/Sulfation_validation.sdf";
		String inputPath_test = "G:/PhaseIIPredictor/Sulfation/Sulfation_test.sdf";
		String outputPath_validation_NS = "G:/PhaseIIPredictor/Sulfation/withNS/Sulfation_validation.sdf";
		String outputPath_test_NS = "G:/PhaseIIPredictor/Sulfation/withNS/Sulfation_test.sdf";
		String outputPath_training_NS = "G:/PhaseIIPredictor/Sulfation/withNS/Sulfation_training.sdf";
		IAtomContainerSet train = readSDFFile(inputPath_train);		
		IAtomContainerSet validation = readSDFFile(inputPath_validation);
		IAtomContainerSet test = readSDFFile(inputPath_test);
		train.add(trainingSet);
		validation.add(validationSet);
		test.add(testSet);
		writeSDF(train, outputPath_training_NS);
		writeSDF(validation, outputPath_validation_NS);
		writeSDF(test,outputPath_test_NS);
		
	}
	
	public void generateThreeDatasets(IAtomContainerSet metabolites) throws Exception{
		int numInstaces = 20;
		metabolites = keepValidMetabolites(metabolites);
		ArrayList<Integer> valdiation_20 = getRandomList(numInstaces, metabolites.getAtomContainerCount(), new ArrayList<Integer>());
		IAtomContainerSet validationSet_metabolite = getMoleculesByIndices(metabolites, valdiation_20);
		ArrayList<Integer> test_20 = getRandomList(numInstaces, metabolites.getAtomContainerCount(), valdiation_20);
		IAtomContainerSet testSet_metabolite = getMoleculesByIndices(metabolites, test_20);
		ArrayList<Integer> training_num = new ArrayList<>();
		for(int i = 0; i < metabolites.getAtomContainerCount(); i++) {
			if(!valdiation_20.contains(i) && !test_20.contains(i) && !training_num.contains(i)) training_num.add(i);
		}
		IAtomContainerSet trainingSet_metabolite = getMoleculesByIndices(metabolites, training_num);
		IAtomContainerSet validationSet = getData(validationSet_metabolite);
		IAtomContainerSet testSet = getData(testSet_metabolite);
		IAtomContainerSet trainingSet = getData(trainingSet_metabolite);
		String outputPath_validation_metabolite = "G:/PhaseIIPredictor/Sulfation/Sulfation_validation_metabolite.sdf";
		String outputPath_test_metabolite = "G:/PhaseIIPredictor/Sulfation/Sulfation_test_metabolite.sdf";
		String outputPath_training_metabolite = "G:/PhaseIIPredictor/Sulfation/Sulfation_training_metabolite.sdf";
		String outputPath_validation = "G:/PhaseIIPredictor/Sulfation/Sulfation_validation.sdf";
		String outputPath_test = "G:/PhaseIIPredictor/Sulfation/Sulfation_test.sdf";
		String outputPath_training = "G:/PhaseIIPredictor/Sulfation/Sulfation_training.sdf";
		writeSDF(validationSet_metabolite, outputPath_validation_metabolite);
		writeSDF(testSet_metabolite, outputPath_test_metabolite);
		writeSDF(trainingSet_metabolite, outputPath_training_metabolite);	
		writeSDF(validationSet, outputPath_validation);
		writeSDF(testSet, outputPath_test);
		writeSDF(trainingSet, outputPath_training);
		
	}
	public IAtomContainerSet getData(IAtomContainerSet molecules) throws Exception{
		IAtomContainerSet substrates = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		//SmartsPattern smp = SmartsPattern.create(this.SULFATE_SMARTS);
		for(int i = 0; i < molecules.getAtomContainerCount(); i++) {
			IAtomContainer oneMole = molecules.getAtomContainer(i);
			IAtomContainer clean = generateMoleculeForDB(oneMole);
			if(clean.isEmpty()) continue;
			substrates.addAtomContainer(clean);
		}
		return substrates;
	}
	
	public IAtomContainerSet keepValidMetabolites(IAtomContainerSet molecules) throws Exception{
		IAtomContainerSet result = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		for(int i = 0; i < molecules.getAtomContainerCount(); i++) {
			IAtomContainer  substrate = generateMoleculeForDB(molecules.getAtomContainer(i));
			if(substrate!=null && !substrate.isEmpty()) {
				result.addAtomContainer(molecules.getAtomContainer(i));
			}
		}
		return result;
	}
	
	public IAtomContainer generateMoleculeForDB(IAtomContainer product) throws Exception{
		IAtomContainer oneMole = product.clone();
		IAtomContainer result = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainer.class);
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(oneMole);
		AtomContainerManipulator.convertImplicitToExplicitHydrogens(oneMole);
		SmartsPattern smp = SmartsPattern.create(this.SULFATE_SMARTS);
		smp.setPrepare(false);
		Mappings mapped = smp.matchAll(oneMole);
		//System.out.println(this.sg.create(oneMole));
		if(mapped.uniqueAtoms().countUnique() == 0) {
			System.out.println(this.sg.create(product));
			System.out.println("Invalid");
			return result;
		}
		ArrayList<Integer> matchedAtomsList = new ArrayList<>();
		for(int[] oneMap : mapped.uniqueAtoms()) {
			for(int i = 0; i < oneMap.length; i++) {
				Integer idx = oneMap[i];
				if(!matchedAtomsList.contains(idx)) matchedAtomsList.add(idx);
			}
		}
		for(int i = 0; i < oneMole.getAtomCount(); i++) {
			Integer idx = oneMole.indexOf(oneMole.getAtom(i));
			if(!matchedAtomsList.contains(idx)) {
				if(!result.contains(oneMole.getAtom(i))) result.addAtom(oneMole.getAtom(i));
			}
		}
		for(int i = 0; i < result.getAtomCount(); i++) {
			IAtom workOnAtom = result.getAtom(i);						
			for(int j = i; j < result.getAtomCount(); j++) {
				IBond oneBond = oneMole.getBond(workOnAtom, result.getAtom(j));
				if(oneBond != null && !result.contains(oneBond)) {
					result.addBond(oneBond);
				}
			}
		}
		//System.out.println(this.sg.create(result));
		//boolean processed = false;
		for(int i = 0; i < result.getAtomCount(); i++) {			
			IAtom workOnAtom = result.getAtom(i);	
			if(workOnAtom.getProperty("JointAtom") != null || workOnAtom.getProperty("FakeAtom") != null) continue;
			List<IAtom> neighborList = oneMole.getConnectedAtomsList(workOnAtom);
			for(int j = 0; j < neighborList.size(); j++) {
				if(matchedAtomsList.contains(oneMole.indexOf(neighborList.get(j)))) {	
					IAtom jointAtom = neighborList.get(j);
					jointAtom.setProperty("JointAtom", true);
					result.addAtom(jointAtom);
					result.addBond(oneMole.getBond(jointAtom, workOnAtom));
					IAtom fakeAtom = new Atom("H");
					fakeAtom.setImplicitHydrogenCount(0);
					IBond fakeBond = new Bond(jointAtom, fakeAtom, IBond.Order.SINGLE);
					fakeAtom.setProperty("FakeAtom", true);
					result.addAtom(fakeAtom);
					result.addBond(fakeBond);
					//processed = true;
					break;
					
				}
			}
			//int check = workOnAtom.getImplicitHydrogenCount();
			//check = 0;
		}

		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(result);
		AtomContainerManipulator.suppressHydrogens(result);
		int count = 0;
		String target_atom_property =  "None";
		for(int i = 0; i < result.getAtomCount(); i++) {
			//System.out.println(result.indexOf(result.getAtom(i)));
			if(result.getAtom(i).getProperty("JointAtom") != null) {
				count++;
				if(target_atom_property.equals("None")) target_atom_property = "<" + (i + 1) + "," + result.getAtom(i).getSymbol() + ";SULT;R" + count + ">";
				else {
					target_atom_property =  target_atom_property + "\n" + "<" + (i + 1) + "," + oneMole.getAtom(i).getSymbol() + ";SULT;R" + count + ">";
				}
				

			}
			
		}	
		result.setProperty("BoM", target_atom_property);
		//System.out.println(this.sg.create(result));
		return result;
	}
	
	public IAtomContainerSet readTsv(String inputPath) throws Exception{
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		ArrayList<String> inchikey_visited = new ArrayList<>();
		BufferedReader br = new BufferedReader(new FileReader(new File(inputPath)));
		String oneLine = br.readLine();//Title line
		while((oneLine = br.readLine()) != null) {
			String[] parseLine = oneLine.split("\t");
			String smiles = parseLine[2];
			if(smiles.contains(".")) continue;
			String inchikey = parseLine[3];
			IAtomContainer oneMole = this.sp.parseSmiles(smiles);
			for(int i = 0; i < oneMole.getAtomCount(); i++) {
				if(oneMole.getAtom(i).getSymbol().equals("C")) { 
					if(!inchikey_visited.contains(inchikey)) {
						resultSet.addAtomContainer(oneMole);
						inchikey_visited.add(inchikey);					
					}
					break;
				}
			}
			
		}
		br.close();
		return resultSet;
	}
	
	public void writeSDF(IAtomContainerSet molecules, String outputPath) throws Exception{
		SDFWriter sdfw = new SDFWriter(new FileOutputStream(outputPath));
		sdfw.write(molecules);
		sdfw.close();
	}
	
	/**
	 * Generate a list of indices of items to get random examples from the dataset
	 * @param numerInstances
	 * @param maxValue
	 * @return
	 */
	public ArrayList<Integer> getRandomList(int numerInstances, int maxValue, ArrayList<Integer> notUse) {
		Random rng = new Random();
		ArrayList<Integer> generatedList = new ArrayList<>();
		while(generatedList.size() < numerInstances) {
			int oneInstance = rng.nextInt(maxValue);
			if(!generatedList.contains(oneInstance) && !notUse.contains(oneInstance)) generatedList.add(oneInstance);
		}
		return generatedList;
	}
	/**
	 * Used to generate validation and test sets specificed by the ArrayList<Integer> indices
	 * @param molecules
	 * @param indices
	 * @return
	 */
	public IAtomContainerSet getMoleculesByIndices(IAtomContainerSet molecules, ArrayList<Integer> indices) {
		IAtomContainerSet resultSet = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		for(int i = 0; i < indices.size(); i++) {
			resultSet.addAtomContainer(molecules.getAtomContainer(indices.get(i)));
		}
		return resultSet;
		
		
	}
	
	
	public IAtomContainerSet readSDFFile(String pathToInputFile) throws Exception {
		IChemObjectBuilder bldr = SilentChemObjectBuilder.getInstance();
		IteratingSDFReader sdfr = new IteratingSDFReader(new FileReader(pathToInputFile),
				bldr);
		Properties prop = new Properties();
		//prop.setProperty("ForceReadAs3DCoordinates", "true");
		PropertiesListener listener = new PropertiesListener(prop);
		sdfr.addChemObjectIOListener(listener);
		sdfr.customizeJob();
		IAtomContainerSet MOLS = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		while (sdfr.hasNext())MOLS.addAtomContainer(sdfr.next());
		sdfr.close();
		return MOLS;

	}
	/**
	 * This function will prepare the non-substrates for UGT conjugation reaction and store them into an IAtomContainerSet
	 * @return
	 * @throws Exception
	 */
	public IAtomContainerSet getNSMolecules() throws Exception{
		String inputPath_NS = "G:/PhaseIIPredictor/UGT/withNS/non-substrates.csv";
		BufferedReader br = new BufferedReader(new FileReader(new File(inputPath_NS)));
		IAtomContainerSet non_substrates = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		String oneLine = br.readLine(); //title line
		while((oneLine = br.readLine()) != null) {
			String smiles = oneLine.split(",")[1];
			try {
				IAtomContainer oneMole = this.sp.parseSmiles(smiles);
				non_substrates.addAtomContainer(oneMole);	
			}catch (Exception e) {
				System.out.println(oneLine);
			}
		}
		br.close();
		return non_substrates;
	}
	
	
	
}
